@@ -1,14 +1,20 @@
 """
 Brain method runs framework.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
-from fiftyone.core.runs import Run, RunConfig, RunInfo, RunResults
+from fiftyone.core.runs import (
+    BaseRun,
+    BaseRunConfig,
+    BaseRunInfo,
+    BaseRunResults,
+)
+from fiftyone.core.odm import patch_brain_runs
 
 
-class BrainInfo(RunInfo):
+class BrainInfo(BaseRunInfo):
     """Information about an brain method that has been run on a dataset.
 
     Args:
@@ -22,7 +28,7 @@ class BrainInfo(RunInfo):
         return BrainMethodConfig
 
 
-class BrainMethodConfig(RunConfig):
+class BrainMethodConfig(BaseRunConfig):
     """Base class for configuring :class:`BrainMethod` instances.
 
     Args:
@@ -30,10 +36,16 @@ class BrainMethodConfig(RunConfig):
             their parsing
     """
 
-    pass
+    @property
+    def type(self):
+        return "brain"
+
+    @property
+    def method(self):
+        return None
 
 
-class BrainMethod(Run):
+class BrainMethod(BaseRun):
     """Base class for brain methods.
 
     Args:
@@ -56,8 +68,12 @@ class BrainMethod(Run):
     def _results_cache_field(cls):
         return "_brain_cache"
 
+    @classmethod
+    def _patch_function(cls):
+        return patch_brain_runs
 
-class BrainResults(RunResults):
+
+class BrainResults(BaseRunResults):
     """Base class for brain method results."""
 
     pass
