@@ -1,18 +1,14 @@
 """
 FiftyOne config.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
+
 import logging
 import os
-
-try:
-    from importlib import metadata as importlib_metadata  # Python 3.8
-except ImportError:
-    import importlib_metadata  # Python < 3.8
-
+from importlib import metadata
 import pytz
 
 import eta
@@ -67,6 +63,12 @@ class FiftyOneConfig(EnvConfig):
         if d is None:
             d = {}
 
+        self.database_compressor = self.parse_string(
+            d,
+            "database_compressor",
+            env_var="FIFTYONE_DATABASE_COMPRESSOR",
+            default=None,
+        )
         self.database_uri = self.parse_string(
             d, "database_uri", env_var="FIFTYONE_DATABASE_URI", default=None
         )
@@ -112,6 +114,24 @@ class FiftyOneConfig(EnvConfig):
         self.plugins_dir = self.parse_string(
             d, "plugins_dir", env_var="FIFTYONE_PLUGINS_DIR", default=None
         )
+        self.plugins_cache_enabled = self.parse_bool(
+            d,
+            "plugins_cache_enabled",
+            env_var="FIFTYONE_PLUGINS_CACHE_ENABLED",
+            default=False,
+        )
+        self.operator_timeout = self.parse_int(
+            d,
+            "operator_timeout",
+            env_var="FIFTYONE_OPERATOR_TIMEOUT",
+            default=600,  # 600 seconds (10 minutes)
+        )
+        self.allow_legacy_orchestrators = self.parse_bool(
+            d,
+            "allow_legacy_orchestrators",
+            env_var="FIFTYONE_ALLOW_LEGACY_ORCHESTRATORS",
+            default=False,
+        )
         self.dataset_zoo_manifest_paths = self.parse_path_array(
             d,
             "dataset_zoo_manifest_paths",
@@ -142,6 +162,30 @@ class FiftyOneConfig(EnvConfig):
             env_var="FIFTYONE_DEFAULT_BATCH_SIZE",
             default=None,
         )
+        self.default_batcher = self.parse_string(
+            d,
+            "default_batcher",
+            env_var="FIFTYONE_DEFAULT_BATCHER",
+            default="latency",
+        )
+        self.batcher_static_size = self.parse_int(
+            d,
+            "batcher_static_size",
+            env_var="FIFTYONE_BATCHER_STATIC_SIZE",
+            default=100,
+        )
+        self.batcher_target_size_bytes = self.parse_int(
+            d,
+            "batcher_target_size_bytes",
+            env_var="FIFTYONE_BATCHER_TARGET_SIZE_BYTES",
+            default=2**20,
+        )
+        self.batcher_target_latency = self.parse_number(
+            d,
+            "batcher_target_latency",
+            env_var="FIFTYONE_BATCHER_TARGET_LATENCY",
+            default=0.2,
+        )
         self.default_sequence_idx = self.parse_string(
             d,
             "default_sequence_idx",
@@ -172,12 +216,6 @@ class FiftyOneConfig(EnvConfig):
             env_var="FIFTYONE_DEFAULT_APP_ADDRESS",
             default="localhost",
         )
-        self.desktop_app = self.parse_bool(
-            d,
-            "desktop_app",
-            env_var="FIFTYONE_DESKTOP_APP",
-            default=False,
-        )
         self.logging_level = self.parse_string(
             d,
             "logging_level",
@@ -206,7 +244,42 @@ class FiftyOneConfig(EnvConfig):
         self.timezone = self.parse_string(
             d, "timezone", env_var="FIFTYONE_TIMEZONE", default=None
         )
-
+        self.default_parallelization_method = self.parse_string(
+            d,
+            "default_parallelization_method",
+            env_var="FIFTYONE_DEFAULT_PARALLELIZATION_METHOD",
+            default=None,
+        )
+        self.default_thread_pool_workers = self.parse_int(
+            d,
+            "default_thread_pool_workers",
+            env_var="FIFTYONE_DEFAULT_THREAD_POOL_WORKERS",
+            default=None,
+        )
+        self.default_process_pool_workers = self.parse_int(
+            d,
+            "default_process_pool_workers",
+            env_var="FIFTYONE_DEFAULT_PROCESS_POOL_WORKERS",
+            default=None,
+        )
+        self.max_thread_pool_workers = self.parse_int(
+            d,
+            "max_thread_pool_workers",
+            env_var="FIFTYONE_MAX_THREAD_POOL_WORKERS",
+            default=None,
+        )
+        self.max_process_pool_workers = self.parse_int(
+            d,
+            "max_process_pool_workers",
+            env_var="FIFTYONE_MAX_PROCESS_POOL_WORKERS",
+            default=None,
+        )
+        self.execution_cache_enabled = self.parse_bool(
+            d,
+            "execution_cache_enabled",
+            env_var="FIFTYONE_EXECUTION_CACHE_ENABLED",
+            default=True,
+        )
         self._init()
 
     @property
@@ -240,6 +313,12 @@ class FiftyOneConfig(EnvConfig):
                 self.default_dataset_dir, "__models__"
             )
 
+        if self.plugins_dir is None:
+            self.plugins_dir = os.path.join(
+                self.default_dataset_dir,
+                "__plugins__",
+            )
+
         if self.default_ml_backend is None:
             installed_packages = _get_installed_packages()
 
@@ -302,7 +381,19 @@ class AppConfig(EnvConfig):
             default="viridis",
         )
         self.grid_zoom = self.parse_int(
-            d, "grid_zoom", env_var="FIFTYONE_APP_GRID_ZOOM", default=5
+            d, "grid_zoom", env_var="FIFTYONE_APP_GRID_ZOOM", default=7
+        )
+        self.enable_query_performance = self.parse_bool(
+            d,
+            "enable_query_performance",
+            env_var="FIFTYONE_APP_ENABLE_QUERY_PERFORMANCE",
+            default=True,
+        )
+        self.default_query_performance = self.parse_bool(
+            d,
+            "default_query_performance",
+            env_var="FIFTYONE_APP_DEFAULT_QUERY_PERFORMANCE",
+            default=True,
         )
         self.loop_videos = self.parse_bool(
             d,
@@ -310,6 +401,24 @@ class AppConfig(EnvConfig):
             env_var="FIFTYONE_APP_LOOP_VIDEOS",
             default=False,
         )
+        self.max_query_time = self.parse_int(
+            d,
+            "max_query_time",
+            env_var="FIFTYONE_APP_MAX_QUERY_TIME",
+            default=60,
+        )
+        self.media_fallback = self.parse_bool(
+            d,
+            "media_fallback",
+            env_var="FIFTYONE_APP_MEDIA_FALLBACK",
+            default=False,
+        )
+        self.disable_frame_filtering = self.parse_bool(
+            d,
+            "disable_frame_filtering",
+            env_var="FIFTYONE_APP_DISABLE_FRAME_FILTERING",
+            default=False,
+        )
         self.multicolor_keypoints = self.parse_bool(
             d,
             "multicolor_keypoints",
@@ -322,6 +431,12 @@ class AppConfig(EnvConfig):
             env_var="FIFTYONE_APP_NOTEBOOK_HEIGHT",
             default=800,
         )
+        self.proxy_url = self.parse_string(
+            d,
+            "proxy_url",
+            env_var="FIFTYONE_APP_PROXY_URL",
+            default=None,
+        )
         self.show_confidence = self.parse_bool(
             d,
             "show_confidence",
@@ -352,12 +467,6 @@ class AppConfig(EnvConfig):
             env_var="FIFTYONE_APP_SHOW_TOOLTIP",
             default=True,
         )
-        self.sidebar_mode = self.parse_string(
-            d,
-            "sidebar_mode",
-            env_var="FIFTYONE_APP_SIDEBAR_MODE",
-            default="fast",
-        )
         self.theme = self.parse_string(
             d,
             "theme",
@@ -418,7 +527,7 @@ class AppConfig(EnvConfig):
         return fop.get_colormap(colorscale, n=n, hex_strs=hex_strs)
 
     def _init(self):
-        supported_color_bys = {"field", "instance", "label"}
+        supported_color_bys = {"field", "instance", "value"}
         default_color_by = "field"
         if self.color_by not in supported_color_bys:
             logger.warning(
@@ -429,17 +538,6 @@ class AppConfig(EnvConfig):
             )
             self.color_by = default_color_by
 
-        supported_sidebar_modes = {"all", "best", "fast"}
-        default_sidebar_mode = "best"
-        if self.sidebar_mode not in supported_sidebar_modes:
-            logger.warning(
-                "Invalid sidebar_mode=%s. Must be one of %s. Defaulting to '%s'",
-                self.sidebar_mode,
-                supported_sidebar_modes,
-                default_sidebar_mode,
-            )
-            self.sidebar_mode = default_sidebar_mode
-
         supported_themes = {"browser", "dark", "light"}
         default_theme = "browser"
         if self.theme not in supported_themes:
@@ -506,6 +604,10 @@ class AnnotationConfig(EnvConfig):
         )
 
         self.backends = self._parse_backends(d)
+        if self.default_backend not in self.backends:
+            self.default_backend = next(
+                iter(sorted(self.backends.keys())), None
+            )
 
     def _parse_backends(self, d):
         d = d.get("backends", {})
@@ -520,13 +622,14 @@ class AnnotationConfig(EnvConfig):
         if "FIFTYONE_ANNOTATION_BACKENDS" in env_vars:
             backends = env_vars["FIFTYONE_ANNOTATION_BACKENDS"].split(",")
 
-            # Declare new backends and omit any others not in `backends`
+            # Special syntax to append rather than override default backends
+            if "*" in backends:
+                backends = set(b for b in backends if b != "*")
+                backends |= set(self._BUILTIN_BACKENDS.keys())
+
             d = {backend: d.get(backend, {}) for backend in backends}
         else:
-            backends = sorted(self._BUILTIN_BACKENDS.keys())
-
-            # Declare builtin backends if necessary
-            for backend in backends:
+            for backend in self._BUILTIN_BACKENDS.keys():
                 if backend not in d:
                     d[backend] = {}
 
@@ -535,13 +638,13 @@ class AnnotationConfig(EnvConfig):
         # `FIFTYONE_<BACKEND>_<PARAMETER>`
         #
 
-        for backend, parameters in d.items():
+        for backend, d_backend in d.items():
             prefix = "FIFTYONE_%s_" % backend.upper()
             for env_name, env_value in env_vars.items():
                 if env_name.startswith(prefix):
                     name = env_name[len(prefix) :].lower()
                     value = _parse_env_value(env_value)
-                    parameters[name] = value
+                    d_backend[name] = value
 
         #
         # Set default parameters for builtin annotation backends
@@ -559,30 +662,173 @@ class AnnotationConfig(EnvConfig):
         return d
 
 
+class EvaluationConfig(EnvConfig):
+    """FiftyOne evaluation configuration settings."""
+
+    _BUILTIN_BACKENDS = {
+        "regression": {
+            "simple": {
+                "config_cls": "fiftyone.utils.eval.regression.SimpleEvaluationConfig",
+            },
+        },
+        "classification": {
+            "simple": {
+                "config_cls": "fiftyone.utils.eval.classification.SimpleEvaluationConfig",
+            },
+            "binary": {
+                "config_cls": "fiftyone.utils.eval.classification.BinaryEvaluationConfig",
+            },
+            "top-k": {
+                "config_cls": "fiftyone.utils.eval.classification.TopKEvaluationConfig",
+            },
+        },
+        "detection": {
+            "activitynet": {
+                "config_cls": "fiftyone.utils.eval.activitynet.ActivityNetEvaluationConfig",
+            },
+            "coco": {
+                "config_cls": "fiftyone.utils.eval.coco.COCOEvaluationConfig",
+            },
+            "open-images": {
+                "config_cls": "fiftyone.utils.eval.openimages.OpenImagesEvaluationConfig",
+            },
+        },
+        "segmentation": {
+            "simple": {
+                "config_cls": "fiftyone.utils.eval.segmentation.SimpleEvaluationConfig",
+            },
+        },
+    }
+
+    def __init__(self, d=None):
+        if d is None:
+            d = {}
+
+        self.default_regression_backend = self.parse_string(
+            d,
+            "default_regression_backend",
+            env_var="FIFTYONE_DEFAULT_REGRESSION_BACKEND",
+            default="simple",
+        )
+        self.default_classification_backend = self.parse_string(
+            d,
+            "default_classification_backend",
+            env_var="DEFAULT_FIFTYONE_CLASSIFICATION_BACKEND",
+            default="simple",
+        )
+        self.default_detection_backend = self.parse_string(
+            d,
+            "default_detection_backend",
+            env_var="DEFAULT_FIFTYONE_DETECTION_BACKEND",
+            default="coco",
+        )
+        self.default_segmentation_backend = self.parse_string(
+            d,
+            "default_segmentation_backend",
+            env_var="DEFAULT_FIFTYONE_SEGMENTATION_BACKEND",
+            default="simple",
+        )
+
+        self.regression_backends = self._parse_backends(d, "regression")
+        if self.default_regression_backend not in self.regression_backends:
+            self.default_regression_backend = next(
+                iter(sorted(self.regression_backends.keys())), None
+            )
+
+        self.classification_backends = self._parse_backends(
+            d, "classification"
+        )
+        if (
+            self.default_classification_backend
+            not in self.classification_backends
+        ):
+            self.default_classification_backend = next(
+                iter(sorted(self.classification_backends.keys())), None
+            )
+
+        self.detection_backends = self._parse_backends(d, "detection")
+        if self.default_detection_backend not in self.detection_backends:
+            self.default_detection_backend = next(
+                iter(sorted(self.detection_backends.keys())), None
+            )
+
+        self.segmentation_backends = self._parse_backends(d, "segmentation")
+        if self.default_segmentation_backend not in self.segmentation_backends:
+            self.default_segmentation_backend = next(
+                iter(sorted(self.segmentation_backends.keys())), None
+            )
+
+    def _parse_backends(self, d, type):
+        TYPE = type.upper()
+
+        d = d.get(f"{type}_backends", {})
+        env_vars = dict(os.environ)
+
+        #
+        # `FIFTYONE_{TYPE}_BACKENDS` can be used to declare which backends
+        # are exposed. This may exclude builtin backends and/or declare new
+        # backends
+        #
+
+        if f"FIFTYONE_{TYPE}_BACKENDS" in env_vars:
+            backends = env_vars[f"FIFTYONE_{TYPE}_BACKENDS"].split(",")
+
+            # Special syntax to append rather than override default backends
+            if "*" in backends:
+                backends = set(b for b in backends if b != "*")
+                backends |= set(self._BUILTIN_BACKENDS[type].keys())
+
+            d = {backend: d.get(backend, {}) for backend in backends}
+        else:
+            for backend in self._BUILTIN_BACKENDS[type].keys():
+                if backend not in d:
+                    d[backend] = {}
+
+        #
+        # Extract parameters from any environment variables of the form
+        # `FIFTYONE_{TYPE}_{BACKEND}_{PARAMETER}`
+        #
+
+        for backend, d_backend in d.items():
+            BACKEND = backend.upper()
+            prefix = f"FIFTYONE_{TYPE}_{BACKEND}_"
+            for env_name, env_value in env_vars.items():
+                if env_name.startswith(prefix):
+                    name = env_name[len(prefix) :].lower()
+                    value = _parse_env_value(env_value)
+                    d_backend[name] = value
+
+        #
+        # Set default parameters for builtin similarity backends
+        #
+
+        for backend, defaults in self._BUILTIN_BACKENDS[type].items():
+            if backend not in d:
+                continue
+
+            d_backend = d[backend]
+            for name, value in defaults.items():
+                if name not in d_backend:
+                    d_backend[name] = value
+
+        return d
+
+
 def locate_config():
     """Returns the path to the :class:`FiftyOneConfig` on disk.
 
     The default location is ``~/.fiftyone/config.json``, but you can override
     this path by setting the ``FIFTYONE_CONFIG_PATH`` environment variable.
 
-    Note that a config file may not actually exist on disk in the default
-    location, in which case the default config settings will be used.
+    Note that a config file may not actually exist on disk.
 
     Returns:
         the path to the :class:`FiftyOneConfig` on disk
-
-    Raises:
-        OSError: if the config path has been customized but the file does not
-            exist on disk
     """
     if "FIFTYONE_CONFIG_PATH" not in os.environ:
         return foc.FIFTYONE_CONFIG_PATH
 
-    config_path = os.environ["FIFTYONE_CONFIG_PATH"]
-    if not os.path.isfile(config_path):
-        raise OSError("Config file '%s' not found" % config_path)
-
-    return config_path
+    return os.environ["FIFTYONE_CONFIG_PATH"]
 
 
 def locate_app_config():
@@ -592,24 +838,15 @@ def locate_app_config():
     override this path by setting the ``FIFTYONE_APP_CONFIG_PATH`` environment
     variable.
 
-    Note that a config file may not actually exist on disk in the default
-    location, in which case the default config settings will be used.
+    Note that the file may not actually exist.
 
     Returns:
         the path to the :class:`AppConfig` on disk
-
-    Raises:
-        OSError: if the App config path has been customized but the file does
-            not exist on disk
     """
     if "FIFTYONE_APP_CONFIG_PATH" not in os.environ:
         return foc.FIFTYONE_APP_CONFIG_PATH
 
-    config_path = os.environ["FIFTYONE_APP_CONFIG_PATH"]
-    if not os.path.isfile(config_path):
-        raise OSError("App config file '%s' not found" % config_path)
-
-    return config_path
+    return os.environ["FIFTYONE_APP_CONFIG_PATH"]
 
 
 def locate_annotation_config():
@@ -619,24 +856,33 @@ def locate_annotation_config():
     override this path by setting the ``FIFTYONE_ANNOTATION_CONFIG_PATH``
     environment variable.
 
-    Note that a config file may not actually exist on disk in the default
-    location, in which case the default config settings will be used.
+    Note that a config file may not actually exist on disk.
 
     Returns:
         the path to the :class:`AnnotationConfig` on disk
-
-    Raises:
-        OSError: if the annotation config path has been customized but the file
-            does not exist on disk
     """
     if "FIFTYONE_ANNOTATION_CONFIG_PATH" not in os.environ:
         return foc.FIFTYONE_ANNOTATION_CONFIG_PATH
 
-    config_path = os.environ["FIFTYONE_ANNOTATION_CONFIG_PATH"]
-    if not os.path.isfile(config_path):
-        raise OSError("Annotation config file '%s' not found" % config_path)
+    return os.environ["FIFTYONE_ANNOTATION_CONFIG_PATH"]
 
-    return config_path
+
+def locate_evaluation_config():
+    """Returns the path to the :class:`EvaluationConfig` on disk.
+
+    The default location is ``~/.fiftyone/evaluation_config.json``, but you can
+    override this path by setting the ``FIFTYONE_EVALUATION_CONFIG_PATH``
+    environment variable.
+
+    Note that a config file may not actually exist on disk.
+
+    Returns:
+        the path to the :class:`EvaluationConfig` on disk
+    """
+    if "FIFTYONE_EVALUATION_CONFIG_PATH" not in os.environ:
+        return foc.FIFTYONE_EVALUATION_CONFIG_PATH
+
+    return os.environ["FIFTYONE_EVALUATION_CONFIG_PATH"]
 
 
 def load_config():
@@ -678,6 +924,19 @@ def load_annotation_config():
     return AnnotationConfig()
 
 
+def load_evaluation_config():
+    """Loads the FiftyOne evaluation config.
+
+    Returns:
+        an :class:`EvaluationConfig` instance
+    """
+    evaluation_config_path = locate_evaluation_config()
+    if os.path.isfile(evaluation_config_path):
+        return EvaluationConfig.from_json(evaluation_config_path)
+
+    return EvaluationConfig()
+
+
 def _parse_env_value(value):
     try:
         return int(value)
@@ -706,9 +965,7 @@ def _parse_env_value(value):
 
 def _get_installed_packages():
     try:
-        return set(
-            d.metadata["Name"] for d in importlib_metadata.distributions()
-        )
+        return set(d.metadata["Name"] for d in metadata.distributions())
     except:
         logger.debug("Failed to get installed packages")
         return set()
