@@ -1,7 +1,7 @@
 """
 Base classes for objects that are backed by database documents.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -13,6 +13,7 @@ import eta.core.serial as etas
 import eta.core.utils as etau
 
 import fiftyone.core.labels as fol
+import fiftyone.core.odm as foo
 from fiftyone.core.singletons import DocumentSingleton
 
 
@@ -188,7 +189,7 @@ class _Document(object):
 
         Raises:
             ValueError: if ``field_name`` is not an allowed field name
-            AttirubteError: if the field does not exist and ``create == False``
+            AttributeError: if the field does not exist and ``create == False``
         """
         self._doc.set_field(
             field_name,
@@ -240,28 +241,46 @@ class _Document(object):
         """
         self._doc.clear_field(field_name)
 
-    def iter_fields(self, include_id=False):
+    def iter_fields(self, include_id=False, include_timestamps=False):
         """Returns an iterator over the ``(name, value)`` pairs of the public
         fields of the document.
 
         Args:
             include_id (False): whether to include the ``id`` field
+            include_timestamps (False): whether to include the ``created_at``
+                and ``last_modified_at`` fields
 
         Returns:
             an iterator that emits ``(name, value)`` tuples
         """
         for field_name in self.field_names:
-            if field_name == "id" and not include_id:
+            if not include_id and field_name == "id":
+                continue
+
+            if not include_timestamps and field_name in (
+                "created_at",
+                "last_modified_at",
+            ):
                 continue
 
             yield field_name, self.get_field(field_name)
 
+    def _iter_label_fields(self):
+        for name, value in self.iter_fields():
+            if isinstance(value, fol.Label):
+                yield name, value
+            elif isinstance(value, foo.BaseEmbeddedDocument):
+                for _name, _value in value.iter_fields():
+                    if isinstance(_value, fol.Label):
+                        yield name + "." + _name, _value
+
     def merge(
         self,
         document,
         fields=None,
         omit_fields=None,
         merge_lists=True,
+        merge_embedded_docs=False,
         overwrite=True,
         expand_schema=True,
         validate=True,
@@ -308,6 +327,9 @@ class _Document(object):
                 (when ``overwrite`` is True) or kept (when ``overwrite`` is
                 False) when their ``id`` matches a label from the provided
                 document
+            merge_embedded_docs (False): whether to merge the attributes of
+                embedded documents (True) rather than merging the entire
+                top-level field (False)
             overwrite (True): whether to overwrite (True) or skip (False)
                 existing fields and label elements
             expand_schema (True): whether to dynamically add new fields
@@ -327,7 +349,10 @@ class _Document(object):
         fields = document._parse_fields(fields=fields, omit_fields=omit_fields)
 
         for src_field, dst_field in fields.items():
-            value = document[src_field]
+            try:
+                value = document[src_field]
+            except KeyError:
+                continue
 
             if value is None:
                 continue
@@ -344,10 +369,26 @@ class _Document(object):
                     _merge_lists(curr_value, value, overwrite=overwrite)
                     continue
 
-                if field_type in fol._LABEL_LIST_FIELDS:
+                if issubclass(field_type, fol._HasLabelList):
                     _merge_labels(curr_value, value, overwrite=overwrite)
                     continue
 
+            if merge_embedded_docs:
+                field_type = type(curr_value)
+
+                if issubclass(
+                    field_type, foo.BaseEmbeddedDocument
+                ) and not issubclass(field_type, fol._HasLabelList):
+                    _merge_embedded_doc(
+                        curr_value,
+                        value,
+                        dst_field,
+                        omit_fields=omit_fields,
+                        merge_lists=merge_lists,
+                        overwrite=overwrite,
+                    )
+                    continue
+
             if (
                 not overwrite
                 and dst_field in existing_field_names
@@ -442,7 +483,8 @@ class _Document(object):
             fields = {
                 f: f
                 for f in self.field_names
-                if f not in ("id", "_dataset_id")
+                if f
+                not in ("id", "_dataset_id", "created_at", "last_modified_at")
             }
         elif etau.is_str(fields):
             fields = {fields: fields}
@@ -644,10 +686,6 @@ class DocumentView(_Document):
         excluded_fields=None,
         filtered_fields=None,
     ):
-        if selected_fields is not None and excluded_fields is not None:
-            selected_fields = selected_fields.difference(excluded_fields)
-            excluded_fields = None
-
         self._view = view
         self._selected_fields = selected_fields
         self._excluded_fields = excluded_fields
@@ -723,7 +761,7 @@ class DocumentView(_Document):
             return False
 
         sf = self._selected_fields
-        if sf is not None and field_name not in sf:
+        if sf is not None and field_name.split(".", 1)[0] not in sf:
             return False
 
         return super().has_field(field_name)
@@ -736,16 +774,14 @@ class DocumentView(_Document):
                 % (field_name, self.__class__.__name__)
             )
 
-        value = super().get_field(field_name)
-
         sf = self._selected_fields
-        if sf is not None and field_name not in sf:
+        if sf is not None and field_name.split(".", 1)[0] not in sf:
             raise AttributeError(
                 "Field '%s' was not selected on this %s"
                 % (field_name, self.__class__.__name__)
             )
 
-        return value
+        return super().get_field(field_name)
 
     def set_field(
         self,
@@ -775,12 +811,13 @@ class DocumentView(_Document):
                 dynamic=dynamic,
             )
 
+            if self._selected_fields is not None:
+                root = field_name.split(".", 1)[0]
+                self._selected_fields.add(root)
+
             if self._excluded_fields is not None:
                 self._excluded_fields.discard(field_name)
 
-            if self._selected_fields is not None:
-                self._selected_fields.add(field_name)
-
     def clear_field(self, field_name):
         # Ensures field exists
         _ = self.get_field(field_name)
@@ -790,7 +827,10 @@ class DocumentView(_Document):
     def to_dict(self, include_private=False):
         d = super().to_dict(include_private=include_private)
 
-        if self._selected_fields or self._excluded_fields:
+        if (
+            self._selected_fields is not None
+            or self._excluded_fields is not None
+        ):
             field_names = set(
                 self._get_field_names(
                     include_private=include_private,
@@ -805,7 +845,10 @@ class DocumentView(_Document):
     def to_mongo_dict(self, include_id=False):
         d = super().to_mongo_dict(include_id=include_id)
 
-        if self._selected_fields or self._excluded_fields:
+        if (
+            self._selected_fields is not None
+            or self._excluded_fields is not None
+        ):
             field_names = set(
                 self._get_field_names(include_private=True, use_db_fields=True)
             )
@@ -864,3 +907,41 @@ def _merge_labels(dst, src, overwrite=True):
     else:
         existing_ids = set(l.id for l in labels)
         labels.extend(l for l in new_labels if l.id not in existing_ids)
+
+
+def _merge_embedded_doc(
+    dst,
+    src,
+    root,
+    omit_fields=None,
+    merge_lists=True,
+    overwrite=True,
+):
+    if src is None:
+        return
+
+    for key, value in src.iter_fields():
+        if value is None:
+            continue
+
+        if omit_fields is not None and root + "." + key in omit_fields:
+            continue
+
+        try:
+            curr_value = dst[key]
+        except KeyError:
+            curr_value = None
+
+        if overwrite or curr_value is None:
+            if merge_lists:
+                field_type = type(curr_value)
+
+                if issubclass(field_type, list):
+                    _merge_lists(curr_value, value, overwrite=overwrite)
+                    continue
+
+                if issubclass(field_type, fol._HasLabelList):
+                    _merge_labels(curr_value, value, overwrite=overwrite)
+                    continue
+
+            dst[key] = value
