@@ -1,28 +1,34 @@
 """
 Labels stored in dataset samples.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
+
 import itertools
 import warnings
+from functools import partial
 
 from bson import ObjectId
 import cv2
 import numpy as np
+import scipy.ndimage as spn
+import skimage.measure as skm
+import skimage.segmentation as sks
 
 import eta.core.frameutils as etaf
 import eta.core.image as etai
 import eta.core.utils as etau
 
-from fiftyone.core.odm import DynamicEmbeddedDocument
 import fiftyone.core.fields as fof
 import fiftyone.core.metadata as fom
 import fiftyone.core.utils as fou
+from fiftyone.core.odm import DynamicEmbeddedDocument, EmbeddedDocument
 
 foue = fou.lazy_import("fiftyone.utils.eta")
 foug = fou.lazy_import("fiftyone.utils.geojson")
+foui = fou.lazy_import("fiftyone.utils.image")
 sg = fou.lazy_import(
     "shapely.geometry", callback=lambda: fou.ensure_package("shapely")
 )
@@ -35,6 +41,12 @@ class _NoDefault(object):
 no_default = _NoDefault()
 
 
+class _HasMedia(object):
+    """Mixin for :class:`Label` classes that contain a media field."""
+
+    _MEDIA_FIELD = None
+
+
 class Label(DynamicEmbeddedDocument):
     """Base class for labels.
 
@@ -327,6 +339,50 @@ class _HasID(Label):
     def _id(self):
         return ObjectId(self.id)
 
+    @_id.setter
+    def _id(self, value):
+        self.id = str(value)
+
+
+class Instance(EmbeddedDocument):
+    """A label instance.
+
+    Args:
+        id (None): the label instance ID
+    """
+
+    id = fof.ObjectIdField(default=lambda: str(ObjectId()), db_field="_id")
+
+    @property
+    def _id(self):
+        return ObjectId(self.id)
+
+    @_id.setter
+    def _id(self, value):
+        if not isinstance(value, ObjectId) and etau.is_str(value):
+            value = ObjectId(value)
+
+        self.id = value
+
+
+class _HasInstance(Label):
+    """Mixin for :class:`Label` classes that contain an instance configuration
+    via an ``instance`` attribute.
+
+    Contrary to the ``id`` field, which is unique to each label, the
+    ``instance`` field is unique to each label instance either temporally or
+    across different modalities, allowing you to identify the same logical
+    label across different samples.
+    """
+
+    @property
+    def instance_id(self):
+        """The label's instance ID, or None if it does not have one."""
+        try:
+            return self.instance.id
+        except AttributeError:
+            return None
+
 
 class _HasLabelList(object):
     """Mixin for :class:`Label` classes that contain a list of :class:`Label`
@@ -370,6 +426,7 @@ class Classifications(_HasLabelList, Label):
 
     Args:
         classifications (None): a list of :class:`Classification` instances
+        logits (None): logits associated with the labels
     """
 
     _LABEL_LIST_FIELD = "classifications"
@@ -378,7 +435,7 @@ class Classifications(_HasLabelList, Label):
     logits = fof.VectorField()
 
 
-class Detection(_HasAttributesDict, _HasID, Label):
+class Detection(_HasAttributesDict, _HasID, _HasMedia, _HasInstance, Label):
     """An object detection.
 
     Args:
@@ -391,18 +448,77 @@ class Detection(_HasAttributesDict, _HasID, Label):
         mask (None): an instance segmentation mask for the detection within
             its bounding box, which should be a 2D binary or 0/1 integer numpy
             array
+        mask_path (None):  the absolute path to the instance segmentation image
+            on disk, which should be a single-channel PNG image where any
+            non-zero values represent the instance's extent
         confidence (None): a confidence in ``[0, 1]`` for the detection
         index (None): an index for the object
+        instance (None): an instance of :class:`Instance` to link this
+            detection label to other similar labels
         attributes ({}): a dict mapping attribute names to :class:`Attribute`
             instances
     """
 
+    _MEDIA_FIELD = "mask_path"
+
     label = fof.StringField()
     bounding_box = fof.ListField(fof.FloatField())
     mask = fof.ArrayField()
+    mask_path = fof.StringField()
     confidence = fof.FloatField()
     index = fof.IntField()
 
+    @property
+    def has_mask(self):
+        """Whether this instance has a mask."""
+        return self.mask is not None or self.mask_path is not None
+
+    def get_mask(self):
+        """Returns the detection mask for this instance.
+
+        Returns:
+            a numpy array, or ``None``
+        """
+        if self.mask is not None:
+            return self.mask
+
+        if self.mask_path is not None:
+            return _read_mask(self.mask_path)
+
+        return None
+
+    def import_mask(self, update=False):
+        """Imports this instance's mask from disk to its :attr:`mask`
+        attribute.
+
+        Args:
+            update (False): whether to clear this instance's :attr:`mask_path`
+                attribute after importing
+        """
+        if self.mask_path is not None:
+            self.mask = _read_mask(self.mask_path)
+
+            if update:
+                self.mask_path = None
+
+    def export_mask(self, outpath, update=False):
+        """Exports this instance's mask to the given path.
+
+        Args:
+            outpath: the path to write the mask
+            update (False): whether to clear this instance's :attr:`mask`
+                attribute and set its :attr:`mask_path` attribute when
+                exporting in-database segmentations
+        """
+        if self.mask_path is not None:
+            etau.copy_file(self.mask_path, outpath)
+        else:
+            _write_mask(self.mask, outpath)
+
+            if update:
+                self.mask = None
+                self.mask_path = outpath
+
     def to_polyline(self, tolerance=2, filled=True):
         """Returns a :class:`Polyline` representation of this instance.
 
@@ -457,13 +573,13 @@ class Detection(_HasAttributesDict, _HasID, Label):
         Returns:
             a :class:`Segmentation`
         """
-        if self.mask is None:
+        if not self.has_mask:
             raise ValueError(
-                "Only detections with their `mask` attributes populated can "
-                "be converted to segmentations"
+                "Only detections with their `mask` or `mask_path` attribute "
+                "populated can be converted to segmentations"
             )
 
-        mask, _ = _parse_to_segmentation_inputs(mask, frame_size, None)
+        mask, target = _parse_segmentation_target(mask, frame_size, target)
         _render_instance(mask, self, target)
         return Segmentation(mask=mask)
 
@@ -570,13 +686,13 @@ class Detections(_HasLabelList, Label):
         Returns:
             a :class:`Segmentation`
         """
-        mask, labels_to_targets = _parse_to_segmentation_inputs(
+        mask, labels_to_targets = _parse_segmentation_mask_targets(
             mask, frame_size, mask_targets
         )
 
         # pylint: disable=not-an-iterable
         for detection in self.detections:
-            if detection.mask is None:
+            if not detection.has_mask:
                 msg = "Skipping detection(s) with no instance mask"
                 warnings.warn(msg)
                 continue
@@ -593,7 +709,7 @@ class Detections(_HasLabelList, Label):
         return Segmentation(mask=mask)
 
 
-class Polyline(_HasAttributesDict, _HasID, Label):
+class Polyline(_HasAttributesDict, _HasID, _HasInstance, Label):
     """A set of semantically related polylines or polygons.
 
     Args:
@@ -603,6 +719,8 @@ class Polyline(_HasAttributesDict, _HasID, Label):
             polyline
         confidence (None): a confidence in ``[0, 1]`` for the polyline
         index (None): an index for the polyline
+        instance (None): an instance of :class:`Instance` to link this
+            polyline label to other similar labels
         closed (False): whether the shapes are closed, i.e., and edge should
             be drawn from the last vertex to the first vertex of each shape
         filled (False): whether the polyline represents polygons, i.e., shapes
@@ -682,14 +800,15 @@ class Polyline(_HasAttributesDict, _HasID, Label):
             frame_size (None): the ``(width, height)`` of the segmentation
                 mask to render. This parameter has no effect if a ``mask`` is
                 provided
-            target (255): the pixel value to use to render the object
+            target (255): the pixel value or RGB hex string to use to render
+                the object
             thickness (1): the thickness, in pixels, at which to render
                 (non-filled) polylines
 
         Returns:
             a :class:`Segmentation`
         """
-        mask, _ = _parse_to_segmentation_inputs(mask, frame_size, None)
+        mask, target = _parse_segmentation_target(mask, frame_size, target)
         _render_polyline(mask, self, target, thickness)
         return Segmentation(mask=mask)
 
@@ -768,14 +887,17 @@ class Polyline(_HasAttributesDict, _HasID, Label):
         if mask.ndim > 2:
             mask = mask[:, :, 0]
 
-        points = _get_polygons(mask.astype(bool), tolerance)
+        points = _get_polygons(
+            mask.astype(bool),
+            tolerance=tolerance,
+        )
 
         return cls(
             label=label, points=points, filled=True, closed=True, **attributes
         )
 
     @classmethod
-    def from_cuboid(cls, vertices, label=None, **attributes):
+    def from_cuboid(cls, vertices, frame_size=None, label=None, **attributes):
         """Constructs a cuboid from its 8 vertices in the format below::
 
                7--------6
@@ -787,8 +909,13 @@ class Polyline(_HasAttributesDict, _HasID, Label):
             |/       |/
             0--------1
 
+        If a ``frame_size`` is provided, ``vertices`` must be absolute pixel
+        coordinates; otherwise ``vertices`` should be normalized coordinates in
+        ``[0, 1] x [0, 1]``.
+
         Args:
             vertices: a list of 8 ``(x, y)`` vertices in the above format
+            frame_size (None): the ``(width, height)`` of the frame
             label (None): the label string
             **attributes: additional arguments for the :class:`Polyline`
 
@@ -796,6 +923,9 @@ class Polyline(_HasAttributesDict, _HasID, Label):
             a :class:`Polyline`
         """
         vertices = np.asarray(vertices)
+        if frame_size is not None:
+            vertices /= np.asarray(frame_size)[np.newaxis, :]
+
         front = vertices[:4]
         back = vertices[4:]
         top = vertices[[3, 2, 6, 7], :]
@@ -804,16 +934,24 @@ class Polyline(_HasAttributesDict, _HasID, Label):
         return cls(label=label, points=faces, closed=True, **attributes)
 
     @classmethod
-    def from_rotated_box(cls, xc, yc, w, h, theta, label=None, **attributes):
+    def from_rotated_box(
+        cls, xc, yc, w, h, theta, frame_size=None, label=None, **attributes
+    ):
         """Constructs a rotated bounding box from its center, dimensions, and
         rotation.
 
+        If a ``frame_size`` is provided, the provided box coordinates must be
+        absolute pixel coordinates; otherwise they should be normalized
+        coordinates in ``[0, 1]``. Note that rotations in normalized
+        coordinates only make sense when the source aspect ratio is square.
+
         Args:
-            xc: the x-center coordinate in ``[0, 1]``
-            yc: the y-center coorindate in ``[0, 1]``
-            w: the box width in ``[0, 1]``
-            y: the box height in ``[0, 1]``
+            xc: the x-center coordinate
+            yc: the y-center coorindate
+            w: the box width
+            y: the box height
             theta: the counter-clockwise rotation of the box in radians
+            frame_size (None): the ``(width, height)`` of the frame
             label (None): the label string
             **attributes: additional arguments for the :class:`Polyline`
 
@@ -823,7 +961,11 @@ class Polyline(_HasAttributesDict, _HasID, Label):
         R = _rotation_matrix(theta)
         x = 0.5 * w * np.array([1, -1, -1, 1])
         y = 0.5 * h * np.array([1, 1, -1, -1])
-        points = (R.dot(np.stack((x, y))).T + np.array((xc, yc))).tolist()
+        points = R.dot(np.stack((x, y))).T + np.array((xc, yc))
+        if frame_size is not None:
+            points /= np.asarray(frame_size)[np.newaxis, :]
+
+        points = points.tolist()
         return cls(label=label, points=[points], closed=True, **attributes)
 
 
@@ -896,7 +1038,7 @@ class Polylines(_HasLabelList, Label):
         Returns:
             a :class:`Segmentation`
         """
-        mask, labels_to_targets = _parse_to_segmentation_inputs(
+        mask, labels_to_targets = _parse_segmentation_mask_targets(
             mask, frame_size, mask_targets
         )
 
@@ -914,7 +1056,7 @@ class Polylines(_HasLabelList, Label):
         return Segmentation(mask=mask)
 
 
-class Keypoint(_HasAttributesDict, _HasID, Label):
+class Keypoint(_HasAttributesDict, _HasID, _HasInstance, Label):
     """A list of keypoints in an image.
 
     Args:
@@ -922,6 +1064,8 @@ class Keypoint(_HasAttributesDict, _HasID, Label):
         points (None): a list of ``(x, y)`` keypoints in ``[0, 1] x [0, 1]``
         confidence (None): a list of confidences in ``[0, 1]`` for each point
         index (None): an index for the keypoints
+        instance (None): an instance of :class:`Instance` to link this
+            keypoint label to other similar labels
         attributes ({}): a dict mapping attribute names to :class:`Attribute`
             instances
     """
@@ -950,6 +1094,27 @@ class Keypoint(_HasAttributesDict, _HasID, Label):
 
         return sg.MultiPoint(points)
 
+    def apply_confidence_threshold(self, confidence_thresh):
+        """Replaces all ``points`` on this instance whose confidence are below
+        the provided threshold with ``np.nan``.
+
+        Use
+        :meth:`filter_keypoints <fiftyone.core.collections.SampleCollection.filter_keypoints`
+        to perform this operation as temporary view rather than a permanent
+        data transformation.
+
+        Args:
+            confidence_thresh: a confidence threshold
+        """
+        if self.confidence is None:
+            return
+
+        _points = np.array(self.points)
+        _confs = np.array(self.confidence)
+        _points[_confs < confidence_thresh] = np.nan
+
+        self.points = _points.tolist()
+
 
 class Keypoints(_HasLabelList, Label):
     """A list of :class:`Keypoint` instances in an image.
@@ -963,12 +1128,6 @@ class Keypoints(_HasLabelList, Label):
     keypoints = fof.ListField(fof.EmbeddedDocumentField(Keypoint))
 
 
-class _HasMedia(object):
-    """Mixin for :class:`Label` classes that contain a media field."""
-
-    _MEDIA_FIELD = None
-
-
 class Segmentation(_HasID, _HasMedia, Label):
     """A semantic segmentation for an image.
 
@@ -978,7 +1137,7 @@ class Segmentation(_HasID, _HasMedia, Label):
     Args:
         mask (None): a numpy array with integer values encoding the semantic
             labels
-        mask_path (None): the path to the segmentation image on disk
+        mask_path (None): the absolute path to the segmentation image on disk
     """
 
     _MEDIA_FIELD = "mask_path"
@@ -1010,7 +1169,6 @@ class Segmentation(_HasID, _HasMedia, Label):
         attribute.
 
         Args:
-            outpath: the path to write the map
             update (False): whether to clear this instance's :attr:`mask_path`
                 attribute after importing
         """
@@ -1114,7 +1272,12 @@ class Segmentation(_HasID, _HasMedia, Label):
         )
         return Detections(detections=detections)
 
-    def to_polylines(self, mask_targets=None, mask_types="stuff", tolerance=2):
+    def to_polylines(
+        self,
+        mask_targets=None,
+        mask_types="stuff",
+        tolerance=2,
+    ):
         """Returns a :class:`Polylines` representation of this instance.
 
         Each ``"stuff"`` class will be converted to a single :class:`Polyline`
@@ -1156,7 +1319,7 @@ class Heatmap(_HasID, _HasMedia, Label):
 
     Args:
         map (None): a 2D numpy array
-        map_path (None): the path to the heatmap image on disk
+        map_path (None): the absolute path to the heatmap image on disk
         range (None): an optional ``[min, max]`` range of the map's values. If
             None is provided, ``[0, 1]`` will be assumed if ``map`` contains
             floating point values, ``[0, 255]`` will be assumed if ``map``
@@ -1415,6 +1578,17 @@ _PATCHES_FIELDS = (
     Polylines,
 )
 
+_INDEX_FIEDS = (
+    Detection,
+    Detections,
+    Polyline,
+    Polylines,
+    Keypoint,
+    Keypoints,
+)
+
+_INSTANCE_FIELDS = (Detection, Polyline, Keypoint)
+
 _SINGLE_LABEL_TO_LIST_MAP = {
     Classification: Classifications,
     Detection: Detections,
@@ -1434,12 +1608,12 @@ _LABEL_LIST_TO_SINGLE_MAP = {
 
 def _read_mask(mask_path):
     # pylint: disable=no-member
-    return etai.read(mask_path, flag=cv2.IMREAD_UNCHANGED)
+    return foui.read(mask_path, flag=cv2.IMREAD_UNCHANGED)
 
 
 def _write_mask(mask, mask_path):
     mask = _mask_to_image(mask)
-    etai.write(mask, mask_path)
+    foui.write(mask, mask_path)
 
 
 def _transform_mask(in_mask, targets_map):
@@ -1471,8 +1645,11 @@ def _transform_mask(in_mask, targets_map):
         dtype = np.uint8
 
     out_mask = np.zeros_like(in_mask, dtype=dtype)
+    objects = _find_slices(in_mask)
     for in_val, out_val in targets_map.items():
-        out_mask[in_mask == in_val] = out_val
+        slices = objects.get(in_val, None)
+        if slices is not None:
+            out_mask[slices][in_mask[slices] == in_val] = out_val
 
     if rgb_out:
         out_mask = _int_array_to_rgb(out_mask)
@@ -1494,12 +1671,12 @@ def _mask_to_image(mask):
 
 def _read_heatmap(map_path):
     # pylint: disable=no-member
-    return etai.read(map_path, flag=cv2.IMREAD_UNCHANGED)
+    return foui.read(map_path, flag=cv2.IMREAD_UNCHANGED)
 
 
 def _write_heatmap(map, map_path, range):
     map = _heatmap_to_image(map, range)
-    etai.write(map, map_path)
+    foui.write(map, map_path)
 
 
 def _heatmap_to_image(map, range):
@@ -1521,26 +1698,53 @@ def _heatmap_to_image(map, range):
     return map.astype(np.uint8)
 
 
-def _parse_to_segmentation_inputs(mask, frame_size, mask_targets):
+def _parse_segmentation_target(mask, frame_size, target):
+    if target is not None:
+        is_rgb = fof.is_rgb_target(target)
+    else:
+        is_rgb = False
+
+    if mask is None:
+        if frame_size is None:
+            raise ValueError("Either `mask` or `frame_size` must be provided")
+
+        if target is not None and not is_rgb and target > 255:
+            dtype = int
+        else:
+            dtype = np.uint8
+
+        width, height = frame_size
+        if is_rgb:
+            mask = np.zeros((height, width, 3), dtype=dtype)
+        else:
+            mask = np.zeros((height, width), dtype=dtype)
+
+    if target is not None and is_rgb:
+        target = _hex_to_rgb(target)
+
+    return mask, target
+
+
+def _parse_segmentation_mask_targets(mask, frame_size, mask_targets):
     if mask_targets is not None:
         is_rgb = fof.is_rgb_mask_targets(mask_targets)
+    else:
+        is_rgb = False
 
     if mask is None:
         if frame_size is None:
             raise ValueError("Either `mask` or `frame_size` must be provided")
 
         if mask_targets is not None and not is_rgb and max(mask_targets) > 255:
-            dtype = np.int
+            dtype = int
         else:
             dtype = np.uint8
 
         width, height = frame_size
-        if mask_targets is not None and is_rgb:
+        if is_rgb:
             mask = np.zeros((height, width, 3), dtype=dtype)
         else:
             mask = np.zeros((height, width), dtype=dtype)
-    else:
-        height, width = mask.shape[:2]
 
     if mask_targets is not None:
         if is_rgb:
@@ -1591,7 +1795,21 @@ def _render_polyline(mask, polyline, target, thickness):
         )
 
 
-def _segmentation_to_detections(segmentation, mask_targets, mask_types):
+def _find_slices(mask):
+    """Return slices that tightly bound each unique object in `mask`."""
+    relabeled, forward, backward = sks.relabel_sequential(mask)
+    slices = spn.find_objects(relabeled)
+    return dict((backward[idx + 1], slc) for idx, slc in enumerate(slices))
+
+
+def _convert_segmentation(segmentation, mask_targets, mask_types, converter):
+    """Convert segmentation to a collection of detections, polylines, etc.
+
+    `converter(label_mask, label, label_type, offset, frame_size)` is
+    a function that returns a list of detections, polylines, etc. It
+    gets called for each value in `mask_targets`, or for all values in
+    the mask if `mask_targets` is `None`.
+    """
     if isinstance(mask_types, dict):
         default = None
     else:
@@ -1602,21 +1820,20 @@ def _segmentation_to_detections(segmentation, mask_targets, mask_types):
     is_rgb = mask.ndim == 3
 
     if is_rgb:
-        array_targets = np.unique(mask.reshape(-1, mask.shape[2]), axis=0)
-        targets = [_rgb_to_hex(t) for t in array_targets]
-    else:
-        targets = np.unique(mask)
-        array_targets = itertools.repeat(None)
+        # convert to int, like in transform_mask
+        mask = _rgb_array_to_int(mask)
+        if mask_targets is not None:
+            mask_targets = {_hex_to_int(k): v for k, v in mask_targets.items()}
 
-    detections = []
-    for target, array_target in zip(targets, array_targets):
-        if target in (0, "#000000"):
-            continue  # skip background
+    mask = mask.squeeze()
+    if mask.ndim != 2:
+        raise ValueError(f"Unsupported mask dimensions: {mask.ndim}")
 
+    objects = _find_slices(mask)
+    results = []
+    for target, slices in objects.items():
         if mask_targets is not None:
             label = mask_targets.get(target, None)
-            if is_rgb and label is None:
-                label = mask_targets.get(target.upper(), None)
 
             if label is None:
                 continue  # skip unknown target
@@ -1624,8 +1841,6 @@ def _segmentation_to_detections(segmentation, mask_targets, mask_types):
             label = str(target)
 
         label_type = mask_types.get(target, None)
-        if is_rgb and label_type is None:
-            label_type = mask_types.get(target.upper(), None)
 
         if label_type is None:
             if default is None:
@@ -1633,104 +1848,81 @@ def _segmentation_to_detections(segmentation, mask_targets, mask_types):
 
             label_type = default
 
-        if is_rgb:
-            label_mask = np.all(mask == array_target.reshape(1, 1, 3), axis=2)
-        else:
-            label_mask = mask == target
+        label_mask = mask[slices] == target
+        offset = list(s.start for s in slices)[::-1]
+        frame_size = mask.shape[:2][::-1]
 
-        if label_type == "stuff":
-            instances = [_parse_stuff_instance(label_mask)]
-        elif label_type == "thing":
-            instances = _parse_thing_instances(label_mask)
-        else:
-            raise ValueError(
-                "Unsupported mask type '%s'. Supported values are "
-                "('stuff', 'thing')"
-            )
+        new_results = converter(
+            label_mask, label, label_type, offset, frame_size
+        )
+        results.extend(new_results)
 
-        for bbox, instance_mask in instances:
-            detections.append(
-                Detection(label=label, bounding_box=bbox, mask=instance_mask)
-            )
+    return results
 
-    return detections
+
+def _mask_to_detections(label_mask, label, label_type, offset, frame_size):
+    if label_type == "stuff":
+        instances = [_parse_stuff_instance(label_mask, offset, frame_size)]
+    elif label_type == "thing":
+        instances = _parse_thing_instances(label_mask, offset, frame_size)
+    else:
+        raise ValueError(
+            "Unsupported mask type '%s'. Supported values are "
+            "('stuff', 'thing')"
+        )
+
+    return list(
+        Detection(label=label, bounding_box=bbox, mask=instance_mask)
+        for bbox, instance_mask in instances
+    )
+
+
+def _mask_to_polylines(
+    label_mask, label, label_type, offset, frame_size, tolerance
+):
+    polygons = _get_polygons(
+        label_mask,
+        tolerance=tolerance,
+        offset=offset,
+        frame_size=frame_size,
+    )
+
+    if label_type == "stuff":
+        polygons = [polygons]
+    elif label_type == "thing":
+        polygons = [[p] for p in polygons]
+    else:
+        raise ValueError(
+            "Unsupported mask type '%s'. Supported values are "
+            "('stuff', 'thing')"
+        )
+
+    return list(
+        Polyline(label=label, points=points, filled=True, closed=True)
+        for points in polygons
+    )
+
+
+def _segmentation_to_detections(segmentation, mask_targets, mask_types):
+    return _convert_segmentation(
+        segmentation, mask_targets, mask_types, _mask_to_detections
+    )
 
 
 def _segmentation_to_polylines(
     segmentation, mask_targets, mask_types, tolerance
 ):
-    if isinstance(mask_types, dict):
-        default = None
-    else:
-        default = mask_types
-        mask_types = {}
-
-    mask = segmentation.get_mask()
-    is_rgb = mask.ndim == 3
-
-    if is_rgb:
-        array_targets = np.unique(mask.reshape(-1, mask.shape[2]), axis=0)
-        targets = [_rgb_to_hex(t) for t in targets]
-    else:
-        targets = np.unique(mask)
-        array_targets = itertools.repeat(None)
-
-    polylines = []
-    for target, array_target in zip(targets, array_targets):
-        if target in (0, "#000000"):
-            continue  # skip background
-
-        if mask_targets is not None:
-            label = mask_targets.get(target, None)
-            if is_rgb and label is None:
-                label = mask_targets.get(target.upper(), None)
-
-            if label is None:
-                continue  # skip unknown target
-        else:
-            label = str(target)
-
-        label_type = mask_types.get(target, None)
-        if is_rgb and label is None:
-            label_type = mask_types.get(target.upper(), None)
-
-        if label_type is None:
-            if default is None:
-                continue  # skip unknown type
-
-            label_type = default
-
-        if is_rgb:
-            label_mask = np.all(mask == array_target.reshape(1, 1, 3), axis=2)
-        else:
-            label_mask = mask == target
-
-        polygons = _get_polygons(label_mask, tolerance)
-
-        if label_type == "stuff":
-            polygons = [polygons]
-        elif label_type == "thing":
-            polygons = [[p] for p in polygons]
-        else:
-            raise ValueError(
-                "Unsupported mask type '%s'. Supported values are "
-                "('stuff', 'thing')"
-            )
-
-        for points in polygons:
-            polyline = Polyline(
-                label=label, points=points, filled=True, closed=True
-            )
-            polylines.append(polyline)
-
-    return polylines
+    converter = partial(_mask_to_polylines, tolerance=tolerance)
+    return _convert_segmentation(
+        segmentation, mask_targets, mask_types, converter
+    )
 
 
 def _hex_to_rgb(hex_str):
     r = int(hex_str[1:3], 16)
     g = int(hex_str[3:5], 16)
     b = int(hex_str[5:7], 16)
-    return np.array([r, g, b])
+    return (r, g, b)
 
 
 def _rgb_to_hex(rgb):
@@ -1767,57 +1959,99 @@ def _int_array_to_rgb(mask):
     return out_mask
 
 
-def _parse_stuff_instance(mask):
+def _parse_stuff_instance(mask, offset=None, frame_size=None):
     cols = np.any(mask, axis=0)
     rows = np.any(mask, axis=1)
     xmin, xmax = np.where(cols)[0][[0, -1]]
     ymin, ymax = np.where(rows)[0][[0, -1]]
 
-    height, width = mask.shape
-    x = xmin / width
-    y = ymin / height
+    if offset is None:
+        x_offset, y_offset = (0, 0)
+    else:
+        x_offset, y_offset = offset
+
+    if frame_size is None:
+        height, width = mask.shape
+    else:
+        width, height = frame_size
+
+    x = (xmin + x_offset) / width
+    y = (ymin + y_offset) / height
     w = (xmax - xmin + 1) / width
     h = (ymax - ymin + 1) / height
 
     bbox = [x, y, w, h]
-    instance_mask = mask[ymin:ymax, xmin:xmax]
+    instance_mask = mask[ymin : (ymax + 1), xmin : (xmax + 1)]
 
     return bbox, instance_mask
 
 
-def _parse_thing_instances(mask):
-    height, width = mask.shape
-    polygons = _get_polygons(mask, 2, abs_coords=True)
+def _parse_thing_instances(mask, offset=None, frame_size=None):
+    if offset is None:
+        x_offset, y_offset = (0, 0)
+    else:
+        x_offset, y_offset = offset
 
+    if frame_size is None:
+        height, width = mask.shape
+    else:
+        width, height = frame_size
+
+    labeled = skm.label(mask)
+    objects = _find_slices(labeled)
     instances = []
-    for p in polygons:
-        p = np.array(p)
-        xmin, ymin = np.floor(p.min(axis=0)).astype(int)
-        xmax, ymax = np.ceil(p.max(axis=0)).astype(int)
+    for target, slc in objects.items():
+        yslice, xslice = slc
+        xmin = xslice.start
+        ymin = yslice.start
+        instance_offset = (
+            offset[0] + xmin,
+            offset[1] + ymin,
+        )
 
-        xmax = max(xmin + 1, xmax)
-        ymax = max(ymin + 1, ymax)
-
-        x = xmin / width
-        y = ymin / height
-        w = (xmax - xmin) / width
-        h = (ymax - ymin) / height
-
-        bbox = [x, y, w, h]
-        instance_mask = mask[ymin:ymax, xmin:xmax]
-
-        instances.append((bbox, instance_mask))
+        # use the labeled image mask so `_parse_stuff_instance()`
+        # can be re-used here
+        instance_mask = labeled[slc] == target
+        instance = _parse_stuff_instance(
+            instance_mask, instance_offset, frame_size
+        )
+        instances.append(instance)
 
     return instances
 
 
-def _get_polygons(mask, tolerance, abs_coords=False):
-    polygons = etai._mask_to_polygons(mask, tolerance=tolerance)
-    if abs_coords:
-        return polygons
+def _get_polygons(
+    mask,
+    tolerance,
+    offset=None,
+    frame_size=None,
+    abs_coords=False,
+):
+    if offset is None:
+        x_offset, y_offset = (0, 0)
+    else:
+        x_offset, y_offset = offset
 
-    height, width = mask.shape
-    return [[(x / width, y / height) for x, y in p] for p in polygons]
+    if abs_coords:
+        height, width = (1, 1)
+    else:
+        if frame_size is None:
+            height, width = mask.shape
+        else:
+            width, height = frame_size
+
+    polygons = etai._mask_to_polygons(mask, tolerance=tolerance)
+    polygons = list(
+        list(
+            (
+                (x + x_offset) / width,
+                (y + y_offset) / height,
+            )
+            for x, y in p
+        )
+        for p in polygons
+    )
+    return polygons
 
 
 def _from_geo_json_single(d):
