@@ -1,21 +1,26 @@
 """
 Base classes for documents that back dataset contents.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
+
 from copy import deepcopy
+from datetime import datetime
 import json
 
+import bson
 from bson import json_util, ObjectId
 import mongoengine
 from pymongo import InsertOne, UpdateOne
 
 import eta.core.serial as etas
 
+import fiftyone.core.fields as fof
 import fiftyone.core.utils as fou
 
+from .database import ensure_connection
 from .utils import serialize_value, deserialize_value
 
 
@@ -297,29 +302,87 @@ class MongoEngineBaseDocument(SerializableDocument):
         return self._get_field_names(include_private=False)
 
     def has_field(self, field_name):
-        # pylint: disable=no-member
-        return field_name in self._fields_ordered
+        try:
+            _ = self._get_field(field_name)
+            return True
+        except (KeyError, AttributeError):
+            return False
 
     def get_field(self, field_name):
+        chunks = field_name.split(".", 1)
+        if len(chunks) > 1:
+            try:
+                value = getattr(self, chunks[0])
+                if value is not None:
+                    return value.get_field(chunks[1])
+            except AttributeError:
+                pass
+
+            if self.has_field(field_name):
+                return None
+
+            raise AttributeError(
+                "%s has no field '%s'" % (self.__class__.__name__, field_name)
+            )
+
         return getattr(self, field_name)
 
-    def set_field(
-        self,
-        field_name,
-        value,
-        create=True,
-        validate=True,
-        dynamic=False,
-    ):
-        if not create and not self.has_field(field_name):
+    def _get_field(self, field_name, allow_missing=False):
+        # pylint: disable=no-member
+        chunks = field_name.split(".", 1)
+        if len(chunks) > 1:
+            field = self._fields.get(chunks[0], None)
+            if hasattr(field, "get_field"):
+                field = field.get_field(chunks[1])
+        else:
+            field = self._fields.get(field_name, None)
+
+        if field is None and not allow_missing:
             raise AttributeError(
                 "%s has no field '%s'" % (self.__class__.__name__, field_name)
             )
 
+        return field
+
+    def set_field(self, field_name, value, create=True):
+        chunks = field_name.split(".", 1)
+
+        field = self._get_field(chunks[0], allow_missing=True)
+        if getattr(field, "read_only", False):
+            raise ValueError("Cannot edit read-only field '%s'" % field.path)
+
+        if len(chunks) > 1:
+            doc = self.get_field(chunks[0])
+
+            # handle syntax: sample["field.0.attr"] = value
+            if isinstance(doc, (mongoengine.base.BaseList, fof.ListField)):
+                chunks = chunks[1].split(".", 1)
+                doc = doc[int(chunks[0])]
+
+            return doc.set_field(chunks[1], value, create=create)
+
+        if not create and not self.has_field(field_name):
+            raise ValueError(
+                "%s has no field '%s'" % (self.__class__.__name__, field_name)
+            )
+
         setattr(self, field_name, value)
 
     def clear_field(self, field_name):
-        if not self.has_field(field_name):
+        chunks = field_name.split(".", 1)
+
+        field = self._get_field(chunks[0], allow_missing=True)
+        if getattr(field, "read_only", False):
+            raise ValueError("Cannot edit read-only field '%s'" % field.path)
+
+        if len(chunks) > 1:
+            value = self.get_field(chunks[0])
+            if value is not None:
+                return value.clear_field(chunks[1])
+
+            if self.has_field(field_name):
+                return
+
             raise AttributeError(
                 "%s has no field '%s'" % (self.__class__.__name__, field_name)
             )
@@ -333,13 +396,11 @@ class MongoEngineBaseDocument(SerializableDocument):
             )
 
     def field_to_mongo(self, field_name):
-        # pylint: disable=no-member
         value = self.get_field(field_name)
-        return self._fields[field_name].to_mongo(value)
+        return self._get_field(field_name).to_mongo(value)
 
     def field_to_python(self, field_name, value):
-        # pylint: disable=no-member
-        return self._fields[field_name].to_python(value)
+        return self._get_field(field_name).to_python(value)
 
     def _get_field_names(self, include_private=False, use_db_fields=False):
         field_names = self._fields_ordered
@@ -537,15 +598,46 @@ class Document(BaseDocument, mongoengine.Document):
 
     meta = {"abstract": True}
 
+    def __init__(self, **kwargs):
+        super().__init__(**kwargs)
+        self._changed_fields = []
+
     @classmethod
     def _doc_name(cls):
         return "Document"
 
+    def copy(self, new_id=False):
+        """Returns a deep copy of the document.
+
+        Args:
+            new_id (False): whether to generate a new ID for the copied
+                document. By default, the ID is left as ``None`` and will be
+                automatically populated when the document is added to the
+                database
+        """
+        doc_copy = super().copy()
+
+        if new_id:
+            # pylint: disable=no-member
+            id_field = self._meta.get("id_field", "id")
+            doc_copy.set_field(id_field, ObjectId())
+            doc_copy._created = True
+
+        return doc_copy
+
+    def reload(self, *fields, **kwargs):
+        """Reloads the document from the database.
+
+        Args:
+            *fields: an optional args list of specific fields to reload
+        """
+        ensure_connection()
+        super().reload(*fields, **kwargs)
+
     def save(
         self,
         upsert=False,
         validate=True,
-        clean=True,
         safe=False,
         **kwargs,
     ):
@@ -558,8 +650,6 @@ class Document(BaseDocument, mongoengine.Document):
             upsert (False): whether to insert the document if it has an ``id``
                 populated but no document with that ID exists in the database
             validate (True): whether to validate the document
-            clean (True): whether to call the document's ``clean()`` method.
-                Only applicable when ``validate`` is True
             safe (False): whether to ``reload()`` the document before raising
                 any errors
 
@@ -571,7 +661,6 @@ class Document(BaseDocument, mongoengine.Document):
                 deferred=False,
                 upsert=upsert,
                 validate=validate,
-                clean=clean,
                 **kwargs,
             )
         except:
@@ -582,12 +671,20 @@ class Document(BaseDocument, mongoengine.Document):
 
         return self
 
+    def copy_with_new_id(self):
+        _copy = self.copy()
+        _id = bson.ObjectId()
+        _copy.id = _id
+        _copy._created = True
+        return _copy
+
     def _save(
         self,
         deferred=False,
         upsert=False,
         validate=True,
         clean=True,
+        enforce_read_only=True,
         **kwargs,
     ):
         # pylint: disable=no-member
@@ -596,36 +693,48 @@ class Document(BaseDocument, mongoengine.Document):
                 "Cannot save an abstract document"
             )
 
-        if self._meta.get("auto_create_index", True):
-            self.ensure_indexes()
-
-        if validate:
-            self.validate(clean=clean)
+        ensure_connection()
 
         id_field = self._meta["id_field"]
-
-        created = self._created
-        if not created:
-            created = "_id" not in self.to_mongo(fields=[id_field])
-
-        doc = self.to_mongo()
-        _id = doc.get("_id", None)
+        _id = self.to_mongo(fields=[id_field]).get("_id", None)
+        created = self._created or _id is None
 
         ops = None
 
         if created:
-            # Save new document
+            # Insert new document
+            if validate:
+                self._validate(clean=clean)
+
+            doc = self.to_mongo()
+
             if _id is None:
                 _id = ObjectId()
                 doc["_id"] = _id
 
-            if deferred:
-                ops = [InsertOne(doc)]
-            else:
-                self._get_collection().insert_one(doc)
+            self[id_field] = self._fields[id_field].to_python(_id)
+            self._created = False
+
+            ops = self._insert(doc, deferred=deferred)
         else:
             # Update existing document
             updates = {}
+
+            if hasattr(self, "_changed_fields"):
+                changed_fields = self._get_changed_fields()
+                roots, paths = self._parse_changed_fields(changed_fields)
+            else:
+                # Changes aren't yet tracked, so validate everything
+                roots, paths = None, None
+
+            if validate:
+                self._validate_updates(
+                    roots,
+                    paths,
+                    clean=clean,
+                    enforce_read_only=enforce_read_only,
+                )
+
             sets, unsets = self._delta()
 
             if sets:
@@ -636,7 +745,11 @@ class Document(BaseDocument, mongoengine.Document):
 
             if updates:
                 ops, updated_existing = self._update(
-                    _id, updates, deferred=deferred, upsert=upsert, **kwargs
+                    _id,
+                    updates,
+                    deferred=deferred,
+                    upsert=upsert,
+                    **kwargs,
                 )
 
                 if not deferred and not upsert and not updated_existing:
@@ -645,17 +758,35 @@ class Document(BaseDocument, mongoengine.Document):
                         % (self._doc_name().lower(), str(_id))
                     )
 
-        # Make sure we store the PK on this document now that it's saved
-        if created or id_field not in self._meta.get("shard_key", []):
-            self[id_field] = self._fields[id_field].to_python(_id)
-
-        self._clear_changed_fields()
-        self._created = False
+            self._clear_changed_fields()
 
         return ops
 
-    def _update(self, _id, updates, deferred=False, upsert=False, **kwargs):
-        """Updates an existing document."""
+    def _insert(self, doc, deferred=False):
+        if deferred:
+            ops = [InsertOne(doc)]
+        else:
+            self._get_collection().insert_one(doc)
+            ops = None
+
+        return ops
+
+    def _update(
+        self,
+        _id,
+        updates,
+        deferred=False,
+        upsert=False,
+        virtual=False,
+        **kwargs,
+    ):
+        if self.has_field("last_modified_at") and not virtual:
+            now = datetime.utcnow()
+            self.last_modified_at = now
+            if "$set" not in updates:
+                updates["$set"] = {}
+            updates["$set"]["last_modified_at"] = now
+
         if deferred:
             ops = self._deferred_updates(_id, updates, upsert)
             updated_existing = None
@@ -665,6 +796,24 @@ class Document(BaseDocument, mongoengine.Document):
 
         return ops, updated_existing
 
+    def _update_last_loaded_at(self):
+        self.last_loaded_at = datetime.utcnow()
+        self.save(virtual=True)
+
+    def _update_last_modified_at(self, last_modified_at=None):
+        if last_modified_at is None:
+            last_modified_at = datetime.utcnow()
+
+        self.last_modified_at = last_modified_at
+        self.save(virtual=True)
+
+    def _update_last_deletion_at(self, last_deletion_at=None):
+        if last_deletion_at is None:
+            last_deletion_at = datetime.utcnow()
+
+        self.last_deletion_at = last_deletion_at
+        self.save(virtual=True)
+
     def _do_updates(self, _id, updates, upsert):
         updated_existing = True
 
@@ -683,6 +832,97 @@ class Document(BaseDocument, mongoengine.Document):
         op = UpdateOne({"_id": _id}, updates, upsert=upsert)
         return [op]
 
+    def _parse_changed_fields(self, db_paths):
+        roots = set()
+        paths = set()
+
+        for db_path in db_paths:
+            self._parse_db_path(db_path, roots, paths)
+
+        return roots, paths
+
+    def _parse_db_path(self, db_path, roots, paths):
+        chunks = db_path.split(".")
+
+        # pylint: disable=no-member
+        root = chunks[0]
+        root = self._reverse_db_field_map.get(root, root)
+
+        roots.add(root)
+        paths.add(root)
+
+        path = root
+        for chunk in chunks[1:]:
+            if chunk.isdigit():
+                continue
+            elif chunk == "_id":
+                chunk = "id"
+
+            path += "." + chunk
+            paths.add(path)
+
+    def _validate_updates(
+        self, roots, paths, clean=True, enforce_read_only=True
+    ):
+        if enforce_read_only and paths is not None:
+            for path in paths:
+                field = self._get_field(path, allow_missing=True)
+                if getattr(field, "read_only", False):
+                    raise ValueError("Cannot edit read-only field '%s'" % path)
+
+        self._validate(fields=roots, clean=clean)
+
+    def _validate(self, fields=None, clean=True):
+        errors = {}
+
+        if clean:
+            try:
+                self.clean()
+            except mongoengine.ValidationError as error:
+                errors["__all__"] = error
+
+        if fields is None:
+            fields = self._fields_ordered
+
+        for name in fields:
+            # pylint: disable=no-member
+            field = self._fields.get(name)
+            if field is None:
+                field = self._dynamic_fields.get(name)
+
+            value = self._data.get(name)
+
+            if value is not None:
+                try:
+                    if isinstance(
+                        field,
+                        (
+                            mongoengine.EmbeddedDocumentField,
+                            mongoengine.GenericEmbeddedDocumentField,
+                        ),
+                    ):
+                        field._validate(value, clean=clean)
+                    else:
+                        field._validate(value)
+                except mongoengine.ValidationError as error:
+                    errors[field.name] = error.errors or error
+                except (ValueError, AttributeError, AssertionError) as error:
+                    errors[field.name] = error
+            elif field.required and not getattr(field, "_auto_gen", False):
+                errors[field.name] = mongoengine.ValidationError(
+                    "Field is required", field_name=field.name
+                )
+
+        if errors:
+            # pylint: disable=no-member
+            pk = "None"
+            if hasattr(self, "pk"):
+                pk = self.pk
+            elif self._instance and hasattr(self._instance, "pk"):
+                pk = self._instance.pk
+            message = f"ValidationError ({self._class_name}:{pk}) "
+            raise mongoengine.ValidationError(message, errors=errors)
+
 
 def _merge_lists(dst, src, overwrite=False):
     dst.extend(v for v in src if v not in dst)
