@@ -1,13 +1,17 @@
 """
 Mixins and helpers for dataset backing documents.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
+
 from collections import OrderedDict
+from datetime import datetime
+import itertools
 
 from bson import ObjectId
+import mongoengine
 from pymongo import UpdateOne
 
 import fiftyone.core.fields as fof
@@ -16,6 +20,7 @@ import fiftyone.core.utils as fou
 
 from .database import get_db_conn
 from .dataset import SampleFieldDocument
+from .embedded_document import DynamicEmbeddedDocument
 from .utils import (
     deserialize_value,
     serialize_value,
@@ -113,18 +118,14 @@ class DatasetMixin(object):
         # pylint: disable=no-member
         return tuple(cls._fields[f].db_field or f for f in field_names)
 
-    def has_field(self, field_name):
-        # pylint: disable=no-member
-        return field_name in self._fields
-
     def get_field(self, field_name):
-        if not self.has_field(field_name):
+        try:
+            return super().get_field(field_name)
+        except AttributeError:
             raise AttributeError(
                 "%s has no field '%s'" % (self._doc_name(), field_name)
             )
 
-        return super().get_field(field_name)
-
     def set_field(
         self,
         field_name,
@@ -132,10 +133,44 @@ class DatasetMixin(object):
         create=True,
         validate=True,
         dynamic=False,
+        _enforce_read_only=True,
     ):
-        validate_field_name(field_name)
+        field = self._get_field(field_name, allow_missing=True)
+        if getattr(field, "read_only", False) and _enforce_read_only:
+            raise ValueError("Cannot edit read-only field '%s'" % field.path)
 
-        if not self.has_field(field_name):
+        chunks = field_name.split(".", 1)
+
+        if len(chunks) > 1:
+            root = chunks[0]
+            doc = self.get_field(root)
+
+            if doc is None and create:
+                doc = DynamicEmbeddedDocument()
+                self[root] = doc
+                self._changed_fields.remove(root)
+
+            # handle sytnax: sample["field.0.attr"] = value
+            if isinstance(doc, (mongoengine.base.BaseList, fof.ListField)):
+                chunks = chunks[1].split(".", 1)
+                doc = doc[int(chunks[0])]
+                field_name = root + "." + chunks[1]
+
+            doc.set_field(chunks[1], value, create=create)
+
+            #
+            # We treat attributes of `DynamicEmbeddedDocument` fields like
+            # top-level fields and automatically declare new attributes on the
+            # datasets's schema
+            #
+            # If the user is setting attributes of other embedded documents
+            # such as `Label` fields, then by convention we don't automatically
+            # declare new attributes on the dataset's schema
+            #
+            if type(doc) != DynamicEmbeddedDocument:
+                return
+
+        if field is None:
             if create:
                 self.add_implied_field(
                     field_name,
@@ -149,8 +184,8 @@ class DatasetMixin(object):
                     "%s has no field '%s'" % (self._doc_name(), field_name)
                 )
         elif value is not None:
-            if validate:
-                self._fields[field_name].validate(value)
+            if validate and field is not None:
+                field.validate(value)
 
             if dynamic:
                 self.add_implied_field(
@@ -161,6 +196,9 @@ class DatasetMixin(object):
                     dynamic=dynamic,
                 )
 
+        if len(chunks) > 1:
+            return
+
         super().__setattr__(field_name, value)
 
     def clear_field(self, field_name):
@@ -168,7 +206,17 @@ class DatasetMixin(object):
 
     @classmethod
     def get_field_schema(
-        cls, ftype=None, embedded_doc_type=None, include_private=False
+        cls,
+        ftype=None,
+        embedded_doc_type=None,
+        subfield=None,
+        read_only=None,
+        info_keys=None,
+        created_after=None,
+        include_private=False,
+        flat=False,
+        unwind=True,
+        mode=None,
     ):
         """Returns a schema dictionary describing the fields of this document.
 
@@ -183,28 +231,48 @@ class DatasetMixin(object):
                 iterable of types to which to restrict the returned schema.
                 Must be subclass(es) of
                 :class:`fiftyone.core.odm.BaseEmbeddedDocument`
+            subfield (None): an optional subfield type or iterable of subfield
+                types to which to restrict the returned schema. Must be
+                subclass(es) of :class:`fiftyone.core.fields.Field`
+            read_only (None): whether to restrict to (True) or exclude (False)
+                read-only fields. By default, all fields are included
+            info_keys (None): an optional key or list of keys that must be in
+                the field's ``info`` dict
+            created_after (None): an optional ``datetime`` specifying a minimum
+                creation date
             include_private (False): whether to include fields that start with
                 ``_`` in the returned schema
+            flat (False): whether to return a flattened schema where all
+                embedded document fields are included as top-level keys
+            unwind (True): whether to traverse into list fields. Only
+                applicable when ``flat=True``
+            mode (None): whether to apply the above constraints before and/or
+                after flattening the schema. Only applicable when ``flat=True``.
+                Supported values are ``("before", "after", "both")``. The
+                default is ``"after"``
 
         Returns:
-             a dictionary mapping field names to field types
+            a dict mapping field names to :class:`fiftyone.core.fields.Field`
+            instances
         """
-        fof.validate_type_constraints(
-            ftype=ftype, embedded_doc_type=embedded_doc_type
+        schema = OrderedDict(
+            (fn, cls._fields[fn])  # pylint: disable=no-member
+            for fn in cls._get_fields_ordered(include_private=include_private)
         )
 
-        schema = OrderedDict()
-        field_names = cls._get_fields_ordered(include_private=include_private)
-        for field_name in field_names:
-            # pylint: disable=no-member
-            field = cls._fields[field_name]
-
-            if fof.matches_type_constraints(
-                field, ftype=ftype, embedded_doc_type=embedded_doc_type
-            ):
-                schema[field_name] = field
-
-        return schema
+        return fof.filter_schema(
+            schema,
+            ftype=ftype,
+            embedded_doc_type=embedded_doc_type,
+            subfield=subfield,
+            read_only=read_only,
+            info_keys=info_keys,
+            created_after=created_after,
+            include_private=include_private,
+            flat=flat,
+            unwind=unwind,
+            mode=mode,
+        )
 
     @classmethod
     def merge_field_schema(
@@ -213,20 +281,22 @@ class DatasetMixin(object):
         expand_schema=True,
         recursive=True,
         validate=True,
+        overwrite=False,
     ):
         """Merges the field schema into this document.
 
         Args:
-            schema: a dictionary mapping field names or
-                ``embedded.field.names``to
+            schema: a dict mapping field names or ``embedded.field.names`` to
                 :class:`fiftyone.core.fields.Field` instances
             expand_schema (True): whether to add new fields to the schema
                 (True) or simply validate that fields already exist with
                 consistent types (False)
             recursive (True): whether to recursively merge embedded document
                 fields
-            validate (True): whether to validate the field against an existing
-                field at the same path
+            validate (True): whether to validate fields against existing fields
+                at the same path
+            overwrite (False): whether to overwrite the editable metadata of
+                existing fields
 
         Returns:
             True/False whether any new fields were added
@@ -236,17 +306,29 @@ class DatasetMixin(object):
                 existing field of the same name or a new field is found but
                 ``expand_schema == False``
         """
+        dataset = cls._dataset
+        dataset_doc = dataset._doc
+        media_type = dataset.media_type
+        is_frame_field = cls._is_frames_doc
+        now = datetime.utcnow()
+
         new_schema = {}
+        new_metadata = {}
 
         for path, field in schema.items():
-            new_fields = cls._merge_field(
+            _new_schema, _new_metadata = cls._merge_field(
                 path,
                 field,
                 validate=validate,
                 recursive=recursive,
+                overwrite=overwrite,
             )
-            if new_fields:
-                new_schema.update(new_fields)
+
+            if _new_schema:
+                new_schema.update(_new_schema)
+
+            if _new_metadata:
+                new_metadata.update(_new_metadata)
 
         if new_schema and not expand_schema:
             raise ValueError(
@@ -254,13 +336,46 @@ class DatasetMixin(object):
                 % (cls._doc_name(), list(new_schema.keys()))
             )
 
-        if not new_schema:
+        if not new_schema and not new_metadata:
             return False
 
-        for path, field in new_schema.items():
-            cls._add_field_schema(path, field)
+        # This fixes https://github.com/voxel51/fiftyone/issues/3185
+        # @todo improve list field updates in general so this isn't necessary
+        cls._reload_fields()
 
-        cls._dataset.save()
+        for path in new_schema.keys():
+            _, _, _, root_doc = cls._parse_path(path)
+            if root_doc is not None and getattr(root_doc, "read_only", False):
+                root = path.rsplit(".", 1)[0]
+                raise ValueError("Cannot edit read-only field '%s'" % root)
+
+            validate_field_name(
+                path,
+                media_type=media_type,
+                is_frame_field=is_frame_field,
+            )
+
+        # Silently skip updating metadata of any read-only fields
+        for path in list(new_metadata.keys()):
+            field = cls._get_field(path, allow_missing=True)
+            if field is not None and (
+                not isinstance(field, fof.Field)
+                or getattr(field, "read_only", False)
+            ):
+                del new_metadata[path]
+
+        for path, field in new_schema.items():
+            # Special syntax for declaring the subfield of a ListField
+            if path.endswith("[]"):
+                path = path[:-2]
+                field = fof.ListField(field=field)
+
+            cls._add_field_schema(path, field, created_at=now)
+
+        for path, d in new_metadata.items():
+            cls._update_field_metadata(path, d)
+
+        dataset_doc.save()
 
         return True
 
@@ -274,6 +389,7 @@ class DatasetMixin(object):
         fields=None,
         description=None,
         info=None,
+        read_only=False,
         expand_schema=True,
         recursive=True,
         validate=True,
@@ -299,6 +415,7 @@ class DatasetMixin(object):
                 :class:`fiftyone.core.fields.EmbeddedDocumentField`
             description (None): an optional description
             info (None): an optional info dict
+            read_only (False): whether the field should be read-only
             expand_schema (True): whether to add new fields to the schema
                 (True) or simply validate that the field already exists with a
                 consistent type (False)
@@ -323,6 +440,7 @@ class DatasetMixin(object):
             fields=fields,
             description=description,
             info=info,
+            read_only=read_only,
             **kwargs,
         )
 
@@ -386,6 +504,7 @@ class DatasetMixin(object):
         fields=None,
         description=None,
         info=None,
+        read_only=False,
         **kwargs,
     ):
         field_name = path.rsplit(".", 1)[-1]
@@ -397,6 +516,7 @@ class DatasetMixin(object):
             fields=fields,
             description=description,
             info=info,
+            read_only=read_only,
             **kwargs,
         )
 
@@ -440,6 +560,12 @@ class DatasetMixin(object):
                     % (cls._doc_name().lower(), path)
                 )
 
+            if field is not None and getattr(field, "read_only", False):
+                raise ValueError(
+                    "Cannot rename read-only %s field '%s'"
+                    % (cls._doc_name().lower(), path)
+                )
+
             if existing_field is not None:
                 raise ValueError(
                     "%s field '%s' already exists"
@@ -479,14 +605,26 @@ class DatasetMixin(object):
                 sample_collection, _paths, _new_paths
             )
 
+        # This fixes https://github.com/voxel51/fiftyone/issues/3185
+        # @todo improve list field updates in general so this isn't necessary
+        if schema_paths:
+            cls._reload_fields()
+
         for path, new_path in schema_paths:
             cls._rename_field_schema(path, new_path)
 
         if new_group_field:
             dataset_doc.group_field = new_group_field
 
+        if is_frame_field:
+            paths = [dataset._FRAMES_PREFIX + p for p in paths]
+            new_paths = [dataset._FRAMES_PREFIX + p for p in new_paths]
+
         dataset_doc.app_config._rename_paths(paths, new_paths)
-        dataset.save()
+        dataset_doc.save()
+
+        if schema_paths:
+            cls._rename_indexes(paths, new_paths)
 
     @classmethod
     def _clone_fields(cls, sample_collection, paths, new_paths):
@@ -504,6 +642,7 @@ class DatasetMixin(object):
         media_type = dataset.media_type
         is_frame_field = cls._is_frames_doc
         is_dataset = isinstance(sample_collection, fod.Dataset)
+        now = datetime.utcnow()
 
         simple_paths = []
         coll_paths = []
@@ -553,10 +692,15 @@ class DatasetMixin(object):
             _paths, _new_paths = zip(*coll_paths)
             cls._clone_fields_collection(sample_collection, _paths, _new_paths)
 
-        for path, new_path in schema_paths:
-            cls._clone_field_schema(path, new_path)
+        # This fixes https://github.com/voxel51/fiftyone/issues/3185
+        # @todo improve list field updates in general so this isn't necessary
+        if schema_paths:
+            cls._reload_fields()
 
-        dataset.save()
+        for path, new_path in schema_paths:
+            cls._clone_field_schema(path, new_path, created_at=now)
+
+        dataset_doc.save()
 
     @classmethod
     def _clear_fields(cls, sample_collection, paths):
@@ -569,6 +713,8 @@ class DatasetMixin(object):
         """
         is_dataset = isinstance(sample_collection, fod.Dataset)
 
+        paths = _remove_nested_paths(paths)
+
         simple_paths = []
         coll_paths = []
 
@@ -581,6 +727,12 @@ class DatasetMixin(object):
                     "%s field '%s' does not exist" % (cls._doc_name(), path)
                 )
 
+            if field is not None and getattr(field, "read_only", False):
+                raise ValueError(
+                    "Cannot rename read-only %s field '%s'"
+                    % (cls._doc_name().lower(), path)
+                )
+
             if is_dataset and is_root_field:
                 simple_paths.append(path)
             else:
@@ -609,6 +761,8 @@ class DatasetMixin(object):
         media_type = dataset.media_type
         is_frame_field = cls._is_frames_doc
 
+        paths = _remove_nested_paths(paths)
+
         del_paths = []
         del_schema_paths = []
 
@@ -642,6 +796,12 @@ class DatasetMixin(object):
                 )
                 continue
 
+            if field is not None and getattr(field, "read_only", False):
+                raise ValueError(
+                    "Cannot delete read-only %s field '%s'"
+                    % (cls._doc_name().lower(), path)
+                )
+
             if (
                 media_type == fom.GROUP
                 and not is_frame_field
@@ -667,15 +827,27 @@ class DatasetMixin(object):
             del_paths.append(path)
             del_schema_paths.append(path)
 
-        if del_paths:
-            cls._delete_fields_simple(del_paths)
+        if not del_paths:
+            return
+
+        cls._delete_fields_simple(del_paths)
+
+        # This fixes https://github.com/voxel51/fiftyone/issues/3185
+        # @todo improve list field updates in general so this isn't necessary
+        if del_schema_paths:
+            cls._reload_fields()
 
         for del_path in del_schema_paths:
             cls._delete_field_schema(del_path)
 
+        if is_frame_field:
+            del_paths = [dataset._FRAMES_PREFIX + p for p in del_paths]
+
+        dataset_doc.app_config._delete_paths(del_paths)
+        dataset_doc.save()
+
         if del_paths:
-            dataset_doc.app_config._delete_paths(del_paths)
-            dataset.save()
+            cls._delete_indexes(del_paths)
 
     @classmethod
     def _remove_dynamic_fields(cls, paths, error_level=0):
@@ -691,6 +863,11 @@ class DatasetMixin(object):
             -   1: log warning if a field cannot be removed
             -   2: ignore fields that cannot be removed
         """
+        dataset = cls._dataset
+        dataset_doc = dataset._doc
+
+        paths = _remove_nested_paths(paths)
+
         del_paths = []
 
         for path in paths:
@@ -728,15 +905,33 @@ class DatasetMixin(object):
                 )
                 continue
 
+            if field is not None and getattr(field, "read_only", False):
+                fou.handle_error(
+                    ValueError(
+                        "Cannot remove read-only %s field '%s'"
+                        % (cls._doc_name().lower(), path)
+                    ),
+                    error_level,
+                )
+                continue
+
             del_paths.append(path)
 
+        if not del_paths:
+            return
+
+        # This fixes https://github.com/voxel51/fiftyone/issues/3185
+        # @todo improve list field updates in general so this isn't necessary
+        cls._reload_fields()
+
         for del_path in del_paths:
             cls._delete_field_schema(del_path)
 
-        if del_paths:
-            dataset = cls._dataset
-            dataset._doc.app_config._delete_paths(del_paths)
-            dataset.save()
+        if cls._is_frames_doc:
+            del_paths = [dataset._FRAMES_PREFIX + p for p in del_paths]
+
+        dataset_doc.app_config._delete_paths(del_paths)
+        dataset_doc.save()
 
     @classmethod
     def _rename_fields_simple(cls, paths, new_paths):
@@ -746,9 +941,12 @@ class DatasetMixin(object):
         _paths, _new_paths = cls._handle_db_fields(paths, new_paths)
 
         rename_expr = dict(zip(_paths, _new_paths))
+        now = datetime.utcnow()
 
         coll = get_db_conn()[cls.__name__]
-        coll.update_many({}, {"$rename": rename_expr})
+        coll.update_many(
+            {}, {"$rename": rename_expr, "$set": {"last_modified_at": now}}
+        )
 
     @classmethod
     def _rename_fields_collection(cls, sample_collection, paths, new_paths):
@@ -781,7 +979,7 @@ class DatasetMixin(object):
 
             view = view.set_field(new_path, expr, _allow_missing=True)
 
-        view = view.mongo([{"$unset": _paths}])
+        view = view.mongo([{"$project": {p: False for p in _paths}}])
 
         #
         # Ideally only the embedded field would be saved, but the `$merge`
@@ -799,6 +997,7 @@ class DatasetMixin(object):
         _paths, _new_paths = cls._handle_db_fields(paths, new_paths)
 
         set_expr = {v: "$" + k for k, v in zip(_paths, _new_paths)}
+        set_expr["last_modified_at"] = datetime.utcnow()
 
         coll = get_db_conn()[cls.__name__]
         coll.update_many({}, [{"$set": set_expr}])
@@ -849,8 +1048,11 @@ class DatasetMixin(object):
 
         _paths = cls._handle_db_fields(paths)
 
+        set_expr = {p: None for p in _paths}
+        set_expr["last_modified_at"] = datetime.utcnow()
+
         coll = get_db_conn()[cls.__name__]
-        coll.update_many({}, {"$set": {p: None for p in _paths}})
+        coll.update_many({}, {"$set": set_expr})
 
     @classmethod
     def _clear_fields_collection(cls, sample_collection, paths):
@@ -882,9 +1084,12 @@ class DatasetMixin(object):
             return
 
         _paths = cls._handle_db_fields(paths)
+        now = datetime.utcnow()
 
         coll = get_db_conn()[cls.__name__]
-        coll.update_many({}, [{"$unset": _paths}])
+        coll.update_many(
+            {}, [{"$unset": _paths}, {"$set": {"last_modified_at": now}}]
+        )
 
     @classmethod
     def _handle_db_field(cls, path, new_path=None):
@@ -917,7 +1122,14 @@ class DatasetMixin(object):
         return tuple(cls._handle_db_field(p) for p in paths)
 
     @classmethod
-    def _merge_field(cls, path, field, validate=True, recursive=True):
+    def _merge_field(
+        cls,
+        path,
+        field,
+        validate=True,
+        recursive=True,
+        overwrite=False,
+    ):
         chunks = path.split(".")
         field_name = chunks[-1]
 
@@ -941,7 +1153,7 @@ class DatasetMixin(object):
 
             doc = schema[chunk]
 
-            if isinstance(doc, fof.ListField):
+            while isinstance(doc, fof.ListField):
                 doc = doc.field
 
             if not isinstance(doc, fof.EmbeddedDocumentField):
@@ -957,31 +1169,62 @@ class DatasetMixin(object):
                     )
                 )
 
-        if isinstance(field, fof.ObjectIdField) and field_name.startswith("_"):
-            field_name = field_name[1:]
-
-        if field_name == "id":
-            return
-
         if field_name in doc._fields:
             existing_field = doc._fields[field_name]
 
             if recursive:
-                if isinstance(existing_field, fof.ListField) and isinstance(
+                is_list_field = False
+
+                while isinstance(existing_field, fof.ListField) and isinstance(
                     field, fof.ListField
                 ):
                     existing_field = existing_field.field
                     field = field.field
+                    is_list_field = True
 
                 if isinstance(existing_field, fof.EmbeddedDocumentField):
                     return existing_field._merge_fields(
-                        path, field, validate=validate, recursive=recursive
+                        path,
+                        field,
+                        validate=validate,
+                        recursive=recursive,
+                        overwrite=overwrite,
                     )
 
+                # Special syntax for declaring the subfield of a ListField
+                if (
+                    is_list_field
+                    and field is not None
+                    and existing_field is None
+                    and path.endswith("[]")
+                ):
+                    return {path: field}, None
+
+            #
+            # In principle, merging an untyped list field into a typed list
+            # field --- eg ListField(StringField) --- should not be allowed, as
+            # the untyped list may contain incompatible values.
+            #
+            # However, we are intentionally skipping validation here so that
+            # the following will work::
+            #
+            #   doc.set_field("tags", [], validate=True)
+            #
+            if is_list_field and field is None:
+                return None, None
+
             if validate:
                 validate_fields_match(path, field, existing_field)
 
-            return
+            if overwrite:
+                # Overwrite existing field parameters
+                new_metadata = {path: fof.get_field_metadata(field)}
+                return None, new_metadata
+
+            return None, None
+
+        if field_name == "id":
+            return None, None
 
         dataset = cls._dataset
         media_type = dataset.media_type
@@ -1008,23 +1251,27 @@ class DatasetMixin(object):
                     "have one group field" % field_name
                 )
 
-        return {path: field}
+        return {path: field}, None
 
     @classmethod
-    def _add_field_schema(cls, path, field):
-        field_name, doc, field_docs = cls._parse_path(path)
+    def _add_field_schema(cls, path, field, created_at=None):
+        if created_at is None:
+            created_at = datetime.utcnow()
+
+        field_name, doc, field_docs, root_doc = cls._parse_path(path)
 
         field = field.copy()
         field.db_field = _get_db_field(field, field_name)
         field.name = field_name
+        field._set_created_at(created_at)
 
         doc._declare_field(cls._dataset, path, field)
-        _add_field_doc(field_docs, field)
+        _add_field_doc(field_docs, root_doc, field)
 
     @classmethod
     def _rename_field_schema(cls, path, new_path):
         same_root, new_field_name = _parse_paths(path, new_path)
-        field_name, doc, field_docs = cls._parse_path(path)
+        field_name, doc, field_docs, _ = cls._parse_path(path)
 
         field = doc._fields[field_name]
         new_db_field = _get_db_field(field, new_field_name)
@@ -1039,24 +1286,56 @@ class DatasetMixin(object):
             doc._undeclare_field(field_name)
             _delete_field_doc(field_docs, field_name)
 
-            _, new_doc, new_field_docs = cls._parse_path(new_path)
+            _, new_doc, new_field_docs, new_root_doc = cls._parse_path(
+                new_path
+            )
             new_doc._declare_field(cls._dataset, new_path, field)
-            _add_field_doc(new_field_docs, field)
+            _add_field_doc(new_field_docs, new_root_doc, field)
 
     @classmethod
-    def _clone_field_schema(cls, path, new_path):
-        field_name, doc, _ = cls._parse_path(path)
+    def _clone_field_schema(cls, path, new_path, created_at=None):
+        field_name, doc, _, _ = cls._parse_path(path)
         field = doc._fields[field_name]
 
-        cls._add_field_schema(new_path, field)
+        cls._add_field_schema(new_path, field, created_at=created_at)
 
     @classmethod
     def _delete_field_schema(cls, path):
-        field_name, doc, field_docs = cls._parse_path(path)
+        field_name, doc, field_docs, _ = cls._parse_path(path)
 
         doc._undeclare_field(field_name)
         _delete_field_doc(field_docs, field_name)
 
+    @classmethod
+    def _rename_indexes(cls, paths, new_paths):
+        updates = _get_index_updates(cls._dataset, paths, new_paths=new_paths)
+
+        for name, new_index_spec in updates.items():
+            cls._dataset.drop_index(name)
+            cls._dataset.create_index(new_index_spec)
+
+    @classmethod
+    def _delete_indexes(cls, paths):
+        updates = _get_index_updates(cls._dataset, paths)
+
+        for name in updates.keys():
+            cls._dataset.drop_index(name)
+
+    @classmethod
+    def _update_field_metadata(cls, path, d):
+        field_name, doc, _, _ = cls._parse_path(path)
+        field_doc = cls._get_field_doc(path)
+        field = doc._fields[field_name]
+
+        for key, value in d.items():
+            setattr(field_doc, key, value)
+            setattr(field, key, value)
+
+    @classmethod
+    def _reload_fields(cls):
+        dataset_doc = cls._dataset._doc
+        dataset_doc.reload(cls._fields_attr())
+
     @classmethod
     def _get_field(cls, path, allow_missing=False, check_default=False):
         chunks = path.split(".")
@@ -1066,7 +1345,7 @@ class DatasetMixin(object):
         try:
             for chunk in chunks[:-1]:
                 doc = doc._fields[chunk]
-                if isinstance(doc, fof.ListField):
+                while isinstance(doc, fof.ListField):
                     doc = doc.field
 
             field = doc._fields[field_name]
@@ -1086,7 +1365,12 @@ class DatasetMixin(object):
         return field, is_default
 
     @classmethod
-    def _get_field_doc(cls, path, allow_missing=False):
+    def _get_field_doc(cls, path, allow_missing=False, reload=False):
+        # This fixes https://github.com/voxel51/fiftyone/issues/3185
+        # @todo improve list field updates in general so this isn't necessary
+        if reload:
+            cls._reload_fields()
+
         chunks = path.split(".")
         field_docs = cls._dataset._doc[cls._fields_attr()]
 
@@ -1117,6 +1401,7 @@ class DatasetMixin(object):
         field_name = chunks[-1]
         doc = cls
         field_docs = cls._dataset._doc[cls._fields_attr()]
+        root_field_doc = None
 
         root = None
         for chunk in chunks[:-1]:
@@ -1128,13 +1413,14 @@ class DatasetMixin(object):
             found = False
             for field_doc in field_docs:
                 if field_doc.name == chunk:
+                    root_field_doc = field_doc
                     field_docs = field_doc.fields
                     found = True
                     break
 
             if not found:
                 if allow_missing:
-                    return None, None, None
+                    return None, None, None, None
 
                 raise ValueError(
                     "Invalid %s field '%s'; field '%s' does not exist"
@@ -1143,12 +1429,12 @@ class DatasetMixin(object):
 
             doc = doc._fields[chunk]
 
-            if isinstance(doc, fof.ListField):
+            while isinstance(doc, fof.ListField):
                 doc = doc.field
 
             if not isinstance(doc, fof.EmbeddedDocumentField):
                 if allow_missing:
-                    return None, None, None
+                    return None, None, None, None
 
                 raise ValueError(
                     "Invalid %s field '%s'; field '%s' is a %s, not an %s"
@@ -1161,7 +1447,7 @@ class DatasetMixin(object):
                     )
                 )
 
-        return field_name, doc, field_docs
+        return field_name, doc, field_docs, root_field_doc
 
     @classmethod
     def _declare_field(cls, dataset, path, field_or_doc):
@@ -1226,16 +1512,31 @@ class DatasetMixin(object):
 
         delattr(cls, field_name)
 
+    def _insert(self, doc, deferred=False):
+        now = datetime.utcnow()
+        self.created_at = now
+        self.last_modified_at = now
+        doc["created_at"] = now
+        doc["last_modified_at"] = now
+        return super()._insert(doc, deferred=deferred)
+
     def _update(
         self,
         _id,
         updates,
         deferred=False,
         upsert=False,
+        virtual=False,
         filtered_fields=None,
         **kwargs,
     ):
-        """Updates an existing document."""
+        if not virtual:
+            now = datetime.utcnow()
+            self.last_modified_at = now
+            if "$set" not in updates:
+                updates["$set"] = {}
+            updates["$set"]["last_modified_at"] = now
+
         extra_updates = self._extract_extra_updates(updates, filtered_fields)
 
         if deferred:
@@ -1462,15 +1763,29 @@ class NoDatasetMixin(object):
 
     def has_field(self, field_name):
         try:
-            return field_name in self._data
+            data = self._data
         except AttributeError:
-            # If `_data` is not initialized
+            # `_data` is not initialized
             return False
 
+        chunks = field_name.split(".", 1)
+        if len(chunks) > 1:
+            value = data.get(chunks[0], None)
+            try:
+                return value.has_field(chunks[1])
+            except AttributeError:
+                return False
+
+        return field_name in data
+
     def get_field(self, field_name):
+        chunks = field_name.split(".", 1)
         try:
-            return self._data[field_name]
-        except KeyError:
+            if len(chunks) > 1:
+                return self._data[chunks[0]].get_field(chunks[1])
+            else:
+                return self._data[field_name]
+        except (KeyError, AttributeError):
             raise AttributeError(
                 "%s has no field '%s'" % (self._doc_name(), field_name)
             )
@@ -1483,6 +1798,11 @@ class NoDatasetMixin(object):
         validate=True,
         dynamic=False,
     ):
+        chunks = field_name.split(".", 1)
+        if len(chunks) > 1:
+            doc = self.get_field(chunks[0])
+            return doc.set_field(chunks[1], value, create=create)
+
         if not create and not self.has_field(field_name):
             raise ValueError(
                 "%s has no field '%s'" % (self._doc_name(), field_name)
@@ -1492,18 +1812,31 @@ class NoDatasetMixin(object):
         self._data[field_name] = value
 
     def clear_field(self, field_name):
+        chunks = field_name.split(".", 1)
+        if len(chunks) > 1:
+            value = self.get_field(chunks[0])
+            if value is not None:
+                return value.clear_field(chunks[1])
+
+            if self.has_field(field_name):
+                return
+
+            raise AttributeError(
+                "%s has no field '%s'" % (self._doc_name(), field_name)
+            )
+
         if field_name in self.default_fields:
             default_value = self._get_default(self.default_fields[field_name])
             self.set_field(field_name, default_value, create=False)
             return
 
-        if not self.has_field(field_name):
-            raise ValueError(
+        try:
+            self._data.pop(field_name)
+        except KeyError:
+            raise AttributeError(
                 "%s has no field '%s'" % (self._doc_name(), field_name)
             )
 
-        self._data.pop(field_name)
-
     def to_dict(self, extended=False):
         d = {}
         for k, v in self._data.items():
@@ -1576,8 +1909,15 @@ def _split_path(path):
     return chunks[0], chunks[1]
 
 
-def _add_field_doc(field_docs, field_or_doc):
-    # This is intentionally implemented to modify ``field_docs`` in place
+def _remove_nested_paths(paths):
+    return [
+        path
+        for path in paths
+        if not any(path.startswith(p + ".") for p in paths)
+    ]
+
+
+def _add_field_doc(field_docs, root_doc, field_or_doc):
     if isinstance(field_or_doc, fof.Field):
         new_field_doc = SampleFieldDocument.from_field(field_or_doc)
     else:
@@ -1588,11 +1928,10 @@ def _add_field_doc(field_docs, field_or_doc):
             field_docs[i] = new_field_doc
             return
 
-    try:
+    if root_doc is not None:
+        root_doc.fields = root_doc.fields + [new_field_doc]
+    else:
         field_docs.append(new_field_doc)
-    except ReferenceError:
-        # mongoengine seems to lose references to things... it's okay
-        pass
 
 
 def _update_field_doc(field_docs, field_name, field):
@@ -1603,8 +1942,58 @@ def _update_field_doc(field_docs, field_name, field):
 
 
 def _delete_field_doc(field_docs, field_name):
-    # This is intentionally implemented to modify ``field_docs`` in place
     for i, field_doc in enumerate(field_docs):
         if field_doc.name == field_name:
             del field_docs[i]
             break
+
+
+def _get_index_updates(dataset, paths, new_paths=None):
+    if not paths:
+        return {}
+
+    update = new_paths is not None
+    if new_paths is None:
+        new_paths = itertools.repeat("")
+
+    has_frame_fields = dataset._has_frame_fields()
+    index_info = dataset.get_index_information()
+    fields_map = dataset._get_db_fields_map(reverse=True)
+    if has_frame_fields:
+        prefix = dataset._FRAMES_PREFIX
+        frame_fields_map = dataset._get_db_fields_map(
+            frames=True, reverse=True
+        )
+
+    updates = {}
+
+    for name, info in index_info.items():
+        is_frame_index = has_frame_fields and name.startswith(prefix)
+
+        modified = False
+        new_index_spec = []
+        for _path, arg in info["key"]:
+            if is_frame_index:
+                _path = prefix + frame_fields_map.get(_path, _path)
+            else:
+                _path = fields_map.get(_path, _path)
+
+            key = (_path, arg)
+
+            for path, new_path in zip(paths, new_paths):
+                if _path == path:
+                    key = (new_path, arg)
+                    modified = True
+                elif _path.startswith(path + "."):
+                    key = (new_path + _path[len(path) :], arg)
+                    modified = True
+
+            new_index_spec.append(key)
+
+        if modified:
+            if update:
+                updates[name] = new_index_spec
+            else:
+                updates[name] = None
+
+    return updates
