@@ -1,7 +1,7 @@
 """
 Utilities for documents.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -9,13 +9,14 @@ from datetime import date, datetime
 import inspect
 import json
 import numbers
-import six
 import sys
 
 from bson import Binary, json_util, ObjectId, SON
 import numpy as np
 import pytz
 
+import eta.core.utils as etau
+
 import fiftyone as fo
 import fiftyone.core.fields as fof
 import fiftyone.core.media as fom
@@ -110,10 +111,13 @@ def deserialize_value(value):
 
         return value
 
-    if isinstance(value, six.binary_type):
+    if isinstance(value, bytes):
         # Serialized array in non-extended format
         return fou.deserialize_numpy_array(value)
 
+    if isinstance(value, list):
+        return [deserialize_value(v) for v in value]
+
     return value
 
 
@@ -183,6 +187,8 @@ def create_field(
     db_field=None,
     description=None,
     info=None,
+    read_only=False,
+    created_at=None,
     **kwargs,
 ):
     """Creates the field defined by the given specification.
@@ -214,6 +220,8 @@ def create_field(
             ``name`` is used
         description (None): an optional description
         info (None): an optional info dict
+        read_only (False): whether the field should be read-only
+        created_at (None): the datetime the field was created
 
     Returns:
         a :class:`fiftyone.core.fields.Field`
@@ -224,9 +232,13 @@ def create_field(
         else:
             db_field = name
 
-    # All user-defined fields are nullable
     field_kwargs = dict(
-        null=True, db_field=db_field, description=description, info=info
+        null=True,  # all user-defined fields are nullable
+        db_field=db_field,
+        description=description,
+        info=info,
+        read_only=read_only,
+        created_at=created_at,
     )
     field_kwargs.update(kwargs)
 
@@ -295,11 +307,15 @@ def get_field_kwargs(field):
     """Constructs the field keyword arguments dictionary for the given field.
 
     Args:
-        field: a :class:`fiftyone.core.fields.Field`
+        field: a :class:`fiftyone.core.fields.Field` or ``str(field)``
+            representation of one
 
     Returns:
         a field specification dict
     """
+    if etau.is_str(field):
+        return _parse_field_str(field)
+
     fields = []
 
     kwargs = {
@@ -308,6 +324,8 @@ def get_field_kwargs(field):
         "db_field": field.db_field,
         "description": field.description,
         "info": field.info,
+        "read_only": field.read_only,
+        "created_at": field.created_at,
     }
 
     if isinstance(field, (fof.ListField, fof.DictField)):
@@ -326,6 +344,24 @@ def get_field_kwargs(field):
     return kwargs
 
 
+def _parse_field_str(field_str):
+    # eg: "fiftyone.core.fields.EmbeddedDocumentField(fiftyone.core.labels.Detections)"
+    chunks = field_str.strip().split("(", 1)
+    ftype = etau.get_class(chunks[0])
+    kwargs = {"ftype": ftype}
+
+    if len(chunks) > 1:
+        param = etau.get_class(chunks[1][:-1])  # remove trailing ")"
+        if issubclass(ftype, fof.EmbeddedDocumentField):
+            kwargs["embedded_doc_type"] = param
+        elif issubclass(ftype, (fof.ListField, fof.DictField)):
+            kwargs["subfield"] = param
+        else:
+            raise ValueError("Failed to parse field string '%s'" % field_str)
+
+    return kwargs
+
+
 def get_implied_field_kwargs(value, dynamic=False):
     """Infers the field keyword arguments dictionary for a field that can hold
     the given value.
@@ -353,7 +389,7 @@ def get_implied_field_kwargs(value, dynamic=False):
     if isinstance(value, numbers.Number):
         return {"ftype": fof.FloatField}
 
-    if isinstance(value, six.string_types):
+    if isinstance(value, str):
         return {"ftype": fof.StringField}
 
     if isinstance(value, ObjectId):
@@ -406,6 +442,8 @@ def _get_field_kwargs(value, field, dynamic):
         "db_field": field.db_field,
         "description": field.description,
         "info": field.info,
+        "read_only": field.read_only,
+        "created_at": field.created_at,
     }
 
     if isinstance(field, (fof.ListField, fof.DictField)):
@@ -459,32 +497,37 @@ def _parse_embedded_doc_list_fields(values, dynamic):
         return []
 
     fields_dict = {}
-    fields = [_parse_embedded_doc_fields(v, dynamic) for v in values]
-    _merge_embedded_doc_fields(fields_dict, fields)
+    for value in values:
+        fields = _parse_embedded_doc_fields(value, dynamic)
+        _merge_embedded_doc_fields(fields_dict, fields)
+
     return _finalize_embedded_doc_fields(fields_dict)
 
 
-def _merge_embedded_doc_fields(fields_dict, fields_list):
-    for fields in fields_list:
-        for field in fields:
-            ftype = field["ftype"]
-            name = field["name"]
+def _merge_embedded_doc_fields(fields_dict, fields):
+    for field in fields:
+        name = field["name"]
+        ftype = field["ftype"]
+        subfield = field.get("subfield", None)
 
-            if name not in fields_dict:
-                fields_dict[name] = field
-                if ftype == fof.EmbeddedDocumentField:
-                    field["fields"] = {f["name"]: f for f in field["fields"]}
-            else:
-                efield = fields_dict[name]
-                etype = efield["ftype"]
-                if etype != ftype:
-                    # @todo could provide an `add_mixed=True` option to declare
-                    # mixed fields like this as a generic `Field`
-                    fields_dict[name] = None
-                elif ftype == fof.EmbeddedDocumentField:
-                    _merge_embedded_doc_fields(
-                        efield["fields"], field["fields"]
-                    )
+        if name not in fields_dict:
+            fields_dict[name] = field
+            if ftype == fof.EmbeddedDocumentField:
+                field["fields"] = {f["name"]: f for f in field["fields"]}
+        else:
+            efield = fields_dict[name]
+            etype = efield["ftype"]
+            if etype != ftype:
+                # @todo could provide an `add_mixed=True` option to declare
+                # mixed fields like this as a generic `Field`
+                fields_dict[name] = None
+            elif etype in (fof.ListField, fof.DictField):
+                if "subfield" in efield and efield["subfield"] != subfield:
+                    efield["subfield"] = None
+                elif subfield is not None:
+                    efield["subfield"] = subfield
+            elif ftype == fof.EmbeddedDocumentField:
+                _merge_embedded_doc_fields(efield["fields"], field["fields"])
 
 
 def _finalize_embedded_doc_fields(fields_dict):
@@ -565,7 +608,7 @@ def _get_field_type(value):
     if isinstance(value, numbers.Number):
         return fof.FloatField
 
-    if isinstance(value, six.string_types):
+    if isinstance(value, str):
         return fof.StringField
 
     if isinstance(value, ObjectId):
@@ -618,7 +661,8 @@ class DocumentRegistry(object):
             pass
 
         # Then full module list
-        for module in sys.modules.values():
+        all_modules = sys.modules.copy().values()
+        for module in all_modules:
             try:
                 cls = self._get_cls(module, name)
                 self._cache[name] = cls
@@ -652,3 +696,37 @@ class DocumentRegistryError(Exception):
 
 
 _document_registry = DocumentRegistry()
+
+
+def load_dataset(id=None, name=None):
+    """Loads the dataset from the database by its unique id or name. Throws
+    an error if neither id nor name is provided.
+
+    Args:
+        id (None): the unique id of the dataset
+        name (None): the name of the dataset
+
+    Returns:
+        a :class:`fiftyone.core.dataset.Dataset`
+    """
+    import fiftyone.core.odm as foo
+    import fiftyone.core.dataset as fod
+
+    if name:
+        return fod.load_dataset(name)
+
+    if not id:
+        raise ValueError("Must provide either id or name")
+
+    db = foo.get_db_conn()
+    try:
+        uid = ObjectId(id)
+    except:
+        # Although _id is an ObjectId by default, it's possible to set it to
+        # something else
+        uid = id
+
+    res = db.datasets.find_one({"_id": uid}, {"name": True})
+    if not res:
+        raise ValueError(f"Dataset with _id={uid} does not exist")
+    return fod.load_dataset(res.get("name"))
