@@ -1,7 +1,7 @@
 """
 Matplotlib plots.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -14,7 +14,6 @@ from matplotlib.widgets import Button, LassoSelector
 from matplotlib.path import Path
 import matplotlib.pyplot as plt
 from mpl_toolkits.axes_grid1 import make_axes_locatable
-from mpl_toolkits.axes_grid1.inset_locator import InsetPosition
 from mpl_toolkits.mplot3d import Axes3D  # pylint: disable=unused-import
 import sklearn.metrics.pairwise as skp
 import sklearn.metrics as skm
@@ -37,7 +36,7 @@ from .utils import (
 logger = logging.getLogger(__name__)
 
 
-_DEFAULT_STYLE = "seaborn-ticks"
+_DEFAULT_STYLE = "default"
 _DEFAULT_LINE_COLOR = "#FF6D04"
 _DEFAULT_CONTINUOUS_COLORSCALE = "viridis"
 
@@ -381,8 +380,8 @@ def plot_roc_curve(
     """Plots a receiver operating characteristic (ROC) curve.
 
     Args:
-        fpr: an array-like of false postive rates
-        tpr: an array-like of true postive rates
+        fpr: an array-like of false positive rates
+        tpr: an array-like of true positive rates
         roc_auc (None): the area under the ROC curve
         title (None): a title for the plot
         ax (None): a matplotlib axis to plot in
@@ -452,11 +451,11 @@ def lines(
                 of array-likes of values for multiple line traces
             -   the name of a sample field or ``embedded.field.name`` of
                 ``samples`` from which to extract values for a single line
-            -   the name of a frame field or ``frames.embbeded.field.name`` of
+            -   the name of a frame field or ``frames.embedded.field.name`` of
                 ``samples`` from which to extract values for per-sample line
                 traces
             -   a :class:`fiftyone.core.expressions.ViewExpression` that
-                resolves to a list (one line plot) or list of lists (muliple
+                resolves to a list (one line plot) or list of lists (multiple
                 line plots) of numeric values to compute from ``samples`` via
                 :meth:`fiftyone.core.collections.SampleCollection.values`
         y (None): the y data to plot. Can be any of the following:
@@ -466,11 +465,11 @@ def lines(
                 of array-likes of values for multiple line traces
             -   the name of a sample field or ``embedded.field.name`` of
                 ``samples`` from which to extract values for a single line
-            -   the name of a frame field or ``frames.embbeded.field.name`` of
+            -   the name of a frame field or ``frames.embedded.field.name`` of
                 ``samples`` from which to extract values for per-sample line
                 traces
             -   a :class:`fiftyone.core.expressions.ViewExpression` that
-                resolves to a list (one line plot) or list of lists (muliple
+                resolves to a list (one line plot) or list of lists (multiple
                 line plots) of numeric values to compute from ``samples`` via
                 :meth:`fiftyone.core.collections.SampleCollection.values`
         samples (None): the :class:`fiftyone.core.collections.SampleCollection`
@@ -1166,8 +1165,7 @@ class InteractiveCollection(InteractiveMatplotlibPlot):
 
         self._buttons = []
         for i, (label, icon_img, callback) in enumerate(button_defs):
-            bax = self.ax.figure.add_axes([0, 0, 1, 1], label=label)
-            bax.set_axes_locator(InsetPosition(self.ax, _button_pos(i)))
+            bax = self.ax.inset_axes(_button_pos(i), label=label)
             button = Button(
                 bax, "", color=color, hovercolor=hovercolor, image=icon_img
             )
