@@ -1,7 +1,7 @@
 """
 Plotly plots.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -103,6 +103,8 @@ def plot_confusion_matrix(
             labels,
             colorscale=colorscale,
             title=title,
+            gt_field=gt_field,
+            pred_field=pred_field,
             **kwargs,
         )
 
@@ -121,23 +123,31 @@ def plot_confusion_matrix(
 
 
 def _plot_confusion_matrix_static(
-    confusion_matrix, labels, colorscale=None, title=None, **kwargs
+    confusion_matrix,
+    labels,
+    colorscale=None,
+    title=None,
+    gt_field=None,
+    pred_field=None,
+    **kwargs,
 ):
     confusion_matrix = np.asarray(confusion_matrix)
     num_rows, num_cols = confusion_matrix.shape
     zlim = [0, confusion_matrix.max()]
+    truth = gt_field or "truth"
+    predicted = pred_field or "predicted"
 
     hover_lines = [
         "<b>count: %{z:d}</b>",
-        "truth: %{y}",
-        "predicted: %{x}",
+        f"{truth}: %{{y}}",
+        f"{predicted}: %{{x}}",
     ]
     hovertemplate = "<br>".join(hover_lines) + "<extra></extra>"
 
     xlabels = labels[:num_cols]
     ylabels = labels[:num_rows]
 
-    # Flip data so plot will have the standard descending diagnoal
+    # Flip data so plot will have the standard descending diagonal
     # Flipping the yaxis via `autorange="reversed"` isn't an option because
     # screenshots don't seem to respect that setting...
     confusion_matrix = np.flip(confusion_matrix, axis=0)
@@ -164,8 +174,6 @@ def _plot_confusion_matrix_static(
             scaleanchor="x",
             scaleratio=1,
         ),
-        xaxis_title="Predicted label",
-        yaxis_title="True label",
         title=title,
     )
 
@@ -214,7 +222,7 @@ def _plot_confusion_matrix_interactive(
     xlabels = labels[:num_cols]
     ylabels = labels[:num_rows]
 
-    # Flip data so plot will have the standard descending diagnoal
+    # Flip data so plot will have the standard descending diagonal
     # Flipping the yaxis via `autorange="reversed"` isn't an option because
     # screenshots don't seem to respect that setting...
     confusion_matrix = np.flip(confusion_matrix, axis=0)
@@ -234,6 +242,8 @@ def _plot_confusion_matrix_interactive(
         xlabels=xlabels,
         ylabels=ylabels,
         zlim=zlim,
+        gt_field=gt_field,
+        pred_field=pred_field,
         colorscale=colorscale,
         link_type="labels",
         init_view=samples,
@@ -326,7 +336,7 @@ def plot_regressions(
             By default, if ``sizes`` is a field name, this name will be used,
             otherwise the tooltip will use "size"
         show_colorbar_title (None): whether to show the colorbar title. By
-            default, a title will be shown only if a value was pasesd to
+            default, a title will be shown only if a value was passed to
             ``labels_title`` or an appropriate default can be inferred from
             the ``labels`` parameter
         **kwargs: optional keyword arguments for
@@ -618,8 +628,8 @@ def plot_roc_curve(
     """Plots a receiver operating characteristic (ROC) curve.
 
     Args:
-        fpr: an array-like of false postive rates
-        tpr: an array-like of true postive rates
+        fpr: an array-like of false positive rates
+        tpr: an array-like of true positive rates
         thresholds (None): an array-like of decision thresholds
         roc_auc (None): the area under the ROC curve
         style ("area"): a plot style to use. Supported values are
@@ -729,11 +739,11 @@ def lines(
                 of array-likes of values for multiple line traces
             -   the name of a sample field or ``embedded.field.name`` of
                 ``samples`` from which to extract values for a single line
-            -   the name of a frame field or ``frames.embbeded.field.name`` of
+            -   the name of a frame field or ``frames.embedded.field.name`` of
                 ``samples`` from which to extract values for per-sample line
                 traces
             -   a :class:`fiftyone.core.expressions.ViewExpression` that
-                resolves to a list (one line plot) or list of lists (muliple
+                resolves to a list (one line plot) or list of lists (multiple
                 line plots) of numeric values to compute from ``samples`` via
                 :meth:`fiftyone.core.collections.SampleCollection.values`
         y (None): the y data to plot. Can be any of the following:
@@ -743,11 +753,11 @@ def lines(
                 of array-likes of values for multiple line traces
             -   the name of a sample field or ``embedded.field.name`` of
                 ``samples`` from which to extract values for a single line
-            -   the name of a frame field or ``frames.embbeded.field.name`` of
+            -   the name of a frame field or ``frames.embedded.field.name`` of
                 ``samples`` from which to extract values for per-sample line
                 traces
             -   a :class:`fiftyone.core.expressions.ViewExpression` that
-                resolves to a list (one line plot) or list of lists (muliple
+                resolves to a list (one line plot) or list of lists (multiple
                 line plots) of numeric values to compute from ``samples`` via
                 :meth:`fiftyone.core.collections.SampleCollection.values`
         samples (None): the :class:`fiftyone.core.collections.SampleCollection`
@@ -1027,7 +1037,7 @@ def scatterplot(
         edges_title (None): a title string to use for ``edges`` in the legend.
             If none is provided, edges are not included in the legend
         show_colorbar_title (None): whether to show the colorbar title. By
-            default, a title will be shown only if a value was pasesd to
+            default, a title will be shown only if a value was passed to
             ``labels_title`` or an appropriate default can be inferred from
             the ``labels`` parameter
         axis_equal (False): whether to set the axes to equal scale
@@ -1211,6 +1221,7 @@ def location_scatterplot(
     sizes_title=None,
     edges_title=None,
     show_colorbar_title=None,
+    map_type="roadmap",
     **kwargs,
 ):
     """Generates an interactive scatterplot of the given location coordinates
@@ -1301,9 +1312,11 @@ def location_scatterplot(
         edges_title (None): a title string to use for ``edges`` in the legend.
             If none is provided, edges are not included in the legend
         show_colorbar_title (None): whether to show the colorbar title. By
-            default, a title will be shown only if a value was pasesd to
+            default, a title will be shown only if a value was passed to
             ``labels_title`` or an appropriate default can be inferred from
             the ``labels`` parameter
+        map_type ("satellite"): the map type to render. Supported values are
+            ``("roadmap", "satellite")``
         **kwargs: optional keyword arguments for
             :meth:`plotly:plotly.graph_objects.Figure.update_layout`
 
@@ -1414,6 +1427,8 @@ def location_scatterplot(
             colorbar_title,
         )
 
+    _set_map_type_to_figure(figure, map_type)
+
     figure.update_layout(**_DEFAULT_LAYOUT)
     figure.update_layout(title=title, **kwargs)
 
@@ -1515,6 +1530,7 @@ def _get_continuous_color(colorscale, value):
     # Returns a string like `rgb(float, float, float)`
 
     hex_to_rgb = lambda c: "rgb" + str(ImageColor.getcolor(c, "RGB"))
+    colorscale = sorted(colorscale, key=lambda x: x[0])
 
     if value <= 0 or len(colorscale) == 1:
         c = colorscale[0][1]
@@ -1564,11 +1580,11 @@ class PlotlyWidgetMixin(object):
             msg = (
                 "Interactive plots are currently only supported in Jupyter "
                 "notebooks. Support outside of notebooks and in Google Colab "
-                "will be included in an upcoming release. In the meantime, "
-                "you can still use this plot, but note that (i) selecting "
-                "data will not trigger callbacks, and (ii) you must manually "
-                "call `plot.show()` to launch a new plot that reflects the "
-                "current state of an attached session.\n\n"
+                "and Databricks will be included in an upcoming release. In "
+                "the meantime, you can still use this plot, but note that (i) "
+                "selecting data will not trigger callbacks, and (ii) you must "
+                "manually call `plot.show()` to launch a new plot that "
+                "reflects the current state of an attached session.\n\n"
                 "See https://docs.voxel51.com/user_guide/plots.html#working-in-notebooks"
                 " for more information."
             )
@@ -1668,12 +1684,12 @@ def _check_plotly_jupyter_environment():
     #
     # Requirements source: https://plotly.com/python/getting-started
     #
-    # There is also a `notebook>=5.3` requirement in Jupyter notebooks, but
+    # There is also a `notebook>=6` requirement in Jupyter notebooks, but
     # we do not explicitly check that here because the requirement for
     # JupyterLab is different and I don't know how to distinguish Jupyter
     # notebooks from JupyterLab right now...
     #
-    fou.ensure_package("ipywidgets>=7.5,<8")
+    fou.ensure_package("ipywidgets>=8,<9")
 
 
 class PlotlyNotebookPlot(PlotlyWidgetMixin, Plot):
@@ -1953,6 +1969,10 @@ class InteractiveHeatmap(PlotlyInteractivePlot):
         zlim (None): a ``[zmin, zmax]`` limit to use for the colorbar
         values_title ("count"): the semantic meaning of the heatmap values.
             Used for tooltips
+        gt_field (None): the name of the ground truth field, if known. Used for
+            tooltips
+        pred_field (None): the name of the predictions field, if known. Used
+            for tooltips
         colorscale (None): a plotly colorscale to use
         grid_opacity (0.1): an opacity value for the grid points
         bg_opacity (0.25): an opacity value for background (unselected) cells
@@ -1968,6 +1988,8 @@ class InteractiveHeatmap(PlotlyInteractivePlot):
         ylabels=None,
         zlim=None,
         values_title="count",
+        gt_field=None,
+        pred_field=None,
         colorscale=None,
         grid_opacity=0.1,
         bg_opacity=0.25,
@@ -1985,6 +2007,8 @@ class InteractiveHeatmap(PlotlyInteractivePlot):
         self.ylabels = ylabels
         self.zlim = zlim
         self.values_title = values_title
+        self.gt_field = gt_field
+        self.pred_field = pred_field
         self.colorscale = colorscale
         self.grid_opacity = grid_opacity
         self.bg_opacity = bg_opacity
@@ -2179,11 +2203,13 @@ class InteractiveHeatmap(PlotlyInteractivePlot):
         xticks = np.arange(num_cols)
         yticks = np.arange(num_rows)
         X, Y = np.meshgrid(xticks, yticks)
+        truth = self.gt_field or "truth"
+        predicted = self.pred_field or "predicted"
 
         hover_lines = [
             "<b>%s: %%{z}</b>" % self.values_title,
-            "truth: %{y}",
-            "predicted: %{x}",
+            f"{truth}: %{{y}}",
+            f"{predicted}: %{{x}}",
         ]
         hovertemplate = "<br>".join(hover_lines) + "<extra></extra>"
 
@@ -2559,6 +2585,60 @@ def _plot_scatter_numeric(
     return figure
 
 
+def _set_map_type_to_figure(
+    figure,
+    map_type,
+):
+    if map_type == "satellite":
+        figure.update_layout(
+            mapbox_style="white-bg",
+            mapbox_layers=[
+                {
+                    "below": "traces",
+                    "sourcetype": "raster",
+                    "sourceattribution": "United States Geological Survey",
+                    "source": [
+                        "https://basemap.nationalmap.gov/arcgis/rest/services/USGSImageryOnly/MapServer/tile/{z}/{y}/{x}"
+                    ],
+                }
+            ],
+            margin={"r": 0, "t": 0, "l": 0, "b": 0},
+        )
+    elif map_type == "roadmap":
+        figure.update_layout(mapbox_style="carto-positron")
+    else:
+        figure.update_layout(mapbox_style="carto-positron")
+        msg = "Unsupported map type '%s'; defaulted to 'roadmap'" % map_type
+        warnings.warn(msg)
+
+
+def _set_map_type_to_figure(
+    figure,
+    map_type,
+):
+    if map_type == "satellite":
+        figure.update_layout(
+            mapbox_style="white-bg",
+            mapbox_layers=[
+                {
+                    "below": "traces",
+                    "sourcetype": "raster",
+                    "sourceattribution": "United States Geological Survey",
+                    "source": [
+                        "https://basemap.nationalmap.gov/arcgis/rest/services/USGSImageryOnly/MapServer/tile/{z}/{y}/{x}"
+                    ],
+                }
+            ],
+            margin={"r": 0, "t": 0, "l": 0, "b": 0},
+        )
+    elif map_type == "roadmap":
+        figure.update_layout(mapbox_style="carto-positron")
+    else:
+        figure.update_layout(mapbox_style="carto-positron")
+        msg = "Unsupported map type '%s'; defaulted to 'roadmap'" % map_type
+        warnings.warn(msg)
+
+
 def _plot_scatter_mapbox_categorical(
     coords,
     labels,
@@ -2643,7 +2723,6 @@ def _plot_scatter_mapbox_categorical(
 
     zoom, (center_lon, center_lat) = _compute_zoom_center(coords)
     figure.update_layout(
-        mapbox_style="carto-positron",
         mapbox=dict(center=dict(lat=center_lat, lon=center_lon), zoom=zoom),
         legend_title_text=colorbar_title,
         legend_itemsizing="constant",
@@ -2741,7 +2820,6 @@ def _plot_scatter_mapbox_categorical_single_trace(
 
     zoom, (center_lon, center_lat) = _compute_zoom_center(coords)
     figure.update_layout(
-        mapbox_style="carto-positron",
         mapbox=dict(center=dict(lat=center_lat, lon=center_lon), zoom=zoom),
     )
 
@@ -2836,7 +2914,6 @@ def _plot_scatter_mapbox_numeric(
 
     zoom, (center_lon, center_lat) = _compute_zoom_center(coords)
     figure.update_layout(
-        mapbox_style="carto-positron",
         mapbox=dict(center=dict(lat=center_lat, lon=center_lon), zoom=zoom),
     )
 
@@ -2913,7 +2990,6 @@ def _plot_scatter_mapbox_density(
 
     zoom, (center_lon, center_lat) = _compute_zoom_center(coords)
     figure.update_layout(
-        mapbox_style="carto-positron",
         mapbox=dict(center=dict(lat=center_lat, lon=center_lon), zoom=zoom),
     )
 
