@@ -1,7 +1,7 @@
 """
 Dataset samples.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -292,6 +292,7 @@ class _SampleMixin(object):
         fields=None,
         omit_fields=None,
         merge_lists=True,
+        merge_embedded_docs=False,
         overwrite=True,
         expand_schema=True,
         validate=True,
@@ -338,6 +339,9 @@ class _SampleMixin(object):
                 (when ``overwrite`` is True) or kept (when ``overwrite`` is
                 False) when their ``id`` matches a label from the provided
                 sample
+            merge_embedded_docs (False): whether to merge the attributes of
+                embedded documents (True) rather than merging the entire
+                top-level field (False)
             overwrite (True): whether to overwrite (True) or skip (False)
                 existing fields and label elements
             expand_schema (True): whether to dynamically add new fields
@@ -368,6 +372,7 @@ class _SampleMixin(object):
             fields=fields,
             omit_fields=omit_fields,
             merge_lists=merge_lists,
+            merge_embedded_docs=merge_embedded_docs,
             overwrite=overwrite,
             expand_schema=expand_schema,
             validate=validate,
@@ -380,6 +385,7 @@ class _SampleMixin(object):
                 fields=frame_fields,
                 omit_fields=omit_frame_fields,
                 merge_lists=merge_lists,
+                merge_embedded_docs=merge_embedded_docs,
                 overwrite=overwrite,
                 expand_schema=expand_schema,
                 validate=validate,
@@ -492,7 +498,7 @@ class Sample(_SampleMixin, Document, metaclass=SampleSingleton):
     Args:
         filepath: the path to the data on disk. The path is converted to an
             absolute path (if necessary) via
-            :func:`fiftyone.core.utils.normalize_path`
+            :func:`fiftyone.core.storage.normalize_path`
         tags (None): a list of tags for the sample
         metadata (None): a :class:`fiftyone.core.metadata.Metadata` instance
         **kwargs: additional fields to dynamically set on the sample
@@ -526,15 +532,17 @@ class Sample(_SampleMixin, Document, metaclass=SampleSingleton):
         d = self._dataset._sample_collection.find_one({"_id": self._id})
         self._doc = self._dataset._sample_dict_to_doc(d)
 
-    def reload(self, hard=False):
+    def reload(self, hard=False, include_frames=True):
         """Reloads the sample from the database.
 
         Args:
             hard (False): whether to reload the sample's schema in addition to
                 its field values. This is necessary if new fields may have been
                 added to the dataset schema
+            include_frames (True): whether to reload any in-memory frames of
+                video samples
         """
-        if self.media_type == fomm.VIDEO:
+        if self.media_type == fomm.VIDEO and include_frames:
             self.frames.reload(hard=hard)
 
         super().reload(hard=hard)
@@ -550,12 +558,15 @@ class Sample(_SampleMixin, Document, metaclass=SampleSingleton):
             )
 
         if self.media_type == fomm.VIDEO:
-            frame_ops = self.frames._save(deferred=deferred)
+            _sample_ops, frame_ops = self.frames._save(deferred=deferred)
         else:
-            frame_ops = []
+            _sample_ops, frame_ops = None, None
 
         sample_ops = super()._save(deferred=deferred)
 
+        if _sample_ops:
+            sample_ops.extend(_sample_ops)
+
         return sample_ops, frame_ops
 
     @classmethod
@@ -739,12 +750,15 @@ class SampleView(_SampleMixin, DocumentView):
 
     def _save(self, deferred=False):
         if self.media_type == fomm.VIDEO:
-            frame_ops = self.frames._save(deferred=deferred)
+            _sample_ops, frame_ops = self.frames._save(deferred=deferred)
         else:
-            frame_ops = []
+            _sample_ops, frame_ops = None, None
 
         sample_ops = super()._save(deferred=deferred)
 
+        if _sample_ops:
+            sample_ops.extend(_sample_ops)
+
         return sample_ops, frame_ops
 
 
@@ -761,13 +775,18 @@ def _apply_confidence_thresh(label, confidence_thresh):
             k: _apply_confidence_thresh(v, confidence_thresh)
             for k, v in label.items()
         }
-    elif isinstance(label, fol._LABEL_LIST_FIELDS):
+    elif isinstance(label, fol.Keypoints):
+        for keypoint in label.keypoints:
+            keypoint.apply_confidence_threshold(confidence_thresh)
+    elif isinstance(label, fol._HasLabelList):
         labels = [
             l
             for l in getattr(label, label._LABEL_LIST_FIELD)
             if l.confidence is not None and l.confidence >= confidence_thresh
         ]
         setattr(label, label._LABEL_LIST_FIELD, labels)
+    elif isinstance(label, fol.Keypoint):
+        label.apply_confidence_threshold(confidence_thresh)
     elif hasattr(label, "confidence"):
         if label.confidence is None or label.confidence < confidence_thresh:
             label = None
@@ -776,8 +795,6 @@ def _apply_confidence_thresh(label, confidence_thresh):
 
 
 def _is_frames_dict(label):
-    return (
-        label
-        and isinstance(label, dict)
-        and fofu.is_frame_number(next(iter(label.keys())))
+    return isinstance(label, dict) and fofu.is_frame_number(
+        next(iter(label.keys()))
     )
