@@ -1,34 +1,40 @@
 """
 Session class for interacting with the FiftyOne App.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
+
 from collections import defaultdict
+from dataclasses import asdict
 from functools import wraps
 import logging
-from packaging.version import Version
-import pkg_resources
+import os
 import time
 import typing as t
-import webbrowser
 from uuid import uuid4
+import webbrowser
+
+from dacite import from_dict
 
 try:
     import IPython.display
 except:
     pass
 
+import eta.core.serial as etas
+
 import fiftyone as fo
+import fiftyone.core.odm.dataset as food
+from fiftyone.core.odm.workspace import default_workspace_factory, Space
 import fiftyone.constants as focn
 import fiftyone.core.dataset as fod
 from fiftyone.core.config import AppConfig
 import fiftyone.core.context as focx
 import fiftyone.core.plots as fop
 import fiftyone.core.service as fos
-from fiftyone.core.spaces import default_spaces, Space
-from fiftyone.core.state import StateDescription
+from fiftyone.core.state import build_color_scheme, StateDescription
 import fiftyone.core.utils as fou
 import fiftyone.core.view as fov
 
@@ -37,7 +43,16 @@ from fiftyone.core.session.events import (
     CaptureNotebookCell,
     CloseSession,
     DeactivateNotebookCell,
+    LabelData,
     ReactivateNotebookCell,
+    Refresh,
+    SelectLabels,
+    SelectSamples,
+    SetColorScheme,
+    SetDatasetColorScheme,
+    SetSample,
+    SetSpaces,
+    SetGroupSlice,
     StateUpdate,
 )
 import fiftyone.core.session.notebooks as fosn
@@ -62,10 +77,6 @@ _session = None
 _server_services = {}
 _subscribed_sessions = defaultdict(set)
 
-_APP_DESKTOP_MESSAGE = """
-Desktop App launched.
-"""
-
 _APP_WEB_MESSAGE = """
 App launched. Point your web browser to http://localhost:{0}
 """
@@ -97,16 +108,40 @@ intended to start an App instance or a remote session from a script, you should
 call `session.wait()` to keep the session (and the script) alive.
 """
 
+_WELCOME_MESSAGE = """
+Welcome to
+
+███████╗██╗███████╗████████╗██╗   ██╗ ██████╗ ███╗   ██╗███████╗
+██╔════╝██║██╔════╝╚══██╔══╝╚██╗ ██╔╝██╔═══██╗████╗  ██║██╔════╝
+█████╗  ██║█████╗     ██║    ╚████╔╝ ██║   ██║██╔██╗ ██║█████╗
+██╔══╝  ██║██╔══╝     ██║     ╚██╔╝  ██║   ██║██║╚██╗██║██╔══╝
+██║     ██║██║        ██║      ██║   ╚██████╔╝██║ ╚████║███████╗
+╚═╝     ╚═╝╚═╝        ╚═╝      ╚═╝    ╚═════╝ ╚═╝  ╚═══╝╚══════╝ v{0}
+
+If you're finding FiftyOne helpful, here's how you can get involved:
+
+|
+|  ⭐⭐⭐ Give the project a star on GitHub ⭐⭐⭐
+|  https://github.com/voxel51/fiftyone
+|
+|  🚀🚀🚀 Join the FiftyOne Discord community 🚀🚀🚀
+|  https://community.voxel51.com/
+|
+"""
+
 
 def launch_app(
     dataset: fod.Dataset = None,
     view: fov.DatasetView = None,
+    sample_id: str = None,
+    group_id: str = None,
     spaces: Space = None,
+    color_scheme: food.ColorScheme = None,
     plots: fop.PlotManager = None,
     port: int = None,
     address: str = None,
     remote: bool = False,
-    desktop: bool = None,
+    browser: str = None,
     height: int = None,
     auto: bool = True,
     config: AppConfig = None,
@@ -121,8 +156,15 @@ def launch_app(
             :class:`fiftyone.core.view.DatasetView` to load
         view (None): an optional :class:`fiftyone.core.view.DatasetView` to
             load
-        spaces (None): an optional :class:`fiftyone.core.spaces.Space` instance
-            defining a space configuration to load
+        sample_id (None): an optional :class:`fiftyone.core.sample.Sample` ID
+            to load in the modal
+        group_id (None): an optional :class:`fiftyone.core.groups.Group` ID
+            to load in the modal
+        spaces (None): an optional :class:`fiftyone.core.odm.workspace.Space`
+            instance defining a space configuration to load
+        color_scheme (None): an optional
+            :class:`fiftyone.core.odm.dataset.ColorScheme` defining a custom
+            color scheme to use
         plots (None): an optional
             :class:`fiftyone.core.plots.manager.PlotManager` to connect to this
             session
@@ -132,9 +174,9 @@ def launch_app(
             ``fiftyone.config.default_app_address`` is used
         remote (False): whether this is a remote session, and opening the App
             should not be attempted
-        desktop (None): whether to launch the App in the browser (False) or as
-            a desktop App (True). If None, ``fiftyone.config.desktop_app`` is
-            used. Not applicable to notebook contexts
+        browser (None): an optional browser to use to open the App. If None,
+            the default browser will be used. Refer to list of supported
+            browsers at https://docs.python.org/3/library/webbrowser.html
         height (None): an optional height, in pixels, at which to render App
             instances in notebook cells. Only applicable in notebook contexts
         auto (True): whether to automatically show a new App window
@@ -147,28 +189,18 @@ def launch_app(
         a :class:`Session`
     """
     global _session  # pylint: disable=global-statement
-
-    #
-    # Note, we always `close_app()` here rather than just calling
-    # `session.open()` if a session already exists, because the app may have
-    # been closed in some way other than `session.close()` --- e.g., the user
-    # closing the GUI --- in which case the underlying Electron process may
-    # still exist; in this case, `session.open()` does not seem to reopen the
-    # app
-    #
-    # @todo this can probably be improved
-    #
-    close_app()
-
     _session = Session(
         dataset=dataset,
         view=view,
+        sample_id=sample_id,
+        group_id=group_id,
         spaces=spaces,
+        color_scheme=color_scheme,
         plots=plots,
         port=port,
         address=address,
         remote=remote,
-        desktop=desktop,
+        browser=browser,
         height=height,
         auto=auto,
         config=config,
@@ -176,14 +208,14 @@ def launch_app(
 
     if _session.remote:
         logger.info(_REMOTE_INSTRUCTIONS.strip().format(_session.server_port))
-    elif _session.desktop:
-        logger.info(_APP_DESKTOP_MESSAGE.strip())
     elif focx.is_notebook_context():
         if not auto:
             logger.info(_APP_NOTEBOOK_MESSAGE.strip())
     else:
         logger.info(_APP_WEB_MESSAGE.strip().format(_session.server_port))
 
+    _log_welcome_message_if_allowed()
+
     return _session
 
 
@@ -244,6 +276,14 @@ class Session(object):
         :attr:`Session.view` property of the session to your
         :class:`fiftyone.core.view.DatasetView`.
 
+    -   To load a specific sample in the modal, simply set the
+        :attr:`Session.sample_id` property of the session to the ID of the
+        :class:`fiftyone.core.sample.Sample`.
+
+    -   To load a specific group in the modal, simply set the
+        :attr:`Session.group_id` property of the session to the ID of the
+        :class:`fiftyone.core.groups.Group`.
+
     -   To attach/remove interactive plots, use the methods exposed on the
         :attr:`Session.plots` property of the session.
 
@@ -271,8 +311,15 @@ class Session(object):
             :class:`fiftyone.core.view.DatasetView` to load
         view (None): an optional :class:`fiftyone.core.view.DatasetView` to
             load
-        spaces (None): an optional :class:`fiftyone.core.spaces.Space` instance
-            defining a space configuration to load
+        sample_id (None): an optional :class:`fiftyone.core.sample.Sample` ID
+            to load in the modal
+        group_id (None): an optional :class:`fiftyone.core.groups.Group` ID
+            to load in the modal
+        spaces (None): an optional :class:`fiftyone.core.odm.workspace.Space`
+            instance defining a space configuration to load
+        color_scheme (None): an optional
+            :class:`fiftyone.core.odm.dataset.ColorScheme` defining a custom
+            color scheme to use
         plots (None): an optional
             :class:`fiftyone.core.plots.manager.PlotManager` to connect to this
             session
@@ -282,9 +329,9 @@ class Session(object):
             ``fiftyone.config.default_app_address`` is used
         remote (False): whether this is a remote session, and opening the App
             should not be attempted
-        desktop (None): whether to launch the App in the browser (False) or as
-            a desktop App (True). If None, ``fiftyone.config.desktop_app`` is
-            used. Not applicable to notebook contexts (e.g., Jupyter and Colab)
+        browser (None): an optional browser to use to open the App. If None,
+            the default browser will be used. Refer to list of supported
+            browsers at https://docs.python.org/3/library/webbrowser.html
         height (None): an optional height, in pixels, at which to render App
             instances in notebook cells. Only applicable in notebook contexts
         auto (True): whether to automatically show a new App window
@@ -298,16 +345,19 @@ class Session(object):
         self,
         dataset: t.Union[fod.Dataset, fov.DatasetView] = None,
         view: fov.DatasetView = None,
-        view_name: str = None,
+        sample_id: str = None,
+        group_id: str = None,
         spaces: Space = None,
+        color_scheme: food.ColorScheme = None,
         plots: fop.PlotManager = None,
         port: int = None,
         address: str = None,
         remote: bool = False,
-        desktop: bool = None,
+        browser: str = None,
         height: int = None,
         auto: bool = True,
         config: AppConfig = None,
+        view_name: str = None,
     ) -> None:
         focx.init_context()
 
@@ -315,7 +365,7 @@ class Session(object):
             view = dataset
             dataset = dataset._root_dataset
 
-        self._validate(dataset, view, spaces, plots, config)
+        self._validate(dataset, view, spaces, color_scheme, plots, config)
 
         if port is None:
             port = fo.config.default_app_port
@@ -331,7 +381,7 @@ class Session(object):
             )
 
         if address is None:
-            if fou.is_docker() or focx.is_databricks_context():
+            if fou.is_container() or focx.is_databricks_context():
                 address = "0.0.0.0"
             else:
                 address = fo.config.default_app_address
@@ -350,13 +400,6 @@ class Session(object):
         self._disable_wait_warning = False
         self._notebook_cells: t.Dict[str, fosn.NotebookCell] = {}
 
-        if desktop is None:
-            desktop = (
-                fo.config.desktop_app
-                if not focx.is_notebook_context()
-                else False
-            )
-
         self.plots = plots
 
         final_view_name = view_name
@@ -364,30 +407,35 @@ class Session(object):
             final_view_name = view.name
 
         if spaces is None:
-            spaces = default_spaces.copy()
+            spaces = default_workspace_factory()
 
         self._state = StateDescription(
-            config=config,
             dataset=view._root_dataset if view is not None else dataset,
             view=view,
             view_name=final_view_name,
+            sample_id=sample_id,
+            group_id=group_id,
+            group_slice=_pull_group_slice(dataset, view),
             spaces=spaces,
+            color_scheme=build_color_scheme(color_scheme, dataset, config),
+            config=config,
         )
         self._client = fosc.Client(
             address=address,
             auto=auto,
-            desktop=desktop,
             port=port,
             remote=remote,
             start_time=self._get_time(),
         )
-        self._client.run(self._state)
+        self._client.open(self._state)
         _attach_listeners(self)
         _register_session(self)
 
         if self.auto and focx.is_notebook_context():
             self.show(height=config.notebook_height)
 
+        self.browser = browser
+
         if self.remote:
             if focx.is_notebook_context():
                 raise ValueError(
@@ -396,21 +444,6 @@ class Session(object):
 
             return
 
-        if self.desktop:
-            if focx.is_notebook_context():
-                raise ValueError(
-                    "Cannot open a Desktop App instance from a %s notebook"
-                    % focx._get_context()
-                )
-
-            if not focn.DEV_INSTALL:
-                import_desktop()
-
-            self._app_service = fos.AppService(
-                server_port=port, server_address=address
-            )
-            return
-
         if not focx.is_notebook_context():
             self.open()
             return
@@ -420,6 +453,7 @@ class Session(object):
         dataset: t.Optional[t.Union[fod.Dataset, fov.DatasetView]],
         view: t.Optional[fov.DatasetView],
         spaces: t.Optional[Space],
+        color_scheme: t.Optional[food.ColorScheme],
         plots: t.Optional[fop.PlotManager],
         config: t.Optional[AppConfig],
     ) -> None:
@@ -441,6 +475,14 @@ class Session(object):
                 % (Space, type(spaces))
             )
 
+        if color_scheme is not None and not isinstance(
+            color_scheme, food.ColorScheme
+        ):
+            raise ValueError(
+                "`color_scheme` must be a %s or None; found %s"
+                % (food.ColorScheme, type(color_scheme))
+            )
+
         if plots is not None and not isinstance(plots, fop.PlotManager):
             raise ValueError(
                 "`plots` must be a %s or None; found %s"
@@ -465,6 +507,7 @@ class Session(object):
                 # logger may already have been garbage-collected
                 print(_WAIT_INSTRUCTIONS)
 
+            self._client.close()
             _unregister_session(self)
         except:
             # e.g. globals were already garbage-collected
@@ -494,15 +537,14 @@ class Session(object):
         """Whether the session is remote."""
         return self._client.remote
 
-    @property
-    def desktop(self) -> bool:
-        """Whether the session is connected to a desktop App."""
-        return self._client.desktop
-
     @property
     def url(self) -> str:
         """The URL of the session."""
-        return focx.get_url(self.server_address, self.server_port)
+        return focx.get_url(
+            self.server_address,
+            self.server_port,
+            proxy_url=self.config.proxy_url,
+        )
 
     @property
     def config(self) -> AppConfig:
@@ -538,16 +580,45 @@ class Session(object):
 
         self._state.config = config
 
+    @property
+    def group_id(self) -> t.Optional[str]:
+        """The current :class:`fiftyone.core.groups.Group` ID in the modal, if
+        any.
+        """
+        return self._state.group_id
+
+    @group_id.setter  # type: ignore
+    def group_id(self, group_id: t.Optional[str]) -> None:
+        if group_id is not None and not isinstance(group_id, str):
+            raise ValueError(f"unexpected group id value '{group_id}'")
+
+        self._state.group_id = group_id
+        self._client.send_event(SetSample(group_id=group_id))
+
+    @property
+    def sample_id(self) -> t.Optional[str]:
+        """The current :class:`fiftyone.core.sample.Sample` ID in the modal, if
+        any.
+        """
+        return self._state.sample_id
+
+    @sample_id.setter  # type: ignore
+    def sample_id(self, sample_id: t.Optional[str]) -> None:
+        if sample_id is not None and not isinstance(sample_id, str):
+            raise ValueError(f"unexpected sample id value '{sample_id}'")
+
+        self._state.sample_id = sample_id
+        self._client.send_event(SetSample(sample_id=sample_id))
+
     @property
     def spaces(self) -> Space:
         """The layout state for the session."""
         return self._state.spaces
 
     @spaces.setter  # type: ignore
-    @update_state()
     def spaces(self, spaces: t.Optional[Space]) -> None:
         if spaces is None:
-            spaces = default_spaces.copy()
+            spaces = default_workspace_factory()
 
         if not isinstance(spaces, Space):
             raise ValueError(
@@ -556,6 +627,35 @@ class Session(object):
             )
 
         self._state.spaces = spaces
+        self._client.send_event(SetSpaces(spaces=spaces.to_dict()))
+
+    def load_workspace(self, workspace: str) -> None:
+        """Loads the given saved workspace.
+
+        Args:
+            workspace: the name of a saved workspace
+        """
+        spaces = self.dataset.load_workspace(workspace)
+        self.spaces = spaces
+
+    @property
+    def color_scheme(self) -> food.ColorScheme:
+        """The color scheme for the session."""
+        return self._state.color_scheme
+
+    @color_scheme.setter  # type: ignore
+    def color_scheme(self, color_scheme: t.Optional[food.ColorScheme]) -> None:
+        if color_scheme is None:
+            color_scheme = build_color_scheme(None, self.dataset, self.config)
+
+        if not isinstance(color_scheme, food.ColorScheme):
+            raise ValueError(
+                "`Session.color_scheme` must be a %s or None; found %s"
+                % (food.ColorScheme, type(color_scheme))
+            )
+
+        self._state.color_scheme = color_scheme
+        self._client.send_event(SetColorScheme.from_odm(color_scheme))
 
     @property
     def _collection(self) -> t.Union[fod.Dataset, fov.DatasetView, None]:
@@ -578,21 +678,32 @@ class Session(object):
                 % (fod.Dataset, type(dataset))
             )
 
-        if dataset is not None:
-            dataset._reload()
-
-        self._state.dataset = dataset
-        self._state.view = None
-        self._state.selected = []
-        self._state.selected_labels = []
-        self._state.spaces = default_spaces
+        self._set_dataset(dataset)
 
     @update_state()
     def clear_dataset(self) -> None:
         """Clears the current :class:`fiftyone.core.dataset.Dataset` from the
         session, if any.
         """
-        self._state.dataset = None
+        self._set_dataset(None)
+
+    def _set_dataset(self, dataset):
+        if dataset is not None:
+            dataset._reload()
+            self._state.group_slice = dataset.group_slice
+        else:
+            self._state.group_slice = None
+
+        self._state.color_scheme = build_color_scheme(
+            None, dataset, self.config
+        )
+        self._state.dataset = dataset
+        self._state.group_id = None
+        self._state.sample_id = None
+        self._state.spaces = default_workspace_factory()
+        self._state.selected = []
+        self._state.selected_labels = []
+        self._state.view = None
 
     @property
     def view(self) -> t.Union[fov.DatasetView, None]:
@@ -610,22 +721,32 @@ class Session(object):
                 % (fov.DatasetView, type(view))
             )
 
-        self._state.view = view
-
-        if view is not None:
-            view._root_dataset._reload()
-            self._state.dataset = view._root_dataset
-            self._state.view_name = view.name
-
-        self._state.selected = []
-        self._state.selected_labels = []
+        self._set_view(view)
 
     @update_state()
     def clear_view(self) -> None:
         """Clears the current :class:`fiftyone.core.view.DatasetView` from the
         session, if any.
         """
-        self._state.view = None
+        self._set_view(None)
+
+    def _set_view(self, view):
+        if view is None:
+            self._state.group_slice = None
+            self._state.view = None
+            self._state.view_name = None
+        else:
+            if view._root_dataset != self.dataset:
+                self._set_dataset(view._root_dataset)
+
+            self._state.group_slice = view.group_slice
+            self._state.view = view
+            self._state.view_name = view.name
+
+        self._state.group_id = None
+        self._state.sample_id = None
+        self._state.selected = []
+        self._state.selected_labels = []
 
     @property
     def has_plots(self) -> bool:
@@ -656,7 +777,7 @@ class Session(object):
 
     def refresh(self) -> None:
         """Refreshes the current App window."""
-        self._client.send_event(StateUpdate(state=self._state, refresh=True))
+        self._client.send_event(Refresh(state=self._state))
 
     @property
     def selected(self) -> t.List[str]:
@@ -666,16 +787,15 @@ class Session(object):
         return list(self._state.selected)
 
     @selected.setter  # type: ignore
-    @update_state()
     def selected(self, sample_ids: t.List[str]) -> None:
         self._state.selected = list(sample_ids) if sample_ids else []
+        self._client.send_event(SelectSamples(sample_ids))
 
-    @update_state()
     def clear_selected(self) -> None:
         """Clears the currently selected samples, if any."""
         self._state.selected = []
+        self._client.send_event(SelectSamples([]))
 
-    @update_state()
     def select_samples(
         self,
         ids: t.Optional[t.Union[str, t.Iterable[str]]] = None,
@@ -694,6 +814,7 @@ class Session(object):
             ids = []
 
         self._state.selected = list(ids)
+        self._client.send_event(SelectSamples(self._state.selected))
 
     @property
     def selected_labels(self) -> t.List[dict]:
@@ -703,6 +824,7 @@ class Session(object):
 
         -   ``label_id``: the ID of the label
         -   ``sample_id``: the ID of the sample containing the label
+        -   ``instance_id``: the instance ID of the label (optional)
         -   ``field``: the field name containing the label
         -   ``frame_number``: the frame number containing the label (only
             applicable to video samples)
@@ -710,15 +832,17 @@ class Session(object):
         return list(self._state.selected_labels)
 
     @selected_labels.setter  # type: ignore
-    @update_state()
     def selected_labels(self, labels: dict) -> None:
         self._state.selected_labels = list(labels) if labels else []
+        self._client.send_event(
+            from_dict(SelectLabels, dict(labels=self._state.selected_labels))
+        )
 
-    @update_state()
     def select_labels(
         self,
         labels: t.Optional[t.List[dict]] = None,
         ids: t.Optional[t.Union[str, t.Iterable[str]]] = None,
+        instance_ids: t.Optional[t.Union[str, t.Iterable[str]]] = None,
         tags: t.Optional[t.Union[str, t.Iterable[str]]] = None,
         fields: t.Optional[t.Union[str, t.Iterable[str]]] = None,
     ) -> None:
@@ -731,15 +855,17 @@ class Session(object):
         Args:
             labels (None): a list of dicts specifying the labels to select
             ids (None): an ID or iterable of IDs of the labels to select
+            instance_ids (None): an instance ID or iterable of instance IDs of
+                the labels to select
             tags (None): a tag or iterable of tags of labels to select
             fields (None): a field or iterable of fields from which to select
         """
         if labels is None and self._collection:
             labels = self._collection._get_selected_labels(
-                ids=ids, tags=tags, fields=fields
+                ids=ids, instance_ids=instance_ids, tags=tags, fields=fields
             )
 
-        self._state.selected_labels = list(labels or [])
+        self.selected_labels = list(labels or [])
 
     @update_state()
     def clear_selected_labels(self) -> None:
@@ -854,11 +980,14 @@ class Session(object):
             type_ = "colab"
         elif focx.is_databricks_context():
             type_ = "databricks"
-        elif self.desktop:
-            type_ = "desktop"
         else:
             type_ = None
 
+        if self.group_id:
+            elements.append(("Group:", self.group_id))
+        elif self.sample_id:
+            elements.append(("Sample:", self.sample_id))
+
         if type_ is None:
             elements.append(("Session URL:", self.url))
         else:
@@ -882,12 +1011,12 @@ class Session(object):
 
         -   Notebooks: calls :meth:`Session.show` to open a new App window in
             the output of your current cell
-        -   Desktop: opens the desktop App, if necessary
         -   Other (non-remote): opens the App in a new browser tab
         """
-        if self.remote:
-            logger.warning("Remote sessions cannot open new App windows")
-            return
+        _register_session(self)
+
+        if not self._client.is_open:
+            self._client.open(self._state)
 
         if self.plots:
             self.plots.connect()
@@ -896,17 +1025,13 @@ class Session(object):
             self.show()
             return
 
-        if self.desktop:
-            self._app_service.start()
-            return
-
         self.open_tab()
 
     def open_tab(self) -> None:
         """Opens the App in a new tab of your browser.
 
-        This method can be called from Jupyter notebooks and in desktop App
-        mode to override the default location of the App.
+        This method can be called from Jupyter notebooks to override the
+        default location of the App.
         """
         if self.remote:
             logger.warning("Remote sessions cannot open new App windows")
@@ -920,7 +1045,11 @@ class Session(object):
             )
             return
 
-        webbrowser.open(self.url, new=2)
+        if self.browser is None:
+            webbrowser.open(self.url, new=2)
+        else:
+            # open in a specified browser
+            webbrowser.get(self.browser).open(self.url, new=2)
 
     @update_state()
     def show(self, height: int = None) -> None:
@@ -931,7 +1060,7 @@ class Session(object):
         Args:
             height (None): a height, in pixels, for the App
         """
-        if not focx.is_notebook_context() or self.desktop:
+        if not focx.is_notebook_context():
             return
 
         self.freeze()
@@ -951,7 +1080,7 @@ class Session(object):
         )
 
         self._notebook_cells[uuid] = cell
-        fosn.display(self._client, cell)
+        fosn.display(self._client, cell, self.config.proxy_url)
 
     def no_show(self) -> fou.SetAttributes:
         """Returns a context manager that temporarily prevents new App
@@ -986,11 +1115,8 @@ class Session(object):
     def wait(self, wait: float = 3) -> None:
         """Blocks execution until the App is closed by the user.
 
-        For browser Apps, all connected windows (tabs) must be closed before
-        this method will unblock.
-
-        For desktop Apps, all positive ``wait`` values are equivalent;
-        execution will immediately unblock when the App is closed.
+        All connected windows (tabs) must be closed before this method will
+        unblock.
 
         Args:
             wait (3): the number of seconds to wait for a new App connection
@@ -1005,25 +1131,21 @@ class Session(object):
             if wait < 0:
                 while True:
                     time.sleep(10)
-            elif self.remote or not self._client.desktop:
+            else:
                 self._wait_closed = False
                 while not self._wait_closed:
                     time.sleep(wait)
-            else:
-                self._app_service.wait()
         except KeyboardInterrupt:
             self._disable_wait_warning = True
             raise
 
     def close(self) -> None:
         """Closes the session and terminates the App, if necessary."""
-        if self.remote:
-            return
-
-        if self._client._connected and focx._get_context() == focx._NONE:
-            self._client.send_event(CloseSession())
+        if self._client.is_open and focx.is_notebook_context():
+            self.freeze()
 
         self.plots.disconnect()
+        self.__del__()
 
     def freeze(self) -> None:
         """Screenshots the active App cell, replacing it with a static image.
@@ -1039,21 +1161,72 @@ class Session(object):
 
 
 def _attach_listeners(session: "Session"):
-    on_close_session: t.Callable[[CloseSession], None] = lambda event: setattr(
+    on_close_session: t.Callable[[CloseSession], None] = lambda _: setattr(
         session, "_wait_closed", True
     )
     session._client.add_event_listener("close_session", on_close_session)
 
+    on_refresh: t.Callable[[Refresh], None] = lambda event: _on_refresh(
+        session, event.state
+    )
+    session._client.add_event_listener("refresh", on_refresh)
+
     on_state_update: t.Callable[[StateUpdate], None] = lambda event: setattr(
         session, "_state", event.state
     )
     session._client.add_event_listener("state_update", on_state_update)
 
+    on_select_samples: t.Callable[
+        [SelectSamples], None
+    ] = lambda event: setattr(session._state, "selected", event.sample_ids)
+    session._client.add_event_listener("select_samples", on_select_samples)
+
+    on_select_labels: t.Callable[
+        [SelectLabels], None
+    ] = lambda event: _on_select_labels(session._state, event)
+    session._client.add_event_listener("select_labels", on_select_labels)
+
+    on_set_color_scheme: t.Callable[
+        [SetColorScheme], None
+    ] = lambda event: setattr(session._state, "color_scheme", event.to_odm())
+    session._client.add_event_listener("set_color_scheme", on_set_color_scheme)
+
+    on_set_dataset_color_scheme: t.Callable[
+        [SetDatasetColorScheme], None
+    ] = lambda _: _on_refresh(session, None)
+    session._client.add_event_listener(
+        "set_dataset_color_scheme", on_set_dataset_color_scheme
+    )
+
+    on_set_group_slice: t.Callable[
+        [SetGroupSlice], None
+    ] = lambda event: setattr(
+        session._state.dataset,
+        "group_slice",
+        event.slice,
+    )
+    session._client.add_event_listener("set_group_slice", on_set_group_slice)
+
+    def on_set_sample(event: SetSample) -> None:
+        session._state.sample_id = event.sample_id
+        session._state.group_id = event.group_id
+
+    session._client.add_event_listener("set_sample", on_set_sample)
+
+    on_set_spaces: t.Callable[[SetSpaces], None] = lambda event: setattr(
+        session._state,
+        "spaces",
+        Space.from_dict(event.spaces),
+    )
+    session._client.add_event_listener("set_spaces", on_set_spaces)
+
     if focx.is_notebook_context() and not focx.is_colab_context():
 
         def on_capture_notebook_cell(event: CaptureNotebookCell) -> None:
             event.subscription in session._notebook_cells and fosn.capture(
-                session._notebook_cells[event.subscription], event
+                session._notebook_cells[event.subscription],
+                event,
+                proxy_url=session.config.proxy_url,
             )
 
         session._client.add_event_listener(
@@ -1070,45 +1243,6 @@ def _attach_listeners(session: "Session"):
         )
 
 
-def import_desktop() -> None:
-    """Attempts to import :mod:`fiftyone.desktop`
-
-    Raises:
-        RuntimeError: If matching ``fiftyone-desktop`` version is not
-        installed
-    """
-    try:
-        # pylint: disable=unused-import
-        import fiftyone.desktop
-    except ImportError as e:
-        raise RuntimeError(
-            "You must `pip install fiftyone[desktop]` in order to launch the "
-            "desktop App"
-        ) from e
-
-    # Get `fiftyone-desktop` requirement for current `fiftyone` install
-    fiftyone_dist = pkg_resources.get_distribution("fiftyone")
-    requirements = fiftyone_dist.requires(extras=["desktop"])
-    desktop_req = [r for r in requirements if r.name == "fiftyone-desktop"][0]
-
-    desktop_dist = pkg_resources.get_distribution("fiftyone-desktop")
-
-    if not desktop_req.specifier.contains(
-        Version(desktop_dist.version).base_version
-    ):
-        raise RuntimeError(
-            "fiftyone==%s requires fiftyone-desktop%s, but you have "
-            "fiftyone-desktop==%s installed.\n"
-            "Run `pip install fiftyone[desktop]` to install the proper "
-            "desktop package version"
-            % (
-                fiftyone_dist.version,
-                desktop_req.specifier,
-                desktop_dist.version,
-            )
-        )
-
-
 def _register_session(session: Session) -> None:
     global _server_services  # pylint: disable=global-statement
     if session.server_port not in _server_services:
@@ -1131,3 +1265,50 @@ def _unregister_session(session: Session) -> None:
         if session.server_port in _server_services:
             service = _server_services.pop(session.server_port)
             service.stop()
+
+
+def _log_welcome_message_if_allowed():
+    """Logs a welcome message the first time this function is called on a
+    machine with a new FiftyOne version installed, if allowed.
+    """
+    if os.environ.get("FIFTYONE_SERVER", None):
+        return
+
+    try:
+        last_version = etas.load_json(focn.WELCOME_PATH)["version"]
+    except:
+        last_version = None
+
+    if focn.VERSION == last_version:
+        return
+
+    logger.info(_WELCOME_MESSAGE.format(focn.VERSION))
+
+    try:
+        etas.write_json({"version": focn.VERSION}, focn.WELCOME_PATH)
+    except:
+        pass
+
+
+def _on_refresh(session: Session, state: t.Optional[StateDescription]):
+    if state:
+        session._state = state
+
+    if session.dataset is not None:
+        session.dataset.reload()
+
+
+def _on_select_labels(state: StateDescription, event: SelectLabels):
+    setattr(state, "selected_labels", [asdict(data) for data in event.labels])
+
+
+def _pull_group_slice(
+    dataset: t.Optional[fod.Dataset], view: t.Optional[fov.DatasetView]
+) -> t.Union[None, str]:
+    if view is not None:
+        return view.group_slice
+
+    if dataset is not None:
+        return dataset.group_slice
+
+    return None
