@@ -1,10 +1,11 @@
 """
 FiftyOne singleton implementations.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
+
 from collections import defaultdict
 import weakref
 
@@ -36,7 +37,15 @@ class DatasetSingleton(type):
             instance.__init__(name=name, _create=_create, *args, **kwargs)
             name = instance.name  # `__init__` may have changed `name`
         else:
-            instance._update_last_loaded_at()
+            try:
+                instance._update_last_loaded_at(
+                    force=kwargs.get("_force_load", False)
+                )
+            except ValueError:
+                instance._deleted = True
+                return cls.__call__(
+                    name=name, _create=_create, *args, **kwargs
+                )
 
         cls._instances[name] = instance
 
@@ -111,7 +120,7 @@ class SampleSingleton(DocumentSingleton):
 
     def _reload_instance(cls, obj):
         # pylint: disable=no-value-for-parameter
-        cls._reload_doc(obj._doc.collection_name, obj.id)
+        cls._reload_doc(obj._doc.collection_name, obj.id, include_frames=True)
 
     def _rename_fields(cls, collection_name, field_names, new_field_names):
         """Renames the field on all in-memory samples in the collection."""
@@ -145,18 +154,22 @@ class SampleSingleton(DocumentSingleton):
             for field_name in field_names:
                 sample._doc._data.pop(field_name, None)
 
-    def _reload_doc(cls, collection_name, sample_id, hard=False):
+    def _reload_doc(
+        cls, collection_name, sample_id, hard=False, include_frames=False
+    ):
         """Reloads the backing document for the given sample if it is
         in-memory.
         """
         if collection_name not in cls._instances:
             return
 
-        sample = cls._instances[collection_name].get(sample_id, None)
+        sample = cls._instances[collection_name].get(str(sample_id), None)
         if sample is not None:
-            sample.reload(hard=hard)
+            sample.reload(hard=hard, include_frames=include_frames)
 
-    def _reload_docs(cls, collection_name, sample_ids=None, hard=False):
+    def _reload_docs(
+        cls, collection_name, sample_ids=None, hard=False, include_frames=False
+    ):
         """Reloads the backing documents for in-memory samples in the
         collection.
 
@@ -168,15 +181,17 @@ class SampleSingleton(DocumentSingleton):
         samples = cls._instances[collection_name]
 
         if sample_ids is not None:
-            sample_ids = set(sample_ids)
+            sample_ids = set(str(_id) for _id in sample_ids)
             for sample in samples.values():
                 if sample.id in sample_ids:
-                    sample.reload(hard=hard)
+                    sample.reload(hard=hard, include_frames=include_frames)
         else:
             for sample in samples.values():
-                sample.reload(hard=hard)
+                sample.reload(hard=hard, include_frames=include_frames)
 
-    def _sync_docs(cls, collection_name, sample_ids, hard=False):
+    def _sync_docs(
+        cls, collection_name, sample_ids, hard=False, include_frames=False
+    ):
         """Syncs the backing documents for all in-memory samples in the
         collection according to the following rules:
 
@@ -188,11 +203,11 @@ class SampleSingleton(DocumentSingleton):
 
         samples = cls._instances[collection_name]
 
-        sample_ids = set(sample_ids)
+        sample_ids = set(str(_id) for _id in sample_ids)
         reset_ids = set()
         for sample in samples.values():
             if sample.id in sample_ids:
-                sample.reload(hard=hard)
+                sample.reload(hard=hard, include_frames=include_frames)
             else:
                 reset_ids.add(sample.id)
                 sample._reset_backing_doc()
@@ -214,7 +229,7 @@ class SampleSingleton(DocumentSingleton):
         if sample_ids is not None:
             samples = cls._instances[collection_name]
             for sample_id in sample_ids:
-                sample = samples.pop(sample_id, None)
+                sample = samples.pop(str(sample_id), None)
                 if sample is not None:
                     sample._reset_backing_doc()
         else:
@@ -222,6 +237,10 @@ class SampleSingleton(DocumentSingleton):
             for sample in samples.values():
                 sample._reset_backing_doc()
 
+    def _clear(cls, collection_name):
+        """Removes all samples for the given collection from the cache."""
+        cls._instances.pop(collection_name, None)
+
 
 class FrameSingleton(DocumentSingleton):
     """Singleton metaclass for :class:`fiftyone.core.frame.Frame`.
@@ -264,7 +283,9 @@ class FrameSingleton(DocumentSingleton):
         """Returns a frame number -> Frame dict containing all in-memory frame
         instances for the specified sample.
         """
-        return dict(cls._instances.get(collection_name, {}).get(sample_id, {}))
+        return dict(
+            cls._instances.get(collection_name, {}).get(str(sample_id), {})
+        )
 
     def _rename_fields(cls, collection_name, field_names, new_field_names):
         """Renames the field on all in-memory frames in the collection."""
@@ -306,7 +327,7 @@ class FrameSingleton(DocumentSingleton):
         if collection_name not in cls._instances:
             return
 
-        frames = cls._instances[collection_name].get(sample_id, {})
+        frames = cls._instances[collection_name].get(str(sample_id), {})
         frame = frames.get(frame_number, None)
         if frame is not None:
             frame.reload(hard=hard)
@@ -324,7 +345,7 @@ class FrameSingleton(DocumentSingleton):
             return
 
         samples = cls._instances[collection_name]
-        frames = samples.get(sample_id, {})
+        frames = samples.get(str(sample_id), {})
 
         if not frames:
             return
@@ -359,7 +380,7 @@ class FrameSingleton(DocumentSingleton):
 
         samples = cls._instances[collection_name]
 
-        sample_ids = set(sample_ids)
+        sample_ids = set(str(_id) for _id in sample_ids)
         reset_ids = set()
         for sample_id, frames in samples.items():
             if sample_id in sample_ids:
@@ -373,7 +394,9 @@ class FrameSingleton(DocumentSingleton):
         for sample_id in reset_ids:
             frames = samples.pop(sample_id)
 
-    def _reload_docs(cls, collection_name, sample_ids=None, hard=False):
+    def _reload_docs(
+        cls, collection_name, sample_ids=None, frame_ids=None, hard=False
+    ):
         """Reloads the backing documents for in-memory frames in the
         collection.
 
@@ -386,11 +409,17 @@ class FrameSingleton(DocumentSingleton):
         samples = cls._instances[collection_name]
 
         if sample_ids is not None:
-            sample_ids = set(sample_ids)
+            sample_ids = set(str(_id) for _id in sample_ids)
             for sample_id in sample_ids:
                 frames = samples.get(sample_id, {})
                 for frame in frames.values():
                     frame.reload(hard=hard)
+        elif frame_ids is not None:
+            frame_ids = set(str(_id) for _id in frame_ids)
+            for frames in samples.values():
+                for frame in frames.values():
+                    if frame.id in frame_ids:
+                        frame.reload(hard=hard)
         else:
             for frames in samples.values():
                 for frame in frames.values():
@@ -408,7 +437,7 @@ class FrameSingleton(DocumentSingleton):
         if sample_ids is not None:
             samples = cls._instances[collection_name]
             for sample_id in sample_ids:
-                frames = samples.pop(sample_id, {})
+                frames = samples.pop(str(sample_id), {})
                 for frame in frames.values():
                     frame._reset_backing_doc()
         else:
@@ -430,7 +459,7 @@ class FrameSingleton(DocumentSingleton):
             return
 
         samples = cls._instances[collection_name]
-        frames = samples.get(sample_id, {})
+        frames = samples.get(str(sample_id), {})
 
         if not frames:
             return
@@ -467,7 +496,7 @@ class FrameSingleton(DocumentSingleton):
 
         samples = cls._instances[collection_name]
 
-        frame_ids = set(frame_ids)
+        frame_ids = set(str(_id) for _id in frame_ids)
         reset = []
 
         if keep:
@@ -485,3 +514,7 @@ class FrameSingleton(DocumentSingleton):
 
         for sample_id, fn in reset:
             samples[sample_id].pop(fn)
+
+    def _clear(cls, collection_name):
+        """Removes all frames for the given collection from the cache."""
+        cls._instances.pop(collection_name, None)
