@@ -1,18 +1,19 @@
 """
 FiftyOne Server queries.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
-import typing as t
+
+from dataclasses import asdict
 from datetime import date, datetime
 from enum import Enum
-import os
+import logging
+import typing as t
 
-import asyncio
-import eta.core.serial as etas
 import eta.core.utils as etau
+from pymongo.errors import ExecutionTimeout
 import strawberry as gql
 from bson import ObjectId, json_util
 
@@ -23,24 +24,31 @@ import fiftyone.core.context as focx
 import fiftyone.core.dataset as fod
 import fiftyone.core.media as fom
 from fiftyone.core.odm import SavedViewDocument
+import fiftyone.core.stages as fosg
 from fiftyone.core.state import SampleField, serialize_fields
 import fiftyone.core.uid as fou
+from fiftyone.core.utils import run_sync_task
 import fiftyone.core.view as fov
 
 import fiftyone.server.aggregate as fosa
 from fiftyone.server.aggregations import aggregate_resolver
+from fiftyone.server.color import ColorBy, ColorScheme
 from fiftyone.server.data import Info
 from fiftyone.server.dataloader import get_dataloader_resolver
-import fiftyone.server.events as fose
-from fiftyone.server.metadata import MediaType
+from fiftyone.server.events import get_state
+from fiftyone.server.indexes import Index, from_dict as indexes_from_dict
+from fiftyone.server.lightning import lightning_resolver
 from fiftyone.server.paginator import Connection, get_paginator_resolver
 from fiftyone.server.samples import (
     SampleFilter,
     SampleItem,
     paginate_samples,
 )
-from fiftyone.server.scalars import BSONArray, JSON
+from fiftyone.server.scalars import BSON, BSONArray, JSON
+from fiftyone.server.stage_definitions import stage_definitions
+from fiftyone.server.exceptions import QueryTimeout
 from fiftyone.server.utils import from_dict
+from fiftyone.server.workspace import Workspace
 
 
 ID = gql.scalar(
@@ -52,10 +60,16 @@ DATASET_FILTER = [{"sample_collection_name": {"$regex": "^samples\\."}}]
 DATASET_FILTER_STAGE = [{"$match": DATASET_FILTER[0]}]
 
 
+@gql.type
+class ActiveFields:
+    exclude: t.Optional[bool]
+    paths: t.Optional[t.List[str]]
+
+
 @gql.type
 class Group:
     name: str
-    media_type: MediaType
+    media_type: str
 
 
 @gql.type
@@ -126,7 +140,6 @@ class BrainRunConfig(RunConfig):
             return cls(
                 embeddings_field=self.embeddings_field,
                 patches_field=self.patches_field,
-                supports_prompts=self.supports_prompts,
             )
         except:
             return None
@@ -201,36 +214,34 @@ class NamedKeypointSkeleton(KeypointSkeleton):
     name: str
 
 
-@gql.enum
-class SidebarMode(Enum):
-    all = "all"
-    best = "best"
-    fast = "fast"
-
-
 @gql.type
 class DatasetAppConfig:
+    active_fields: t.Optional[ActiveFields]
+    color_scheme: t.Optional[ColorScheme]
+    disable_frame_filtering: t.Optional[bool] = None
+    dynamic_groups_target_frame_rate: int = 30
+    grid_media_field: str = "filepath"
     media_fields: t.Optional[t.List[str]]
+    modal_media_field: str = "filepath"
+    media_fallback: bool = False
     plugins: t.Optional[JSON]
     sidebar_groups: t.Optional[t.List[SidebarGroup]]
-    sidebar_mode: t.Optional[SidebarMode]
-    modal_media_field: t.Optional[str] = gql.field(default="filepath")
-    grid_media_field: t.Optional[str] = "filepath"
     spaces: t.Optional[JSON]
 
 
 @gql.type
 class Dataset:
     id: gql.ID
+    dataset_id: gql.ID
     name: str
     created_at: t.Optional[date]
     last_loaded_at: t.Optional[datetime]
     persistent: bool
     group_media_types: t.Optional[t.List[Group]]
     group_field: t.Optional[str]
-    group_slice: t.Optional[str]
     default_group_slice: t.Optional[str]
-    media_type: t.Optional[MediaType]
+    media_type: t.Optional[str]
+    parent_media_type: t.Optional[str]
     mask_targets: t.List[NamedTargets]
     default_mask_targets: t.Optional[t.List[Target]]
     sample_fields: t.List[SampleField]
@@ -247,20 +258,58 @@ class Dataset:
     app_config: t.Optional[DatasetAppConfig]
     info: t.Optional[JSON]
 
-    @gql.field
-    def stages(self, slug: t.Optional[str] = None) -> t.Optional[BSONArray]:
-        if not slug:
-            return None
+    estimated_frame_count: t.Optional[int]
+    estimated_sample_count: int
+    frame_indexes: t.Optional[t.List[Index]]
+    sample_indexes: t.Optional[t.List[Index]]
 
-        for view in self.saved_views:
-            if view.slug == slug:
-                return view.stage_dicts()
+    frame_collection_name: gql.Private[t.Optional[str]]
+    sample_collection_name: gql.Private[t.Optional[str]]
+
+    @gql.field
+    def stages(
+        self, slug: t.Optional[str] = None, view: t.Optional[BSONArray] = None
+    ) -> t.Optional[BSONArray]:
+        if slug:
+            for view in self.saved_views:
+                if view.slug == slug:
+                    return view.stage_dicts()
+
+        return view or []
+
+    @gql.field
+    async def estimated_sample_count(self, info: Info = None) -> int:
+        return await info.context.db[
+            self.sample_collection_name
+        ].estimated_document_count()
+
+    @gql.field
+    async def estimated_frame_count(
+        self, info: Info = None
+    ) -> t.Optional[int]:
+        if self.frame_collection_name:
+            return await info.context.db[
+                self.frame_collection_name
+            ].estimated_document_count()
+
+    @gql.field
+    async def workspace(
+        self, slug: t.Optional[str], info: Info
+    ) -> t.Optional[Workspace]:
+        if slug:
+            doc = await info.context.db["workspaces"].find_one({"slug": slug})
+
+            if doc:
+                doc["id"] = doc.pop("_id")
+                doc["dataset_id"] = doc.pop("_dataset_id")
+                return from_dict(Workspace, doc)
 
         return None
 
     @staticmethod
     def modifier(doc: dict) -> dict:
         doc["id"] = doc.pop("_id")
+        doc["dataset_id"] = doc["id"]
         doc["default_mask_targets"] = _convert_targets(
             doc.get("default_mask_targets", {})
         )
@@ -268,9 +317,9 @@ class Dataset:
             NamedTargets(name=name, targets=_convert_targets(targets))
             for name, targets in doc.get("mask_targets", {}).items()
         ]
-        doc["sample_fields"] = _flatten_fields(
-            [], doc.get("sample_fields", [])
-        )
+        flat = _flatten_fields([], doc.get("sample_fields", []))
+        doc["sample_fields"] = flat
+
         doc["frame_fields"] = _flatten_fields([], doc.get("frame_fields", []))
         doc["brain_methods"] = list(doc.get("brain_methods", {}).values())
         doc["evaluations"] = list(doc.get("evaluations", {}).values())
@@ -279,25 +328,28 @@ class Dataset:
             dict(name=name, **data)
             for name, data in doc.get("skeletons", {}).items()
         )
-        doc["group_media_types"] = [
-            Group(name=name, media_type=media_type)
-            for name, media_type in doc.get("group_media_types", {}).items()
-        ]
+        doc["group_media_types"] = []
         doc["default_skeletons"] = doc.get("default_skeletons", None)
+
+        # gql private fields must always be present
+        doc.setdefault("frame_collection_name", None)
+
         return doc
 
     @classmethod
     async def resolver(
         cls,
         name: str,
-        view: t.Optional[BSONArray],
-        info: Info,
+        info: Info = None,
         saved_view_slug: t.Optional[str] = gql.UNSET,
+        view: t.Optional[BSONArray] = None,
     ) -> t.Optional["Dataset"]:
         return await serialize_dataset(
             dataset_name=name,
             serialized_view=view,
             saved_view_slug=saved_view_slug,
+            dicts=False,
+            update_last_loaded_at=True,
         )
 
 
@@ -306,13 +358,6 @@ dataset_dataloader = get_dataloader_resolver(
 )
 
 
-@gql.enum
-class ColorBy(Enum):
-    field = "field"
-    instance = "instance"
-    label = "label"
-
-
 @gql.enum
 class Theme(Enum):
     browser = "browser"
@@ -326,7 +371,10 @@ class AppConfig:
     color_pool: t.List[str]
     colorscale: str
     grid_zoom: int
+    enable_query_performance: bool
+    default_query_performance: bool
     loop_videos: bool
+    multicolor_keypoints: bool
     notebook_height: int
     plugins: t.Optional[JSON]
     show_confidence: bool
@@ -334,17 +382,25 @@ class AppConfig:
     show_label: bool
     show_skeletons: bool
     show_tooltip: bool
-    sidebar_mode: SidebarMode
     theme: Theme
     timezone: t.Optional[str]
     use_frame_number: bool
     spaces: t.Optional[JSON]
+    disable_frame_filtering: bool = False
+    max_query_time: t.Optional[int] = None
+    media_fallback: bool = False
+
+
+@gql.type
+class SchemaResult:
+    field_schema: t.List[SampleField]
+    frame_field_schema: t.List[SampleField]
 
 
 @gql.type
 class Query(fosa.AggregateQuery):
-
     aggregations = gql.field(resolver=aggregate_resolver)
+    lightning = gql.field(resolver=lightning_resolver)
 
     @gql.field
     def colorscale(self) -> t.Optional[t.List[t.List[int]]]:
@@ -355,7 +411,7 @@ class Query(fosa.AggregateQuery):
 
     @gql.field
     def config(self) -> AppConfig:
-        config = fose.get_state().config
+        config = get_state().config
         d = config.serialize()
         d["timezone"] = fo.config.timezone
         return from_dict(AppConfig, d)
@@ -372,7 +428,11 @@ class Query(fosa.AggregateQuery):
     def do_not_track(self) -> bool:
         return fo.config.do_not_track
 
-    dataset: Dataset = gql.field(resolver=Dataset.resolver)
+    @gql.field
+    async def estimated_dataset_count(self, info: Info = None) -> int:
+        return await info.context.db.datasets.estimated_document_count()
+
+    dataset = gql.field(resolver=Dataset.resolver)
     datasets: Connection[Dataset, str] = gql.field(
         resolver=get_paginator_resolver(
             Dataset, "created_at", DATASET_FILTER_STAGE, "datasets"
@@ -387,37 +447,64 @@ class Query(fosa.AggregateQuery):
         first: t.Optional[int] = 20,
         after: t.Optional[str] = None,
         filter: t.Optional[SampleFilter] = None,
-    ) -> Connection[SampleItem, str]:
-        return await paginate_samples(
-            dataset, view, None, first, after, sample_filter=filter
-        )
+        filters: t.Optional[BSON] = None,
+        extended_stages: t.Optional[BSON] = None,
+        pagination_data: t.Optional[bool] = True,
+        sort_by: t.Optional[str] = None,
+        desc: t.Optional[bool] = False,
+        hint: t.Optional[str] = None,
+        max_query_time: t.Optional[int] = None,
+        dynamic_group: t.Optional[BSON] = None,
+    ) -> t.Annotated[
+        t.Union[Connection[SampleItem, str], QueryTimeout],
+        gql.union("PaginateSamplesResponse"),
+    ]:
+
+        try:
+            return await paginate_samples(
+                dataset,
+                view,
+                filters,
+                first,
+                after,
+                sample_filter=filter,
+                extended_stages=extended_stages,
+                pagination_data=pagination_data,
+                sort_by=sort_by,
+                desc=desc,
+                hint=hint,
+                max_query_time=max_query_time,
+                dynamic_group=dynamic_group,
+            )
+        except ExecutionTimeout:
+            return QueryTimeout(query_time=max_query_time)
 
     @gql.field
     async def sample(
-        self, dataset: str, view: BSONArray, filter: SampleFilter
+        self,
+        dataset: str,
+        view: BSONArray,
+        filter: SampleFilter,
+        filters: t.Optional[JSON] = None,
     ) -> t.Optional[SampleItem]:
         samples = await paginate_samples(
-            dataset, view, None, 1, sample_filter=filter
+            dataset,
+            view,
+            filters,
+            1,
+            sample_filter=filter,
+            pagination_data=False,
         )
         if samples.edges:
             return samples.edges[0].node
 
         return None
 
-    @gql.field
-    def teams_submission(self) -> bool:
-        isfile = os.path.isfile(foc.TEAMS_PATH)
-        if isfile:
-            submitted = etas.load_json(foc.TEAMS_PATH)["submitted"]
-        else:
-            submitted = False
-
-        return submitted
+    stage_definitions = gql.field(stage_definitions)
 
     @gql.field
     def uid(self) -> str:
-        uid, _ = fou.get_user_id()
-        return uid
+        return fou.get_user_id()
 
     @gql.field
     def version(self) -> str:
@@ -425,10 +512,64 @@ class Query(fosa.AggregateQuery):
 
     @gql.field
     def saved_views(self, dataset_name: str) -> t.Optional[t.List[SavedView]]:
-        ds = fod.load_dataset(dataset_name)
-        return [
-            SavedView.from_doc(view_doc) for view_doc in ds._doc.saved_views
-        ]
+        try:
+            ds = fod.load_dataset(dataset_name)
+            ds.reload()  # reload to avoid singleton issues
+            return [
+                SavedView.from_doc(view_doc)
+                for view_doc in ds._doc.saved_views
+            ]
+        except:
+            return None
+
+    @gql.field
+    def schema_for_view_stages(
+        self,
+        dataset_name: str,
+        view_stages: BSONArray,
+    ) -> SchemaResult:
+        try:
+            ds = fod.load_dataset(dataset_name)
+            if view_stages:
+                view = fov.DatasetView._build(ds, view_stages or [])
+
+                if ds.media_type == fom.VIDEO:
+                    frame_schema = serialize_fields(
+                        view.get_frame_field_schema(flat=True)
+                    )
+                    field_schema = serialize_fields(
+                        view.get_field_schema(flat=True)
+                    )
+                    return SchemaResult(
+                        field_schema=field_schema,
+                        frame_field_schema=frame_schema,
+                    )
+
+                return SchemaResult(
+                    field_schema=serialize_fields(
+                        view.get_field_schema(flat=True)
+                    ),
+                    frame_field_schema=[],
+                )
+            if ds.media_type == fom.VIDEO:
+                frames_field_schema = serialize_fields(
+                    ds.get_frame_field_schema(flat=True)
+                )
+                field_schema = serialize_fields(ds.get_field_schema(flat=True))
+                return SchemaResult(
+                    field_schema=field_schema,
+                    frame_field_schema=frames_field_schema,
+                )
+
+            return SchemaResult(
+                field_schema=serialize_fields(ds.get_field_schema(flat=True)),
+                frame_field_schema=[],
+            )
+        except Exception as e:
+            return SchemaResult(
+                field_schema=[],
+                frame_field_schema=[],
+            )
 
 
 def _flatten_fields(
@@ -436,7 +577,13 @@ def _flatten_fields(
 ) -> t.List[t.Dict]:
     result = []
     for field in fields:
-        key = field.pop("name")
+        key = field.pop("name", None)
+        if key is None:
+            # Issues with concurrency can cause this to happen.
+            # Until it's fixed, just ignore these fields to avoid throwing hard
+            # errors when loading in the app.
+            logging.debug("Skipping field with no name: %s", field)
+            continue
         field_path = path + [key]
         field["path"] = ".".join(field_path)
         result.append(field)
@@ -456,17 +603,27 @@ async def serialize_dataset(
     dataset_name: str,
     serialized_view: BSONArray,
     saved_view_slug: t.Optional[str] = None,
+    dicts=True,
+    update_last_loaded_at=False,
 ) -> Dataset:
     def run():
-        dataset = fod.load_dataset(dataset_name)
+        if not fod.dataset_exists(dataset_name):
+            return None
+
+        dataset = fo.Dataset(dataset_name, _create=False, _force_load=True)
         dataset.reload()
         view_name = None
         try:
             doc = dataset._get_saved_view_doc(saved_view_slug, slug=True)
             view = dataset.load_saved_view(doc.name)
             view_name = view.name
+            if serialized_view:
+                for stage in serialized_view:
+                    view = view.add_stage(fosg.ViewStage._from_dict(stage))
         except:
-            view = fov.DatasetView._build(dataset, serialized_view or [])
+            view: fov.DatasetView = fov.DatasetView._build(
+                dataset, serialized_view or []
+            )
 
         doc = dataset._doc.to_dict(no_dereference=True)
         Dataset.modifier(doc)
@@ -475,27 +632,46 @@ async def serialize_dataset(
         data.view_name = view_name
         data.saved_view_slug = saved_view_slug
 
+        group_media_types = view._get_group_media_types() or {}
+        data.group_media_types = [
+            Group(name=name, media_type=media_type)
+            for name, media_type in group_media_types.items()
+        ]
+
         collection = dataset.view()
         if view is not None:
+            # unique id for for the relay global store
+            #
+            # until a schema is with respect to a view and not a dataset this
+            # is required
+            data.id = ObjectId()
             if view._dataset != dataset:
                 d = view._dataset._serialize()
                 data.media_type = d["media_type"]
                 data.view_cls = etau.get_class_name(view)
 
-            if view.media_type != data.media_type:
-                data.media_type = view.media_type
+            data.parent_media_type = view._parent_media_type
+            data.media_type = view.media_type
 
             collection = view
 
         data.sample_fields = serialize_fields(
             collection.get_field_schema(flat=True)
         )
+
         data.frame_fields = serialize_fields(
             collection.get_frame_field_schema(flat=True)
         )
 
-        if dataset.media_type == fom.GROUP:
-            data.group_slice = collection.group_slice
+        if dicts:
+            saved_views = []
+            for view in data.saved_views:
+                view_dict = asdict(view)
+                view_dict["view_name"] = view.view_name()
+                view_dict["stage_dicts"] = view.stage_dicts()
+                saved_views.append(view_dict)
+
+            data.saved_views = saved_views
 
         for brain_method in data.brain_methods:
             try:
@@ -523,8 +699,34 @@ async def serialize_dataset(
                 supports_least_similarity,
             )
 
+        _assign_estimated_counts(data, dataset)
+        _assign_lightning_info(data, dataset)
+
         return data
 
-    loop = asyncio.get_running_loop()
+    return await run_sync_task(run)
 
-    return await loop.run_in_executor(None, run)
+
+def _assign_estimated_counts(dataset: Dataset, fo_dataset: fo.Dataset):
+    setattr(
+        dataset,
+        "estimated_sample_count",
+        fo_dataset._sample_collection.estimated_document_count(),
+    )
+    setattr(
+        dataset,
+        "estimated_frame_count",
+        (
+            fo_dataset._frame_collection.estimated_document_count()
+            if fo_dataset._frame_collection_name
+            else None
+        ),
+    )
+
+
+def _assign_lightning_info(dataset: Dataset, fo_dataset: fo.Dataset):
+    dataset.sample_indexes, dataset.frame_indexes = indexes_from_dict(
+        fo_dataset.get_index_information(
+            include_stats=True, _keep_index_names=True
+        )
+    )
