@@ -1,14 +1,14 @@
 """
 FiftyOne Server /aggregation and /tagging routes
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
+
 from starlette.endpoints import HTTPEndpoint
 from starlette.requests import Request
 
-import fiftyone as fo
 import fiftyone.core.aggregations as foa
 import fiftyone.core.view as fov
 
@@ -26,7 +26,9 @@ class Aggregate(HTTPEndpoint):
         stages = data.get("view", None)
         aggregations = data.get("aggregations", [])
 
-        view = fosv.get_view(dataset, stages=stages, filters=filters)
+        view = await fosv.get_view(
+            dataset, stages=stages, filters=filters, awaitable=True
+        )
 
         if sample_ids:
             view = fov.make_optimized_select_view(view, sample_ids)
