@@ -1,7 +1,7 @@
 """
 FiftyOne dataset types.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -706,7 +706,7 @@ class FiftyOneVideoLabelsDataset(VideoLabelsDataset):
 class BDDDataset(ImageLabelsDataset):
     """A labeled dataset consisting of images and their associated multitask
     predictions saved in
-    `Berkeley DeepDrive (BDD) format <https://bdd-data.berkeley.edu>`_.
+    `Berkeley DeepDrive (BDD) format <http://bdd-data.berkeley.edu>`_.
 
     See :ref:`this page <BDDDataset-import>` for importing datasets of this
     type, and see :ref:`this page <BDDDataset-export>` for exporting datasets
@@ -887,3 +887,14 @@ class LegacyFiftyOneDataset(Dataset):
         import fiftyone.utils.data as foud
 
         return foud.LegacyFiftyOneDatasetExporter
+
+
+class PlacesDataset(ImageClassificationDataset):
+    """A labeled dataset consisting of images and their associated labels
+    from the `Places dataset <http://places2.csail.mit.edu/index.html>`.
+    """
+
+    def get_dataset_importer_cls(self):
+        import fiftyone.utils.places as foup
+
+        return foup.PlacesDatasetImporter
