@@ -2,7 +2,7 @@
 Utilities for working with the
 `ActivityNet dataset <http://activity-net.org/index.html>`.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -54,9 +54,8 @@ def download_activitynet_split(
         copy_files (True): whether to move (False) or create copies (True) of
             the source files when populating ``dataset_dir``. This is only
             relevant when a ``source_dir`` is provided
-        num_workers (None): the number of threads to use when downloading
-            individual video. By default, ``multiprocessing.cpu_count()`` is
-            used
+        num_workers (None): a suggested number of threads to use when
+            downloading individual videos
         shuffle (False): whether to randomly shuffle the order in which samples
             are chosen for partial downloads
         seed (None): a random seed to use when shuffling
@@ -371,7 +370,7 @@ class ActivityNetDatasetManager(object):
         return self.info.existing_split_sample_ids(split)
 
     def split_sample_ids(self, split):
-        return self.info.split_sample_ids[split]
+        return self.info.split_sample_ids(split)
 
     def process_source(self, source_dir, copy_files):
         source_dir = os.path.expanduser(source_dir)
@@ -727,6 +726,12 @@ class ActivityNetDatasetManager(object):
             prev_errors = {}
 
         for video, e in download_errors.items():
+            if not isinstance(e, (str, int, float, bool)) and e != None:
+                try:
+                    e = str(e)
+                except:
+                    e = "Cannot parse error message"
+
             if e in prev_errors:
                 prev_errors[e].append(video)
             else:
