@@ -1,20 +1,20 @@
 """
 Annotation utilities.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
 from collections import defaultdict, OrderedDict
 from copy import deepcopy
 import getpass
+import inspect
 import logging
 import os
 
 from bson import ObjectId
 
 import eta.core.annotations as etaa
-import eta.core.image as etai
 import eta.core.utils as etau
 
 import fiftyone as fo
@@ -28,6 +28,7 @@ import fiftyone.core.metadata as fom
 import fiftyone.core.utils as fou
 import fiftyone.core.validation as fov
 import fiftyone.utils.eta as foue
+import fiftyone.utils.image as foui
 
 
 logger = logging.getLogger(__name__)
@@ -91,6 +92,7 @@ def annotate(
     The natively provided backends and their associated config classes are:
 
     -   ``"cvat"``: :class:`fiftyone.utils.cvat.CVATBackendConfig`
+    -   ``"labelstudio"``: :class:`fiftyone.utils.labelstudio.LabelStudioBackendConfig`
     -   ``"labelbox"``: :class:`fiftyone.utils.labelbox.LabelboxBackendConfig`
 
     See :ref:`this page <requesting-annotations>` for more information about
@@ -160,7 +162,11 @@ def annotate(
             define their per-field attributes (in addition to any per-class
             attributes)
         mask_targets (None): a dict mapping pixel values to semantic label
-            strings. Only applicable when annotating semantic segmentations
+            strings. Only applicable when annotating semantic segmentations.
+            All new label fields must have mask targets provided via one of the
+            supported methods. For existing label fields, if mask targets are
+            not provided by this argument nor ``label_schema``, any applicable
+            mask targets stored on your dataset will be used, if available
         allow_additions (True): whether to allow new labels to be added. Only
             applicable when editing existing label fields
         allow_deletions (True): whether to allow labels to be deleted. Only
@@ -187,13 +193,6 @@ def annotate(
     Returns:
         an :class:`AnnotationResults`
     """
-    # @todo support this?
-    if samples._dataset._is_frames:
-        raise ValueError("Annotating frames views is not supported")
-
-    # @todo support this?
-    if samples._dataset._is_clips:
-        raise ValueError("Annotating clips views is not supported")
 
     # Convert to equivalent regular view containing the same labels
     if samples._dataset._is_patches:
@@ -214,14 +213,21 @@ def annotate(
     anno_backend.ensure_requirements()
 
     supported_media_types = anno_backend.supported_media_types
-    if samples.media_type == fomm.GROUP:
-        raise fomm.SelectGroupSlicesError(supported_media_types)
-    elif samples.media_type not in supported_media_types:
+    if samples.media_type not in supported_media_types:
+        if samples.media_type == fomm.GROUP:
+            raise fomm.SelectGroupSlicesError(supported_media_types)
+
         raise fomm.MediaTypeError(
             "The '%s' backend does not supported annotating '%s' collections"
             % (anno_backend.config.name, samples.media_type)
         )
 
+    if samples._is_clips and not anno_backend.supports_clips_views:
+        raise ValueError(
+            "Backend '%s' does not support annotating clips views"
+            % anno_backend.config.name
+        )
+
     label_schema, samples = _build_label_schema(
         samples,
         anno_backend,
@@ -248,6 +254,8 @@ def annotate(
         samples, anno_key, launch_editor=launch_editor
     )
 
+    results._finalize_id_map()
+
     anno_backend.save_run_results(samples, anno_key, results)
 
     return results
@@ -266,6 +274,9 @@ def _parse_config(name, label_schema, **kwargs):
     if name is None:
         name = fo.annotation_config.default_backend
 
+    if inspect.isclass(name):
+        return name(None, label_schema, **kwargs)
+
     backends = fo.annotation_config.backends
 
     if name not in backends:
@@ -461,7 +472,11 @@ def _build_label_schema(
 
         if _label_type == "segmentation":
             _mask_targets, _classes = _get_mask_targets(
-                samples, mask_targets, _label_field, _label_info
+                samples,
+                mask_targets,
+                _label_field,
+                _label_info,
+                _existing_field,
             )
         else:
             _mask_targets = None
@@ -619,14 +634,10 @@ def _get_label_type(samples, backend, label_type, label_field, label_info):
     if "type" in label_info:
         label_type = label_info["type"]
 
-    field_name, is_frame_field = samples._handle_frame_field(label_field)
+    field = samples.get_field(label_field)
+    is_frame_field = samples._is_frame_field(label_field)
 
-    if is_frame_field:
-        schema = samples.get_frame_field_schema()
-    else:
-        schema = samples.get_field_schema()
-
-    if field_name not in schema:
+    if field is None:
         if label_type is None:
             raise ValueError(
                 "You must specify a type for new label field '%s'"
@@ -635,7 +646,7 @@ def _get_label_type(samples, backend, label_type, label_field, label_info):
 
         return label_type, is_frame_field, False, False
 
-    _existing_type = _get_backend_field_type(backend, schema[field_name])
+    _existing_type = _get_backend_field_type(backend, field)
     _multiple_types = isinstance(_existing_type, list)
 
     if label_type is not None:
@@ -834,13 +845,30 @@ def _parse_classes_dict(
     return {"classes": classes, "attributes": attributes}
 
 
-def _get_mask_targets(samples, mask_targets, label_field, label_info):
+def _get_mask_targets(
+    samples,
+    mask_targets,
+    label_field,
+    label_info,
+    existing_field,
+):
     if "mask_targets" in label_info:
         mask_targets = label_info["mask_targets"]
 
     if mask_targets is None:
-        mask_targets = {i: str(i) for i in range(1, 256)}
-        mask_targets[0] = "background"
+        if not existing_field:
+            raise ValueError(
+                "You must provide mask targets for new label field '%s'"
+                % label_field
+            )
+
+        if label_field in samples.mask_targets:
+            mask_targets = samples.mask_targets[label_field]
+        elif samples.default_mask_targets:
+            mask_targets = samples.default_mask_targets
+        else:
+            mask_targets = {i: str(i) for i in range(1, 256)}
+            mask_targets[0] = "background"
 
     classes = [c for v, c in mask_targets.items() if v != 0]
 
@@ -923,11 +951,12 @@ def _format_attributes(backend, attributes):
         if not attr:
             attr = backend.recommend_attr_tool(name, None)
 
-        attr_type = attr.get("type", None)
-        values = attr.get("values", None)
-        default = attr.get("default", None)
-        mutable = attr.get("mutable", None)
-        read_only = attr.get("read_only", None)
+        _attr = deepcopy(attr)
+        attr_type = _attr.pop("type", None)
+        values = _attr.pop("values", None)
+        default = _attr.pop("default", None)
+        mutable = _attr.pop("mutable", None)
+        read_only = _attr.pop("read_only", None)
 
         if attr_type is None:
             raise ValueError(
@@ -946,7 +975,7 @@ def _format_attributes(backend, attributes):
                 )
             )
 
-        _attr = {"type": attr_type}
+        _attr["type"] = attr_type
 
         # Parse `values` property
         if values is not None:
@@ -986,6 +1015,7 @@ def load_annotations(
     dest_field=None,
     unexpected="prompt",
     cleanup=False,
+    progress=None,
     **kwargs,
 ):
     """Downloads the labels from the given annotation run from the annotation
@@ -1000,7 +1030,7 @@ def load_annotations(
         anno_key: an annotation key
         dest_field (None): an optional name of a new destination field
             into which to load the annotations, or a dict mapping field names
-            in the run's label schema to new desination field names
+            in the run's label schema to new destination field names
         unexpected ("prompt"): how to deal with any unexpected labels that
             don't match the run's label schema when importing. The supported
             values are:
@@ -1014,6 +1044,9 @@ def load_annotations(
                 or ``None`` if there aren't any
         cleanup (False): whether to delete any informtation regarding this run
             from the annotation backend after loading the annotations
+        progress (None): whether to render a progress bar (True/False), use the
+            default value ``fiftyone.config.show_progress_bars`` (None), or a
+            progress callback function to invoke instead
         **kwargs: keyword arguments for the run's
             :meth:`fiftyone.core.annotation.AnnotationMethodConfig.load_credentials`
             method
@@ -1026,6 +1059,8 @@ def load_annotations(
     results = dataset.load_annotation_results(anno_key, **kwargs)
     annotations = results.backend.download_annotations(results)
     label_schema = results.config.label_schema
+    is_clips_view = results._is_clips
+    is_frames_view = results._is_frames
 
     unexpected_annos = defaultdict(dict)
 
@@ -1050,6 +1085,9 @@ def load_annotations(
         elif dest_field is not None:
             label_field = dest_field.get(label_field, label_field)
 
+        if is_frames_view:
+            label_field, _ = dataset._handle_frame_field(label_field)
+
         if expected_type and expected_type not in anno_dict:
             anno_dict[expected_type] = {}
 
@@ -1063,6 +1101,7 @@ def load_annotations(
                         results,
                         label_field,
                         label_info=label_info,
+                        progress=progress,
                     )
                 else:
                     _merge_labels(
@@ -1074,6 +1113,7 @@ def load_annotations(
                         label_info=label_info,
                         global_attrs=global_attrs,
                         class_attrs=class_attrs,
+                        progress=progress,
                     )
             else:
                 # Unexpected labels
@@ -1081,7 +1121,11 @@ def load_annotations(
                     new_field = None
                 elif unexpected == "prompt":
                     new_field = _prompt_field(
-                        dataset, anno_type, label_field, label_schema
+                        dataset,
+                        anno_type,
+                        label_field,
+                        label_schema,
+                        is_frames_view,
                     )
                 elif unexpected == "return":
                     new_field = None
@@ -1102,10 +1146,21 @@ def load_annotations(
                     )
 
                     if anno_type == "scalar":
-                        _merge_scalars(dataset, annos, results, new_field)
+                        _merge_scalars(
+                            dataset,
+                            annos,
+                            results,
+                            new_field,
+                            progress=progress,
+                        )
                     else:
                         _merge_labels(
-                            dataset, annos, results, new_field, anno_type
+                            dataset,
+                            annos,
+                            results,
+                            new_field,
+                            anno_type,
+                            progress=progress,
                         )
                 else:
                     if label_field:
@@ -1119,7 +1174,15 @@ def load_annotations(
                         logger.info("Skipping labels of type '%s'", anno_type)
 
                     if unexpected == "return":
-                        unexpected_annos[label_field][anno_type] = annos
+                        if is_clips_view:
+                            _annos = results._to_sample_annos(annos)
+                            unexpected_annos[label_field][anno_type] = _annos
+                        elif is_frames_view:
+                            _label_field = dataset._FRAMES_PREFIX + label_field
+                            _annos = results._to_sample_annos(annos)
+                            unexpected_annos[_label_field][anno_type] = _annos
+                        else:
+                            unexpected_annos[label_field][anno_type] = annos
 
     results.backend.save_run_results(dataset, anno_key, results)
 
@@ -1167,7 +1230,12 @@ def _get_writeable_attributes(attributes):
     return [k for k, v in attributes.items() if not v.get("read_only", False)]
 
 
-def _prompt_field(dataset, label_type, label_field, label_schema):
+def _prompt_field(
+    dataset, label_type, label_field, label_schema, is_frames_view
+):
+    if label_field and is_frames_view:
+        label_field = dataset._FRAMES_PREFIX + label_field
+
     if label_field:
         new_field = input(
             "Found unexpected labels of type '%s' when loading annotations "
@@ -1188,35 +1256,31 @@ def _prompt_field(dataset, label_type, label_field, label_schema):
     if label_type != "scalar":
         fo_label_type = _LABEL_TYPES_MAP[label_type]
 
-    if label_field is not None:
-        _, is_frame_field = dataset._handle_frame_field(label_field)
+    if label_field:
+        is_frame_field = dataset._is_frame_field(label_field)
     else:
-        _, is_frame_field = dataset._handle_frame_field(new_field)
-
-    if is_frame_field:
-        schema = dataset.get_frame_field_schema()
-    else:
-        schema = dataset.get_field_schema()
+        is_frame_field = dataset._is_frame_field(new_field)
 
     while True:
-        is_good_field = new_field not in label_schema
+        if is_frame_field and not dataset._is_frame_field(new_field):
+            new_field = dataset._FRAMES_PREFIX + new_field
+
+        if is_frames_view:
+            _new_field, _ = dataset._handle_frame_field(new_field)
+        else:
+            _new_field = new_field
+
+        is_good_field = _new_field not in label_schema
 
         if is_good_field:
-            if is_frame_field:
-                if not dataset._is_frame_field(new_field):
-                    new_field = dataset._FRAMES_PREFIX + new_field
-
-                field, _ = dataset._handle_frame_field(new_field)
-            else:
-                field = new_field
-
-            if field not in schema:
+            _field = dataset.get_field(new_field)
+            if _field is None:
                 break  # new field
 
             try:
-                field_type = schema[field].document_type
+                field_type = _field.document_type
             except:
-                field_type = type(schema[field])
+                field_type = type(_field)
 
             if label_type == "scalar":
                 # As long as it is not an embedded document field, assume the
@@ -1248,23 +1312,43 @@ def _prompt_field(dataset, label_type, label_field, label_schema):
         if not new_field:
             break
 
+    if is_frames_view:
+        new_field, _ = dataset._handle_frame_field(new_field)
+
     return new_field
 
 
-def _merge_scalars(dataset, anno_dict, results, label_field, label_info=None):
+def _merge_scalars(
+    dataset, anno_dict, results, label_field, label_info=None, progress=None
+):
     if label_info is None:
         label_info = {}
 
     allow_additions = label_info.get("allow_additions", True)
     allow_deletions = label_info.get("allow_deletions", True)
 
-    is_frame_field = dataset._is_frame_field(label_field)
-
-    # Retrieve a view that contains all samples involved in the annotation run
     id_map = results.id_map.get(label_field, {})
     uploaded_ids = set(k for k, v in id_map.items() if v is not None)
     sample_ids = list(uploaded_ids | set(anno_dict.keys()))
 
+    is_clips_view = results._is_clips
+    if is_clips_view:
+        _sample_ids_to_clip_ids = results._sample_ids_to_clip_ids()
+
+        # Map clip IDs to sample IDs
+        sample_ids = results._to_sample_ids(sample_ids)
+
+    is_frames_view = results._is_frames
+    if is_frames_view:
+        _label_field = dataset._FRAMES_PREFIX + label_field
+
+        # Map frame IDs to sample IDs
+        sample_ids = results._to_sample_ids(sample_ids)
+    else:
+        _label_field = label_field
+
+    is_frame_field = dataset._is_frame_field(_label_field)
+
     if dataset.media_type == fomm.GROUP:
         view = dataset.select_group_slices(_allow_mixed=True)
     else:
@@ -1273,20 +1357,32 @@ def _merge_scalars(dataset, anno_dict, results, label_field, label_info=None):
     view = dataset.select(sample_ids)
 
     if is_frame_field:
-        field, _ = view._handle_frame_field(label_field)
+        field, _ = view._handle_frame_field(_label_field)
         if view.has_frame_field(field):
-            view = view.select_fields(label_field)
+            view = view.select_fields(_label_field)
     else:
-        field = label_field
+        field = _label_field
         if view.has_sample_field(field):
-            view = view.select_fields(label_field)
+            view = view.select_fields(_label_field)
 
     num_additions = 0
     num_deletions = 0
 
-    logger.info("Loading scalars for field '%s'...", label_field)
-    for sample in view.iter_samples(progress=True):
-        sample_annos = anno_dict.get(sample.id, None)
+    logger.info("Loading scalars for field '%s'...", _label_field)
+    for sample in view.iter_samples(progress=progress, autosave=True):
+        if is_clips_view:
+            _clip_ids = _sample_ids_to_clip_ids.get(sample.id, None)
+            if _clip_ids:
+                sample_annos = {}
+                for _clip_id, _annos in anno_dict.items():
+                    if _clip_id in _clip_ids:
+                        sample_annos.update(_annos)
+            else:
+                sample_annos = None
+        elif is_frames_view:
+            sample_annos = anno_dict
+        else:
+            sample_annos = anno_dict.get(sample.id, None)
 
         if is_frame_field:
             images = sample.frames.values()
@@ -1322,8 +1418,6 @@ def _merge_scalars(dataset, anno_dict, results, label_field, label_info=None):
                     # Edit value
                     image[field] = new_value
 
-        sample.save()
-
     if num_additions > 0 and not allow_additions:
         logger.warning(
             "Ignored %d added scalars in field '%s' because "
@@ -1350,6 +1444,7 @@ def _merge_labels(
     label_info=None,
     global_attrs=None,
     class_attrs=None,
+    progress=None,
 ):
     if label_info is None:
         label_info = {}
@@ -1362,18 +1457,41 @@ def _merge_labels(
     allow_index_edits = label_info.get("allow_index_edits", True)
     allow_spatial_edits = label_info.get("allow_spatial_edits", True)
 
+    sample_ids = list(anno_dict.keys())
+
+    is_clips_view = results._is_clips
+    if is_clips_view:
+        _sample_ids_to_clip_ids = results._sample_ids_to_clip_ids()
+        _frame_ids_to_clip_ids = {}  # will populate below
+
+        # Map clip IDs to sample IDs
+        sample_ids = results._to_sample_ids(sample_ids)
+
+    is_frames_view = results._is_frames
+    if is_frames_view:
+        _label_field = dataset._FRAMES_PREFIX + label_field
+
+        # Map frame IDs to sample IDs
+        sample_ids = results._to_sample_ids(sample_ids)
+    else:
+        _label_field = label_field
+
     fo_label_type = _LABEL_TYPES_MAP[label_type]
-    if issubclass(fo_label_type, fol._LABEL_LIST_FIELDS):
+    if issubclass(fo_label_type, fol._HasLabelList):
         is_list = True
         list_field = fo_label_type._LABEL_LIST_FIELD
     else:
         is_list = False
 
-    _ensure_label_field(dataset, label_field, fo_label_type)
+    _ensure_label_field(dataset, _label_field, fo_label_type)
 
-    is_frame_field = dataset._is_frame_field(label_field)
+    is_frame_field = dataset._is_frame_field(_label_field)
 
-    if is_frame_field and label_type in _TRACKABLE_TYPES:
+    if (
+        is_frame_field
+        and not is_frames_view
+        and label_type in _TRACKABLE_TYPES
+    ):
         if not existing_field:
             # Always include keyframe info when importing new video tracks
             only_keyframes = True
@@ -1382,7 +1500,7 @@ def _merge_labels(
 
     id_map = results.id_map.get(label_field, {})
 
-    if is_frame_field:
+    if is_frame_field and not is_frames_view:
         field, _ = dataset._handle_frame_field(label_field)
         added_id_map = defaultdict(lambda: defaultdict(list))
     else:
@@ -1392,7 +1510,11 @@ def _merge_labels(
     # Record existing label IDs
     existing_ids = set()
     for sample_id, sample_labels in id_map.items():
-        if is_frame_field:
+        if is_clips_view:
+            for frame_id in sample_labels.keys():
+                _frame_ids_to_clip_ids[frame_id] = sample_id
+
+        if is_frame_field and not is_frames_view:
             for frame_id, frame_labels in sample_labels.items():
                 for label_id in _to_list(frame_labels):
                     existing_ids.add((sample_id, frame_id, label_id))
@@ -1404,7 +1526,11 @@ def _merge_labels(
     anno_ids = set()
     anno_id_counts = defaultdict(int)
     for sample_id, sample_labels in anno_dict.items():
-        if is_frame_field:
+        if is_clips_view:
+            for frame_id in sample_labels.keys():
+                _frame_ids_to_clip_ids[frame_id] = sample_id
+
+        if is_frame_field and not is_frames_view:
             for frame_id, frame_labels in sample_labels.items():
                 for label_id in frame_labels.keys():
                     anno_ids.add((sample_id, frame_id, label_id))
@@ -1420,7 +1546,7 @@ def _merge_labels(
     merge_ids = anno_ids - new_ids
 
     #
-    # Manually prevent duplicate label IDs from being imported by regnerating
+    # Manually prevent duplicate label IDs from being imported by regenerating
     # any newly added IDs that are duplicates.
     #
     # Duplicate IDs can happen when copying annotations or splitting/merging
@@ -1429,7 +1555,7 @@ def _merge_labels(
     #
     dup_ids = set(_id for _id, count in anno_id_counts.items() if count > 1)
     if dup_ids:
-        if is_frame_field:
+        if is_frame_field and not is_frames_view:
             for sample_id, frame_id, label_id in list(new_ids):
                 if label_id in dup_ids:
                     # Regenerate duplicate label ID
@@ -1452,12 +1578,12 @@ def _merge_labels(
                     new_ids.discard((sample_id, label_id))
                     new_ids.add((sample_id, new_label_id))
 
-    logger.info("Loading labels for field '%s'...", label_field)
+    logger.info("Loading labels for field '%s'...", _label_field)
 
     # Delete labels that were deleted in the annotation task
     if delete_ids and allow_deletions:
         _del_ids = [key[-1] for key in delete_ids]
-        dataset.delete_labels(ids=_del_ids, fields=label_field)
+        dataset.delete_labels(ids=_del_ids, fields=_label_field)
 
     if dataset.media_type == fomm.GROUP:
         view = dataset.select_group_slices(_allow_mixed=True)
@@ -1465,11 +1591,23 @@ def _merge_labels(
         view = dataset.view()
 
     # Add/merge labels from the annotation task
-    sample_ids = list(anno_dict.keys())
-    view = view.select(sample_ids).select_fields(label_field)
-    for sample in view.iter_samples(progress=True):
+    view = view.select(sample_ids).select_fields(_label_field)
+    for sample in view.iter_samples(progress=progress, autosave=True):
         sample_id = sample.id
-        sample_annos = anno_dict[sample_id]
+
+        if is_clips_view:
+            _clip_ids = _sample_ids_to_clip_ids.get(sample_id, None)
+            if _clip_ids:
+                sample_annos = {}
+                for _clip_id, _annos in anno_dict.items():
+                    if _clip_id in _clip_ids:
+                        sample_annos.update(_annos)
+            else:
+                sample_annos = None
+        elif is_frames_view:
+            sample_annos = anno_dict
+        else:
+            sample_annos = anno_dict[sample_id]
 
         if is_frame_field:
             images = sample.frames.values()
@@ -1485,6 +1623,9 @@ def _merge_labels(
             else:
                 image_annos = sample_annos
 
+            if is_clips_view:
+                clip_id = _frame_ids_to_clip_ids.get(frame_id, None)
+
             image_label = image[field]
 
             if image_label is None and allow_additions:
@@ -1495,7 +1636,11 @@ def _merge_labels(
                         **{list_field: list(image_annos.values())}
                     )
 
-                    if is_frame_field:
+                    if is_clips_view:
+                        added_id_map[clip_id][frame_id].extend(label_ids)
+                    elif is_frames_view:
+                        added_id_map[frame_id].extend(label_ids)
+                    elif is_frame_field:
                         added_id_map[sample_id][frame_id].extend(label_ids)
                     else:
                         added_id_map[sample_id].extend(label_ids)
@@ -1503,7 +1648,11 @@ def _merge_labels(
                     label_id, anno_label = next(iter(image_annos.items()))
                     image[field] = anno_label
 
-                    if is_frame_field:
+                    if is_clips_view:
+                        added_id_map[clip_id][frame_id] = label_id
+                    elif is_frames_view:
+                        added_id_map[frame_id] = label_id
+                    elif is_frame_field:
                         added_id_map[sample_id][frame_id] = label_id
                     else:
                         added_id_map[sample_id] = label_id
@@ -1516,7 +1665,11 @@ def _merge_labels(
                 # Merge labels that existed before and after annotation
                 for label in labels:
                     label_id = label.id
-                    if is_frame_field:
+                    if is_clips_view:
+                        key = (clip_id, frame_id, label_id)
+                    elif is_frames_view:
+                        key = (frame_id, label_id)
+                    elif is_frame_field:
                         key = (sample_id, frame_id, label_id)
                     else:
                         key = (sample_id, label_id)
@@ -1538,7 +1691,11 @@ def _merge_labels(
                 # Add new labels to label list fields
                 if is_list and allow_additions:
                     for label_id, anno_label in image_annos.items():
-                        if is_frame_field:
+                        if is_clips_view:
+                            key = (clip_id, frame_id, label_id)
+                        elif is_frames_view:
+                            key = (frame_id, label_id)
+                        elif is_frame_field:
                             key = (sample_id, frame_id, label_id)
                         else:
                             key = (sample_id, label_id)
@@ -1547,20 +1704,21 @@ def _merge_labels(
                             continue
 
                         labels.append(anno_label)
-
-                        if is_frame_field:
+                        if is_clips_view:
+                            added_id_map[clip_id][frame_id].append(label_id)
+                        elif is_frames_view:
+                            added_id_map[frame_id].append(label_id)
+                        elif is_frame_field:
                             added_id_map[sample_id][frame_id].append(label_id)
                         else:
                             added_id_map[sample_id].append(label_id)
 
-        sample.save()
-
     if new_ids and not allow_additions:
         logger.warning(
             "Ignored %d added labels in field '%s' because "
             "`allow_additions=False`",
             len(new_ids),
-            label_field,
+            _label_field,
         )
 
     if delete_ids and not allow_deletions:
@@ -1568,7 +1726,7 @@ def _merge_labels(
             "Ignored %d deleted labels in field '%s' because "
             "`allow_deletions=False`",
             len(delete_ids),
-            label_field,
+            _label_field,
         )
 
     # Record newly added IDs so that re-imports of this run will be properly
@@ -1888,6 +2046,11 @@ class AnnotationBackend(foa.AnnotationMethod):
             "subclass must implement supported_attr_types"
         )
 
+    @property
+    def supports_clips_views(self):
+        """Whether this backend supports annotating clips views."""
+        return False
+
     @property
     def supports_keyframes(self):
         """Whether this backend supports uploading only keyframes when editing
@@ -2148,6 +2311,28 @@ class AnnotationResults(foa.AnnotationResults):
         """Deletes all information for this run from the annotation backend."""
         raise NotImplementedError("subclass must implement cleanup()")
 
+    @property
+    def _is_clips(self):
+        """Whether this annotation run was performed on a clips view."""
+        return "_clips" in self.id_map
+
+    @property
+    def _is_frames(self):
+        """Whether this annotation run was perfromed on a frames view."""
+        return "_frames" in self.id_map
+
+    def _finalize_id_map(self):
+        """Internal method that is (only) called prior to saving annotation
+        results for the first time.
+        """
+        if self.samples._is_clips:
+            id_map = dict(zip(*self.samples.values(["id", "sample_id"])))
+            self.id_map["_clips"] = id_map
+
+        if self.samples._is_frames:
+            id_map = dict(zip(*self.samples.values(["id", "sample_id"])))
+            self.id_map["_frames"] = id_map
+
     def _update_id_map(self, label_field, new_id_map):
         """Adds the given label IDs into this object's :attr:`id_map`.
 
@@ -2195,6 +2380,49 @@ class AnnotationResults(foa.AnnotationResults):
                     id_map.get(sample_id, None), label_ids
                 )
 
+    def _to_sample_ids(self, ids):
+        if self._is_clips:
+            id_map = self.id_map["_clips"]
+            sample_ids = set(id_map.get(id, None) for id in ids)
+            sample_ids.discard(None)
+            return list(sample_ids)
+
+        if self._is_frames:
+            id_map = self.id_map["_frames"]
+            sample_ids = set(id_map.get(id, None) for id in ids)
+            sample_ids.discard(None)
+            return list(sample_ids)
+
+        return ids
+
+    def _sample_ids_to_clip_ids(self):
+        id_map = defaultdict(set)
+        for _clip_id, _sample_id in self.id_map["_clips"].items():
+            id_map[_sample_id].add(_clip_id)
+
+        return id_map
+
+    def _to_sample_annos(self, annos):
+        if self._is_clips:
+            id_map = self.id_map["_clips"]
+            _annos = defaultdict(dict)
+            for clip_id, clip_annos in annos.items():
+                sample_id = id_map[clip_id]
+                _annos[sample_id].update(clip_annos)
+
+            return dict(_annos)
+
+        if self._is_frames:
+            id_map = self.id_map["_frames"]
+            _annos = defaultdict(dict)
+            for frame_id, label_ids in annos.items():
+                sample_id = id_map[frame_id]
+                _annos[sample_id][frame_id] = label_ids
+
+            return dict(_annos)
+
+        return annos
+
     def _format_label_ids(self, ids1, ids2):
         return _unwrap(_to_list(ids1) + _to_list(ids2))
 
@@ -2230,14 +2458,15 @@ class AnnotationAPI(object):
         pass
 
     def _prompt_username_password(self, backend, username=None, password=None):
-        prefix = "FIFTYONE_%s_" % backend.upper()
-        logger.info(
-            "Please enter your login credentials.\nYou can avoid this in the "
-            "future by setting your `%sUSERNAME` and `%sPASSWORD` environment "
-            "variables",
-            prefix,
-            prefix,
-        )
+        if backend:
+            prefix = "FIFTYONE_%s_" % backend.upper()
+            logger.info(
+                "Please enter your login credentials.\nYou can avoid this in "
+                "the future by setting your `%sUSERNAME` and `%sPASSWORD` "
+                "environment variables",
+                prefix,
+                prefix,
+            )
 
         if username is None:
             username = input("Username: ")
@@ -2248,12 +2477,13 @@ class AnnotationAPI(object):
         return username, password
 
     def _prompt_api_key(self, backend):
-        prefix = "FIFTYONE_%s_" % backend.upper()
-        logger.info(
-            "Please enter your API key.\nYou can avoid this in the future by "
-            "setting your `%sKEY` environment variable",
-            prefix,
-        )
+        if backend:
+            prefix = "FIFTYONE_%s_" % backend.upper()
+            logger.info(
+                "Please enter your API key.\nYou can avoid this in the future "
+                "by setting your `%sAPI_KEY` environment variable",
+                prefix,
+            )
 
         return getpass.getpass(prompt="API key: ")
 
@@ -2287,7 +2517,13 @@ class DrawConfig(etaa.AnnotationConfig):
 
 
 def draw_labeled_images(
-    samples, output_dir, rel_dir=None, label_fields=None, config=None, **kwargs
+    samples,
+    output_dir,
+    rel_dir=None,
+    label_fields=None,
+    config=None,
+    progress=None,
+    **kwargs,
 ):
     """Renders annotated versions of the images in the collection with the
     specified label data overlaid to the given directory.
@@ -2307,11 +2543,14 @@ def draw_labeled_images(
             This argument allows for populating nested subdirectories in
             ``output_dir`` that match the shape of the input paths. The path is
             converted to an absolute path (if necessary) via
-            :func:`fiftyone.core.utils.normalize_path`
+            :func:`fiftyone.core.storage.normalize_path`
         label_fields (None): a label field or list of label fields to render.
-            If omitted, all compatiable fields are rendered
+            If omitted, all compatible fields are rendered
         config (None): an optional :class:`DrawConfig` configuring how to draw
             the labels
+        progress (None): whether to render a progress bar (True/False), use the
+            default value ``fiftyone.config.show_progress_bars`` (None), or a
+            progress callback function to invoke instead
         **kwargs: optional keyword arguments specifying parameters of the
             default :class:`DrawConfig` to override
 
@@ -2328,7 +2567,7 @@ def draw_labeled_images(
     output_ext = fo.config.default_image_ext
 
     outpaths = []
-    for sample in samples.iter_samples(progress=True):
+    for sample in samples.iter_samples(progress=progress):
         outpath = filename_maker.get_output_path(
             sample.filepath, output_ext=output_ext
         )
@@ -2350,7 +2589,7 @@ def draw_labeled_image(
         sample: a :class:`fiftyone.core.sample.Sample`
         outpath: the path to write the annotated image
         label_fields (None): a label field or list of label fields to render.
-            If omitted, all compatiable fields are rendered
+            If omitted, all compatible fields are rendered
         config (None): an optional :class:`DrawConfig` configuring how to draw
             the labels
         **kwargs: optional keyword arguments specifying parameters of the
@@ -2359,16 +2598,22 @@ def draw_labeled_image(
     config = _parse_draw_config(config, kwargs)
 
     fov.validate_image_sample(sample)
-    img = etai.read(sample.filepath)
+    img = foui.read(sample.filepath)
 
     image_labels = _to_image_labels(sample, label_fields=label_fields)
 
     anno_img = etaa.annotate_image(img, image_labels, annotation_config=config)
-    etai.write(anno_img, outpath)
+    foui.write(anno_img, outpath)
 
 
 def draw_labeled_videos(
-    samples, output_dir, rel_dir=None, label_fields=None, config=None, **kwargs
+    samples,
+    output_dir,
+    rel_dir=None,
+    label_fields=None,
+    config=None,
+    progress=None,
+    **kwargs,
 ):
     """Renders annotated versions of the videos in the collection with the
     specified label data overlaid to the given directory.
@@ -2388,11 +2633,14 @@ def draw_labeled_videos(
             This argument allows for populating nested subdirectories in
             ``output_dir`` that match the shape of the input paths. The path is
             converted to an absolute path (if necessary) via
-            :func:`fiftyone.core.utils.normalize_path`
+            :func:`fiftyone.core.storage.normalize_path`
         label_fields (None): a label field or list of label fields to render.
-            If omitted, all compatiable fields are rendered
+            If omitted, all compatible fields are rendered
         config (None): an optional :class:`DrawConfig` configuring how to draw
             the labels
+        progress (None): whether to render a progress bar (True/False), use the
+            default value ``fiftyone.config.show_progress_bars`` (None), or a
+            progress callback function to invoke instead
         **kwargs: optional keyword arguments specifying parameters of the
             default :class:`DrawConfig` to override
 
@@ -2412,7 +2660,7 @@ def draw_labeled_videos(
     num_videos = len(samples)
 
     outpaths = []
-    for idx, sample in enumerate(samples, 1):
+    for idx, sample in enumerate(samples.iter_samples(progress=progress), 1):
         if is_clips:
             logger.info("Drawing labels for clip %d/%d", idx, num_videos)
             base, ext = os.path.splitext(sample.filepath)
@@ -2443,7 +2691,7 @@ def draw_labeled_video(
         support (None): an optional ``[first, last]`` range of frames to
             render
         label_fields (None): a label field or list of label fields to render.
-            If omitted, all compatiable fields are rendered
+            If omitted, all compatible fields are rendered
         config (None): an optional :class:`DrawConfig` configuring how to draw
             the labels
         **kwargs: optional keyword arguments specifying parameters of the
@@ -2538,11 +2786,7 @@ def _to_video_labels(sample, label_fields=None):
 
 def _get_sample_labels(sample, label_fields):
     if label_fields is None:
-        return {
-            name: value
-            for name, value in sample.iter_fields()
-            if isinstance(value, fol.Label)
-        }
+        return dict(sample._iter_label_fields())
 
     if label_fields:
         return {name: sample[name] for name in label_fields}
@@ -2557,11 +2801,7 @@ def _get_frame_labels(sample, frame_label_fields):
     frames = {}
     for frame_number, frame in sample.frames.items():
         if frame_label_fields is None:
-            frames[frame_number] = {
-                name: value
-                for name, value in frame.iter_fields()
-                if isinstance(value, fol.Label)
-            }
+            frames[frame_number] = dict(frame._iter_label_fields())
         else:
             frames[frame_number] = {
                 name: frame[name] for name in frame_label_fields
