@@ -2,7 +2,7 @@
 Utilities for working with datasets in
 `COCO format <https://cocodataset.org/#format-data>`_.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -11,7 +11,6 @@ import csv
 from datetime import datetime
 from itertools import groupby
 import logging
-import multiprocessing
 import multiprocessing.dummy
 import os
 import random
@@ -29,6 +28,7 @@ import eta.core.web as etaw
 import fiftyone.core.fields as fof
 import fiftyone.core.labels as fol
 import fiftyone.core.metadata as fom
+import fiftyone.core.storage as fos
 import fiftyone.core.utils as fou
 import fiftyone.utils.data as foud
 import fiftyone.utils.eta as foue
@@ -45,7 +45,7 @@ def add_coco_labels(
     sample_collection,
     label_field,
     labels_or_path,
-    classes,
+    categories,
     label_type="detections",
     coco_id_field=None,
     include_annotation_id=False,
@@ -68,7 +68,7 @@ def add_coco_labels(
             {
                 "id": 1,
                 "image_id": 1,
-                "category_id": 2,
+                "category_id": 1,
                 "bbox": [260, 177, 231, 199],
 
                 # optional
@@ -88,7 +88,7 @@ def add_coco_labels(
             {
                 "id": 1,
                 "image_id": 1,
-                "category_id": 2,
+                "category_id": 1,
                 "bbox": [260, 177, 231, 199],
                 "segmentation": [...],
 
@@ -109,7 +109,7 @@ def add_coco_labels(
             {
                 "id": 1,
                 "image_id": 1,
-                "category_id": 2,
+                "category_id": 1,
                 "keypoints": [224, 226, 2, ...],
                 "num_keypoints": 10,
 
@@ -129,7 +129,14 @@ def add_coco_labels(
             will be created if necessary
         labels_or_path: a list of COCO annotations or the path to a JSON file
             containing such data on disk
-        classes: the list of class label strings
+        categories: can be any of the following:
+
+            -   a list of category dicts in the format of
+                :meth:`parse_coco_categories` specifying the classes and their
+                category IDs
+            -   a dict mapping class IDs to class labels
+            -   a list of class labels whose 1-based ordering is assumed to
+                correspond to the category IDs in the provided COCO labels
         label_type ("detections"): the type of labels to load. Supported values
             are ``("detections", "segmentations", "keypoints")``
         coco_id_field (None): this parameter determines how to map the
@@ -194,6 +201,15 @@ def add_coco_labels(
     view.compute_metadata()
     widths, heights = view.values(["metadata.width", "metadata.height"])
 
+    if isinstance(categories, dict):
+        classes_map = categories
+    elif not categories:
+        classes_map = {}
+    elif isinstance(categories[0], dict):
+        classes_map = {c["id"]: c["name"] for c in categories}
+    else:
+        classes_map = {i: label for i, label in enumerate(categories, 1)}
+
     labels = []
     for _coco_objects, width, height in zip(coco_objects, widths, heights):
         frame_size = (width, height)
@@ -202,7 +218,7 @@ def add_coco_labels(
             _labels = _coco_objects_to_detections(
                 _coco_objects,
                 frame_size,
-                classes,
+                classes_map,
                 None,
                 False,
                 include_annotation_id,
@@ -212,7 +228,7 @@ def add_coco_labels(
                 _labels = _coco_objects_to_polylines(
                     _coco_objects,
                     frame_size,
-                    classes,
+                    classes_map,
                     None,
                     tolerance,
                     include_annotation_id,
@@ -221,7 +237,7 @@ def add_coco_labels(
                 _labels = _coco_objects_to_detections(
                     _coco_objects,
                     frame_size,
-                    classes,
+                    classes_map,
                     None,
                     True,
                     include_annotation_id,
@@ -230,7 +246,7 @@ def add_coco_labels(
             _labels = _coco_objects_to_keypoints(
                 _coco_objects,
                 frame_size,
-                classes,
+                classes_map,
                 None,
                 include_annotation_id,
             )
@@ -408,7 +424,8 @@ class COCODetectionDatasetImporter(
 
         self._label_types = _label_types
         self._info = None
-        self._classes = None
+        self._classes_map = None
+        self._class_ids = None
         self._license_map = None
         self._supercategory_map = None
         self._image_paths_map = None
@@ -450,16 +467,16 @@ class COCODetectionDatasetImporter(
             coco_objects = self._annotations.get(image_id, [])
             frame_size = (width, height)
 
-            if self.classes is not None and self.only_matching:
+            if self.only_matching and self._class_ids is not None:
                 coco_objects = _get_matching_objects(
-                    coco_objects, self.classes, self._classes
+                    coco_objects, self._class_ids
                 )
 
             if "detections" in self._label_types:
                 detections = _coco_objects_to_detections(
                     coco_objects,
                     frame_size,
-                    self._classes,
+                    self._classes_map,
                     self._supercategory_map,
                     False,  # no segmentations
                     self.include_annotation_id,
@@ -472,7 +489,7 @@ class COCODetectionDatasetImporter(
                     segmentations = _coco_objects_to_polylines(
                         coco_objects,
                         frame_size,
-                        self._classes,
+                        self._classes_map,
                         self._supercategory_map,
                         self.tolerance,
                         self.include_annotation_id,
@@ -481,7 +498,7 @@ class COCODetectionDatasetImporter(
                     segmentations = _coco_objects_to_detections(
                         coco_objects,
                         frame_size,
-                        self._classes,
+                        self._classes_map,
                         self._supercategory_map,
                         True,  # load segmentations
                         self.include_annotation_id,
@@ -494,7 +511,7 @@ class COCODetectionDatasetImporter(
                 keypoints = _coco_objects_to_keypoints(
                     coco_objects,
                     frame_size,
-                    self._classes,
+                    self._classes_map,
                     self._supercategory_map,
                     self.include_annotation_id,
                 )
@@ -548,7 +565,7 @@ class COCODetectionDatasetImporter(
         if self.labels_path is not None and os.path.isfile(self.labels_path):
             (
                 info,
-                classes,
+                classes_map,
                 supercategory_map,
                 images,
                 annotations,
@@ -556,11 +573,11 @@ class COCODetectionDatasetImporter(
                 self.labels_path, extra_attrs=self.extra_attrs
             )
 
-            if classes is not None:
-                info["classes"] = classes
+            if classes_map is not None:
+                info["classes"] = _to_classes(classes_map)
 
             image_ids = _get_matching_image_ids(
-                classes,
+                classes_map,
                 images,
                 annotations,
                 image_ids=self.image_ids,
@@ -571,18 +588,18 @@ class COCODetectionDatasetImporter(
             )
 
             filenames = [
-                fou.normpath(images[_id]["file_name"]) for _id in image_ids
+                fos.normpath(images[_id]["file_name"]) for _id in image_ids
             ]
 
             _image_ids = set(image_ids)
             image_dicts_map = {
-                fou.normpath(i["file_name"]): i
+                fos.normpath(i["file_name"]): i
                 for _id, i in images.items()
                 if _id in _image_ids
             }
         else:
             info = {}
-            classes = None
+            classes_map = None
             supercategory_map = None
             image_dicts_map = {}
             annotations = None
@@ -596,8 +613,14 @@ class COCODetectionDatasetImporter(
         else:
             license_map = None
 
+        if self.only_matching and self.classes is not None:
+            class_ids = _get_class_ids(self.classes, classes_map)
+        else:
+            class_ids = None
+
         self._info = info
-        self._classes = classes
+        self._classes_map = classes_map
+        self._class_ids = class_ids
         self._license_map = license_map
         self._supercategory_map = supercategory_map
         self._image_paths_map = image_paths_map
@@ -614,9 +637,6 @@ class COCODetectionDatasetExporter(
 ):
     """Exporter that writes COCO detection datasets to disk.
 
-    This class currently only supports exporting detections and instance
-    segmentations.
-
     See :ref:`this page <COCODetectionDataset-export>` for format details.
 
     Args:
@@ -672,13 +692,16 @@ class COCODetectionDatasetExporter(
             generate an output path for each exported image. This argument
             allows for populating nested subdirectories that match the shape of
             the input paths. The path is converted to an absolute path (if
-            necessary) via :func:`fiftyone.core.utils.normalize_path`
+            necessary) via :func:`fiftyone.core.storage.normalize_path`
         abs_paths (False): whether to store absolute paths to the images in the
             exported labels
         image_format (None): the image format to use when writing in-memory
             images to disk. By default, ``fiftyone.config.default_image_ext``
             is used
         classes (None): the list of possible class labels
+        categories (None): a list of category dicts in the format of
+            :meth:`parse_coco_categories` specifying the classes and their
+            category IDs
         info (None): a dict of info as returned by
             :meth:`load_coco_detection_annotations` to include in the exported
             JSON. If not provided, this info will be extracted when
@@ -689,6 +712,8 @@ class COCODetectionDatasetExporter(
             -   ``True``: export all extra attributes found
             -   ``False``: do not export extra attributes
             -   a name or list of names of specific attributes to export
+        coco_id (None): the name of a sample field containing the COCO IDs of
+            each image
         annotation_id (None): the name of a label field containing the COCO
             annotation ID of each label
         iscrowd ("iscrowd"): the name of a detection attribute that indicates
@@ -711,8 +736,10 @@ class COCODetectionDatasetExporter(
         abs_paths=False,
         image_format=None,
         classes=None,
+        categories=None,
         info=None,
         extra_attrs=True,
+        coco_id=None,
         annotation_id=None,
         iscrowd="iscrowd",
         num_decimals=None,
@@ -740,19 +767,22 @@ class COCODetectionDatasetExporter(
         self.abs_paths = abs_paths
         self.image_format = image_format
         self.classes = classes
+        self.categories = categories
         self.info = info
         self.extra_attrs = extra_attrs
+        self.coco_id = coco_id
         self.annotation_id = annotation_id
         self.iscrowd = iscrowd
         self.num_decimals = num_decimals
         self.tolerance = tolerance
 
         self._image_id = None
+        self._image_id_map = None
         self._anno_id = None
         self._images = None
         self._annotations = None
         self._classes = None
-        self._dynamic_classes = classes is None
+        self._dynamic_classes = None
         self._labels_map_rev = None
         self._has_labels = None
         self._media_exporter = None
@@ -786,6 +816,11 @@ class COCODetectionDatasetExporter(
         if self.info is None:
             self.info = sample_collection.info
 
+        if self.coco_id is not None:
+            self._image_id_map = dict(
+                zip(*sample_collection.values(["filepath", self.coco_id]))
+            )
+
     def export_sample(self, image_or_path, label, metadata=None):
         out_image_path, uuid = self._media_exporter.export(image_or_path)
 
@@ -797,12 +832,24 @@ class COCODetectionDatasetExporter(
         else:
             file_name = uuid
 
-        # @todo would be nice to support using existing COCO ID here
-        self._image_id += 1
+        if self._image_id_map is not None:
+            image_id = self._image_id_map.get(image_or_path, None)
+            if image_id is None:
+                msg = (
+                    "Ignoring sample with filepath '%s' that has no image ID"
+                    % image_or_path
+                )
+                warnings.warn(msg)
+                return
+
+            image_id = int(image_id)
+        else:
+            self._image_id += 1
+            image_id = self._image_id
 
         self._images.append(
             {
-                "id": self._image_id,
+                "id": image_id,
                 "file_name": file_name,
                 "height": metadata.height,
                 "width": metadata.width,
@@ -850,7 +897,7 @@ class COCODetectionDatasetExporter(
             obj = COCOObject.from_label(
                 label,
                 metadata,
-                image_id=self._image_id,
+                image_id=image_id,
                 category_id=category_id,
                 extra_attrs=self.extra_attrs,
                 id_attr=self.annotation_id,
@@ -866,38 +913,36 @@ class COCODetectionDatasetExporter(
 
     def close(self, *args):
         if self._dynamic_classes:
-            classes = sorted(self._classes)
-            labels_map_rev = _to_labels_map_rev(classes)
+            labels_map_rev = _to_labels_map_rev(sorted(self._classes))
             for anno in self._annotations:
                 anno["category_id"] = labels_map_rev[anno["category_id"]]
-        else:
-            classes = self.classes
+        elif self.categories is None:
+            labels_map_rev = _to_labels_map_rev(self.classes)
+
+        _info = self.info or {}
+        _date_created = datetime.now().replace(microsecond=0).isoformat()
 
-        date_created = datetime.now().replace(microsecond=0).isoformat()
         info = {
-            "year": self.info.get("year", ""),
-            "version": self.info.get("version", ""),
-            "description": self.info.get("year", "Exported from FiftyOne"),
-            "contributor": self.info.get("contributor", ""),
-            "url": self.info.get("url", "https://voxel51.com/fiftyone"),
-            "date_created": self.info.get("date_created", date_created),
+            "year": _info.get("year", ""),
+            "version": _info.get("version", ""),
+            "contributor": _info.get("contributor", ""),
+            "url": _info.get("url", "https://voxel51.com/fiftyone"),
+            "date_created": _info.get("date_created", _date_created),
         }
 
-        licenses = self.info.get("licenses", [])
+        licenses = _info.get("licenses", [])
 
-        supercategory_map = {
-            c["name"]: c.get("supercategory", None)
-            for c in self.info.get("categories", [])
-        }
-
-        categories = [
-            {
-                "id": i,
-                "name": l,
-                "supercategory": supercategory_map.get(l, None),
-            }
-            for i, l in enumerate(classes)
-        ]
+        if self.categories is not None:
+            categories = self.categories
+        else:
+            categories = [
+                {
+                    "id": i,
+                    "name": c,
+                    "supercategory": None,
+                }
+                for c, i in sorted(labels_map_rev.items(), key=lambda t: t[1])
+            ]
 
         labels = {
             "info": info,
@@ -914,10 +959,17 @@ class COCODetectionDatasetExporter(
         self._media_exporter.close()
 
     def _parse_classes(self):
-        if self._dynamic_classes:
+        if self.categories is not None:
+            self._labels_map_rev = _parse_categories(
+                self.categories, classes=self.classes
+            )
+            self._dynamic_classes = False
+        elif self.classes is None:
             self._classes = set()
+            self._dynamic_classes = True
         else:
             self._labels_map_rev = _to_labels_map_rev(self.classes)
+            self._dynamic_classes = False
 
 
 class COCOObject(object):
@@ -964,7 +1016,7 @@ class COCOObject(object):
     def to_polyline(
         self,
         frame_size,
-        classes=None,
+        classes_map=None,
         supercategory_map=None,
         tolerance=None,
         include_id=False,
@@ -974,7 +1026,7 @@ class COCOObject(object):
 
         Args:
             frame_size: the ``(width, height)`` of the image
-            classes (None): the list of classes
+            classes_map (None): a dict mapping class IDs to class labels
             supercategory_map (None): a dict mapping class names to category
                 dicts
             tolerance (None): a tolerance, in pixels, when generating
@@ -991,7 +1043,7 @@ class COCOObject(object):
             return None
 
         label, attributes = self._get_object_label_and_attributes(
-            classes, supercategory_map, include_id
+            classes_map, supercategory_map, include_id
         )
         attributes.update(self.attributes)
 
@@ -1011,7 +1063,7 @@ class COCOObject(object):
     def to_keypoints(
         self,
         frame_size,
-        classes=None,
+        classes_map=None,
         supercategory_map=None,
         include_id=False,
     ):
@@ -1020,7 +1072,7 @@ class COCOObject(object):
 
         Args:
             frame_size: the ``(width, height)`` of the image
-            classes (None): the list of classes
+            classes_map (None): a dict mapping class IDs to class labels
             supercategory_map (None): a dict mapping class names to category
                 dicts
             include_id (False): whether to include the COCO ID of the object as
@@ -1034,25 +1086,36 @@ class COCOObject(object):
             return None
 
         label, attributes = self._get_object_label_and_attributes(
-            classes, supercategory_map, include_id
+            classes_map, supercategory_map, include_id
         )
         attributes.update(self.attributes)
 
         width, height = frame_size
 
         points = []
+        visible = []
         for x, y, v in fou.iter_batches(self.keypoints, 3):
             if v == 0:
                 points.append((float("nan"), float("nan")))
             else:
                 points.append((x / width, y / height))
 
-        return fol.Keypoint(label=label, points=points, **attributes)
+            visible.append(v)
+        if "visible" in attributes:
+            logger.debug(
+                "Found a custom attribute named 'visible' which is a "
+                "reserved name. Ignoring the custom attribute"
+            )
+            attributes.pop("visible")
+
+        return fol.Keypoint(
+            label=label, points=points, visible=visible, **attributes
+        )
 
     def to_detection(
         self,
         frame_size,
-        classes=None,
+        classes_map=None,
         supercategory_map=None,
         load_segmentation=False,
         include_id=False,
@@ -1062,7 +1125,7 @@ class COCOObject(object):
 
         Args:
             frame_size: the ``(width, height)`` of the image
-            classes (None): the list of classes
+            classes_map (None): a dict mapping class IDs to class labels
             supercategory_map (None): a dict mapping class names to category
                 dicts
             load_segmentation (False): whether to load the segmentation mask
@@ -1078,7 +1141,7 @@ class COCOObject(object):
             return None
 
         label, attributes = self._get_object_label_and_attributes(
-            classes, supercategory_map, include_id
+            classes_map, supercategory_map, include_id
         )
         attributes.update(self.attributes)
 
@@ -1241,7 +1304,7 @@ class COCOObject(object):
             x, y, w, h = label.bounding_box
             bbox = [x * width, y * height, w * width, h * height]
 
-            if label.mask is not None:
+            if label.has_mask:
                 segmentation = _instance_to_coco_segmentation(
                     label, frame_size, iscrowd=iscrowd, tolerance=tolerance
                 )
@@ -1282,6 +1345,7 @@ class COCOObject(object):
         attributes.pop(id_attr, None)  # okay if `id_attr` is None
         attributes.pop(iscrowd, None)
         attributes.pop("area", None)
+        attributes.pop("visible", None)
 
         return cls(
             id=_id,
@@ -1303,10 +1367,10 @@ class COCOObject(object):
         return str(self.category_id)
 
     def _get_object_label_and_attributes(
-        self, classes, supercategory_map, include_id
+        self, classes_map, supercategory_map, include_id
     ):
-        if classes:
-            label = classes[self.category_id]
+        if classes_map:
+            label = classes_map[self.category_id]
         else:
             label = str(self.category_id)
 
@@ -1347,7 +1411,7 @@ def load_coco_detection_annotations(json_path, extra_attrs=True):
         a tuple of
 
         -   info: a dict of dataset info
-        -   classes: a list of classes
+        -   classes_map: a dict mapping class IDs to labels
         -   supercategory_map: a dict mapping class labels to category dicts
         -   images: a dict mapping image IDs to image dicts
         -   annotations: a dict mapping image IDs to list of
@@ -1374,9 +1438,9 @@ def _parse_coco_detection_annotations(d, extra_attrs=True):
 
     # Load classes
     if categories is not None:
-        classes, supercategory_map = parse_coco_categories(categories)
+        classes_map, supercategory_map = parse_coco_categories(categories)
     else:
-        classes = None
+        classes_map = None
         supercategory_map = None
 
     # Load image metadata
@@ -1395,17 +1459,14 @@ def _parse_coco_detection_annotations(d, extra_attrs=True):
     else:
         annotations = None
 
-    return info, classes, supercategory_map, images, annotations
+    return info, classes_map, supercategory_map, images, annotations
 
 
 def parse_coco_categories(categories):
     """Parses the COCO categories list.
 
-    The returned ``classes`` contains all class IDs from ``[0, max_id]``,
-    inclusive.
-
     Args:
-        categories: a dict of the form::
+        categories: a list of dict of the form::
 
             [
                 ...
@@ -1422,25 +1483,15 @@ def parse_coco_categories(categories):
     Returns:
         a tuple of
 
-        -   classes: a list of classes
+        -   classes_map: a dict mapping class IDs to labels
         -   supercategory_map: a dict mapping class labels to category dicts
     """
-    cat_map = {c["id"]: c for c in categories}
+    classes_map = {
+        c["id"]: c["name"] if "name" in c else str(c["id"]) for c in categories
+    }
+    supercategory_map = {c["name"]: c for c in categories}
 
-    classes = []
-    supercategory_map = {}
-    for cat_id in range(max(cat_map, default=-1) + 1):
-        category = cat_map.get(cat_id, None)
-        try:
-            name = category["name"]
-        except:
-            name = str(cat_id)
-
-        classes.append(name)
-        if category is not None:
-            supercategory_map[name] = category
-
-    return classes, supercategory_map
+    return classes_map, supercategory_map
 
 
 def download_coco_dataset_split(
@@ -1485,9 +1536,8 @@ def download_coco_dataset_split(
             -   the path to a text (newline-separated), JSON, or CSV file
                 containing the list of image IDs to load in either of the first
                 two formats
-        num_workers (None): the number of processes to use when downloading
-            individual images. By default, ``multiprocessing.cpu_count()`` is
-            used
+        num_workers (None): a suggested number of threads to use when
+            downloading individual images
         shuffle (False): whether to randomly shuffle the order in which samples
             are chosen for partial downloads
         seed (None): a random seed to use when shuffling
@@ -1617,12 +1667,15 @@ def download_coco_dataset_split(
         d = etas.load_json(full_anno_path)
         (
             _,
-            all_classes,
+            all_classes_map,
             _,
             images,
             annotations,
         ) = _parse_coco_detection_annotations(d, extra_attrs=True)
 
+        if all_classes_map is not None:
+            all_classes = _to_classes(all_classes_map)
+
         if image_ids is not None:
             # Start with specific images
             image_ids = _parse_image_ids(image_ids, images, split=split)
@@ -1633,7 +1686,7 @@ def download_coco_dataset_split(
         if classes is not None:
             # Filter by specified classes
             all_ids, any_ids = _get_images_with_classes(
-                image_ids, annotations, classes, all_classes
+                image_ids, annotations, classes, all_classes_map
             )
         else:
             all_ids = image_ids
@@ -1699,7 +1752,7 @@ def download_coco_dataset_split(
     num_samples = len(downloaded_filenames)  # total downloaded
 
     #
-    # Write usable annotations file to `anno_path`, if necesary
+    # Write usable annotations file to `anno_path`, if necessary
     #
 
     if not os.path.isfile(anno_path):
@@ -1711,7 +1764,8 @@ def download_coco_dataset_split(
 
             categories = d.get("categories", None)
             if categories is not None:
-                all_classes, _ = parse_coco_categories(categories)
+                all_classes_map, _ = parse_coco_categories(categories)
+                all_classes = _to_classes(all_classes_map)
             else:
                 all_classes = None
 
@@ -1797,7 +1851,7 @@ def _parse_include_license(include_license):
 
 
 def _get_matching_image_ids(
-    all_classes,
+    classes_map,
     images,
     annotations,
     image_ids=None,
@@ -1813,7 +1867,7 @@ def _get_matching_image_ids(
 
     if classes is not None:
         all_ids, any_ids = _get_images_with_classes(
-            image_ids, annotations, classes, all_classes
+            image_ids, annotations, classes, classes_map
         )
     else:
         all_ids = image_ids
@@ -1852,8 +1906,7 @@ def _get_existing_ids(images_dir, images, image_ids):
 
 
 def _download_images(images_dir, image_ids, images, num_workers):
-    if num_workers is None:
-        num_workers = multiprocessing.cpu_count()
+    num_workers = fou.recommend_thread_pool_workers(num_workers)
 
     tasks = []
     for image_id in image_ids:
@@ -1882,7 +1935,7 @@ def _do_download(args):
 
 
 def _get_images_with_classes(
-    image_ids, annotations, target_classes, all_classes
+    image_ids, annotations, target_classes, classes_map
 ):
     if annotations is None:
         logger.warning("Dataset is unlabeled; ignoring classes requirement")
@@ -1891,11 +1944,12 @@ def _get_images_with_classes(
     if etau.is_str(target_classes):
         target_classes = [target_classes]
 
-    bad_classes = [c for c in target_classes if c not in all_classes]
+    labels_map_rev = {c: i for i, c in classes_map.items()}
+
+    bad_classes = [c for c in target_classes if c not in labels_map_rev]
     if bad_classes:
         raise ValueError("Unsupported classes: %s" % bad_classes)
 
-    labels_map_rev = _to_labels_map_rev(all_classes)
     class_ids = {labels_map_rev[c] for c in target_classes}
 
     all_ids = []
@@ -1980,47 +2034,46 @@ def _load_image_ids_json(json_path):
     return [_id for _id in etas.load_json(json_path)]
 
 
-def _make_images_list(images_dir):
-    logger.info("Computing image metadata for '%s'", images_dir)
-
-    image_paths = foud.parse_images_dir(images_dir)
-
-    images = []
-    with fou.ProgressBar() as pb:
-        for idx, image_path in pb(enumerate(image_paths)):
-            metadata = fom.ImageMetadata.build_for(image_path)
-            images.append(
-                {
-                    "id": idx,
-                    "file_name": os.path.basename(image_path),
-                    "height": metadata.height,
-                    "width": metadata.width,
-                    "license": None,
-                    "coco_url": None,
-                }
-            )
-
-    return images
-
-
 def _to_labels_map_rev(classes):
-    return {c: i for i, c in enumerate(classes)}
+    return {c: i for i, c in enumerate(classes, 1)}
 
 
-def _get_matching_objects(coco_objects, target_classes, all_classes):
-    if etau.is_str(target_classes):
-        target_classes = [target_classes]
+def _to_classes(classes_map):
+    return [classes_map[i] for i in sorted(classes_map.keys())]
 
-    labels_map_rev = _to_labels_map_rev(all_classes)
-    class_ids = {labels_map_rev[c] for c in target_classes}
 
+def _get_class_ids(classes, classes_map):
+    if etau.is_str(classes):
+        classes = [classes]
+
+    labels_map_rev = {c: i for i, c in classes_map.items()}
+    class_ids = {labels_map_rev[c] for c in classes}
+
+    return class_ids
+
+
+def _get_matching_objects(coco_objects, class_ids):
     return [obj for obj in coco_objects if obj.category_id in class_ids]
 
 
+def _parse_categories(categories, classes=None):
+    classes_map, _ = parse_coco_categories(categories)
+
+    if classes is None:
+        return {c: i for i, c in classes_map.items()}
+
+    if etau.is_str(classes):
+        classes = {classes}
+    else:
+        classes = set(classes)
+
+    return {c: i for i, c in classes_map.items() if c in classes}
+
+
 def _coco_objects_to_polylines(
     coco_objects,
     frame_size,
-    classes,
+    classes_map,
     supercategory_map,
     tolerance,
     include_id,
@@ -2029,7 +2082,7 @@ def _coco_objects_to_polylines(
     for coco_obj in coco_objects:
         polyline = coco_obj.to_polyline(
             frame_size,
-            classes=classes,
+            classes_map=classes_map,
             supercategory_map=supercategory_map,
             tolerance=tolerance,
             include_id=include_id,
@@ -2047,7 +2100,7 @@ def _coco_objects_to_polylines(
 def _coco_objects_to_detections(
     coco_objects,
     frame_size,
-    classes,
+    classes_map,
     supercategory_map,
     load_segmentations,
     include_id,
@@ -2056,14 +2109,14 @@ def _coco_objects_to_detections(
     for coco_obj in coco_objects:
         detection = coco_obj.to_detection(
             frame_size,
-            classes=classes,
+            classes_map=classes_map,
             supercategory_map=supercategory_map,
             load_segmentation=load_segmentations,
             include_id=include_id,
         )
 
         if detection is not None and (
-            not load_segmentations or detection.mask is not None
+            not load_segmentations or detection.has_mask
         ):
             detections.append(detection)
 
@@ -2076,7 +2129,7 @@ def _coco_objects_to_detections(
 def _coco_objects_to_keypoints(
     coco_objects,
     frame_size,
-    classes,
+    classes_map,
     supercategory_map,
     include_id,
 ):
@@ -2084,7 +2137,7 @@ def _coco_objects_to_keypoints(
     for coco_obj in coco_objects:
         keypoint = coco_obj.to_keypoints(
             frame_size,
-            classes=classes,
+            classes_map=classes_map,
             supercategory_map=supercategory_map,
             include_id=include_id,
         )
@@ -2159,6 +2212,10 @@ def _coco_segmentation_to_mask(segmentation, bbox, frame_size):
     if isinstance(segmentation, list):
         # Polygon -- a single object might consist of multiple parts, so merge
         # all parts into one mask RLE code
+        segmentation = _normalize_coco_segmentation(segmentation)
+        if len(segmentation) == 0:
+            return None
+
         rle = mask_utils.merge(
             mask_utils.frPyObjects(segmentation, height, width)
         )
@@ -2177,6 +2234,22 @@ def _coco_segmentation_to_mask(segmentation, bbox, frame_size):
     ]
 
 
+def _normalize_coco_segmentation(segmentation):
+    # Filter out empty segmentations
+    # For polygons of 4 points (1 pixel), duplicate to convert to valid polygon
+    _segmentation = []
+    for seg in segmentation:
+        if len(seg) == 0:
+            continue
+
+        if len(seg) == 4:
+            seg *= 4
+
+        _segmentation.append(seg)
+
+    return _segmentation
+
+
 def _polyline_to_coco_segmentation(polyline, frame_size, iscrowd="iscrowd"):
     if polyline.get_attribute_value(iscrowd, None):
         seg = polyline.to_segmentation(frame_size=frame_size, target=1)
@@ -2218,12 +2291,37 @@ def _instance_to_coco_segmentation(
 def _make_coco_keypoints(keypoint, frame_size):
     width, height = frame_size
 
-    # @todo true COCO format would set v = 1/2 based on whether the keypoints
-    # lie within the object's segmentation, but we'll be lazy for now
-
     keypoints = []
-    for x, y in keypoint.points:
-        keypoints.extend((int(x * width), int(y * height), 2))
+    num_points = len(keypoint.points)
+    visibility = [None] * num_points
+    if "visible" in keypoint:
+        if isinstance(keypoint.visible, list):
+            if len(keypoint.visible) == num_points:
+                visibility = keypoint.visible
+            else:
+                logger.warning(
+                    "Ignoring 'visible' attribute of length %d for keypoint %s."
+                    " The length does not match the 'points' attribute of"
+                    " length %d."
+                    % (len(keypoint.visible), str(keypoint.id), num_points)
+                )
+        if isinstance(keypoint.visible, int):
+            visibility = [keypoint.visible] * num_points
+
+    for visible, (x, y) in zip(visibility, keypoint.points):
+        if np.isnan(x) or np.isnan(y):
+            _x = 0
+            _y = 0
+            _visible = 0
+        else:
+            _x = int(x * width)
+            _y = int(y * height)
+            _visible = 2
+
+        if visible is not None:
+            _visible = visible
+
+        keypoints.extend((_x, _y, _visible))
 
     return keypoints
 
