@@ -1,17 +1,21 @@
 """
 Classification evaluation.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
+from copy import deepcopy
+import inspect
 import itertools
 import warnings
 
 import numpy as np
 import sklearn.metrics as skm
 
-import fiftyone.core.evaluation as foe
+import eta.core.utils as etau
+
+import fiftyone as fo
 from fiftyone.core.expressions import ViewField as F
 import fiftyone.core.fields as fof
 import fiftyone.core.labels as fol
@@ -19,7 +23,11 @@ import fiftyone.core.plots as fop
 import fiftyone.core.utils as fou
 import fiftyone.core.validation as fov
 
-from .base import BaseEvaluationResults
+from .base import (
+    BaseEvaluationMethod,
+    BaseEvaluationMethodConfig,
+    BaseClassificationResults,
+)
 
 
 def evaluate_classifications(
@@ -29,7 +37,9 @@ def evaluate_classifications(
     eval_key=None,
     classes=None,
     missing=None,
-    method="simple",
+    method=None,
+    custom_metrics=None,
+    progress=None,
     **kwargs,
 ):
     """Evaluates the classification predictions in the given collection with
@@ -42,7 +52,7 @@ def evaluate_classifications(
     You can customize the evaluation method by passing additional
     parameters for the method's config class as ``kwargs``.
 
-    The supported ``method`` values and their associated configs are:
+    The natively provided ``method`` values and their associated configs are:
 
     -   ``"simple"``: :class:`SimpleEvaluationConfig`
     -   ``"top-k"``: :class:`TopKEvaluationConfig`
@@ -72,19 +82,33 @@ def evaluate_classifications(
             observed ground truth/predicted labels are used
         missing (None): a missing label string. Any None-valued labels are
             given this label for results purposes
-        method ("simple"): a string specifying the evaluation method to use.
-            Supported values are ``("simple", "binary", "top-k")``
+        method (None): a string specifying the evaluation method to use. The
+            supported values are
+            ``fo.evaluation_config.classification_backends.keys()`` and the
+            default is ``fo.evaluation_config.default_classification_backend``
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
+        progress (None): whether to render a progress bar (True/False), use the
+            default value ``fiftyone.config.show_progress_bars`` (None), or a
+            progress callback function to invoke instead
         **kwargs: optional keyword arguments for the constructor of the
             :class:`ClassificationEvaluationConfig` being used
 
     Returns:
         a :class:`ClassificationResults`
     """
+    fov.validate_non_grouped_collection(samples)
     fov.validate_collection_label_fields(
         samples, (pred_field, gt_field), fol.Classification, same_type=True
     )
 
-    config = _parse_config(pred_field, gt_field, method, **kwargs)
+    config = _parse_config(
+        pred_field,
+        gt_field,
+        method,
+        custom_metrics=custom_metrics,
+        **kwargs,
+    )
     eval_method = config.build()
     eval_method.ensure_requirements()
 
@@ -92,14 +116,24 @@ def evaluate_classifications(
     eval_method.register_samples(samples, eval_key)
 
     results = eval_method.evaluate_samples(
-        samples, eval_key=eval_key, classes=classes, missing=missing
+        samples,
+        eval_key=eval_key,
+        classes=classes,
+        missing=missing,
+        progress=progress,
     )
-    eval_method.save_run_results(samples, eval_key, results)
+    eval_method.compute_custom_metrics(samples, eval_key, results)
+
+    if eval_key is not None:
+        eval_method.save_run_results(samples, eval_key, results)
+        eval_method.add_fields_to_sidebar_group(
+            samples, eval_key, omit_fields=(pred_field, gt_field)
+        )
 
     return results
 
 
-class ClassificationEvaluationConfig(foe.EvaluationMethodConfig):
+class ClassificationEvaluationConfig(BaseEvaluationMethodConfig):
     """Base class for configuring :class:`ClassificationEvaluation`
     instances.
 
@@ -108,15 +142,22 @@ class ClassificationEvaluationConfig(foe.EvaluationMethodConfig):
             :class:`fiftyone.core.labels.Classification` instances
         gt_field: the name of the field containing the ground truth
             :class:`fiftyone.core.labels.Classification` instances
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
     """
 
-    def __init__(self, pred_field, gt_field, **kwargs):
+    def __init__(self, pred_field, gt_field, custom_metrics=None, **kwargs):
         super().__init__(**kwargs)
         self.pred_field = pred_field
         self.gt_field = gt_field
+        self.custom_metrics = custom_metrics
+
+    @property
+    def type(self):
+        return "classification"
 
 
-class ClassificationEvaluation(foe.EvaluationMethod):
+class ClassificationEvaluation(BaseEvaluationMethod):
     """Base class for classification evaluation methods.
 
     Args:
@@ -137,7 +178,7 @@ class ClassificationEvaluation(foe.EvaluationMethod):
         raise NotImplementedError("subclass must implement register_samples()")
 
     def evaluate_samples(
-        self, samples, eval_key=None, classes=None, missing=None
+        self, samples, eval_key=None, classes=None, missing=None, progress=None
     ):
         """Evaluates the predicted classifications in the given samples with
         respect to the specified ground truth labels.
@@ -150,13 +191,16 @@ class ClassificationEvaluation(foe.EvaluationMethod):
                 purposes
             missing (None): a missing label string. Any None-valued labels are
                 given this label for results purposes
+            progress (None): whether to render a progress bar (True/False), use
+                the default value ``fiftyone.config.show_progress_bars``
+                (None), or a progress callback function to invoke instead
 
         Returns:
             a :class:`ClassificationResults` instance
         """
         raise NotImplementedError("subclass must implement evaluate_samples()")
 
-    def get_fields(self, samples, eval_key):
+    def get_fields(self, samples, eval_key, include_custom_metrics=True):
         is_frame_field = samples._is_frame_field(self.config.gt_field)
 
         fields = [eval_key]
@@ -164,13 +208,20 @@ class ClassificationEvaluation(foe.EvaluationMethod):
         if is_frame_field:
             fields.append(samples._FRAMES_PREFIX + eval_key)
 
+        if include_custom_metrics:
+            fields.extend(self.get_custom_metric_fields(samples, eval_key))
+
         return fields
 
     def rename(self, samples, eval_key, new_eval_key):
         dataset = samples._dataset
 
-        in_fields = self.get_fields(dataset, eval_key)
-        out_fields = self.get_fields(dataset, new_eval_key)
+        in_fields = self.get_fields(
+            dataset, eval_key, include_custom_metrics=False
+        )
+        out_fields = self.get_fields(
+            dataset, new_eval_key, include_custom_metrics=False
+        )
 
         in_sample_fields, in_frame_fields = fou.split_frame_fields(in_fields)
         out_sample_fields, out_frame_fields = fou.split_frame_fields(
@@ -185,6 +236,10 @@ class ClassificationEvaluation(foe.EvaluationMethod):
             fields = dict(zip(in_frame_fields, out_frame_fields))
             dataset.rename_frame_fields(fields)
 
+        self.rename_custom_metrics(samples, eval_key, new_eval_key)
+
+        samples._rename_sidebar_group(eval_key, new_eval_key)
+
     def cleanup(self, samples, eval_key):
         dataset = samples._dataset
         is_frame_field = samples._is_frame_field(self.config.gt_field)
@@ -194,6 +249,10 @@ class ClassificationEvaluation(foe.EvaluationMethod):
         if is_frame_field:
             dataset.delete_frame_field(eval_key, error_level=1)
 
+        self.cleanup_custom_metrics(samples, eval_key)
+
+        samples._delete_empty_sidebar_group(eval_key)
+
     def _validate_run(self, samples, eval_key, existing_info):
         self._validate_fields_match(eval_key, "pred_field", existing_info)
         self._validate_fields_match(eval_key, "gt_field", existing_info)
@@ -207,6 +266,8 @@ class SimpleEvaluationConfig(ClassificationEvaluationConfig):
             :class:`fiftyone.core.labels.Classification` instances
         gt_field: the name of the field containing the ground truth
             :class:`fiftyone.core.labels.Classification` instances
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
     """
 
     @property
@@ -245,7 +306,7 @@ class SimpleEvaluation(ClassificationEvaluation):
             dataset.add_sample_field(eval_key, fof.BooleanField)
 
     def evaluate_samples(
-        self, samples, eval_key=None, classes=None, missing=None
+        self, samples, eval_key=None, classes=None, missing=None, progress=None
     ):
         pred_field = self.config.pred_field
         gt_field = self.config.gt_field
@@ -314,10 +375,16 @@ class TopKEvaluationConfig(ClassificationEvaluationConfig):
         gt_field: the name of the field containing the ground truth
             :class:`fiftyone.core.labels.Classification` instances
         k (5): the top-k value to use when assessing accuracy
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
     """
 
-    def __init__(self, pred_field, gt_field, k=5, **kwargs):
-        super().__init__(pred_field, gt_field, **kwargs)
+    def __init__(
+        self, pred_field, gt_field, k=5, custom_metrics=None, **kwargs
+    ):
+        super().__init__(
+            pred_field, gt_field, custom_metrics=custom_metrics, **kwargs
+        )
         self.k = k
 
     @property
@@ -360,7 +427,7 @@ class TopKEvaluation(ClassificationEvaluation):
             dataset.add_sample_field(eval_key, fof.BooleanField)
 
     def evaluate_samples(
-        self, samples, eval_key=None, classes=None, missing=None
+        self, samples, eval_key=None, classes=None, missing=None, progress=None
     ):
         if classes is None:
             raise ValueError(
@@ -512,6 +579,8 @@ class BinaryEvaluationConfig(ClassificationEvaluationConfig):
         pred_field: the name of the field containing the predicted
             :class:`fiftyone.core.labels.Classification` instances
         gt_field: the name of the field containing the ground truth
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
     """
 
     @property
@@ -562,7 +631,7 @@ class BinaryEvaluation(ClassificationEvaluation):
             dataset.add_sample_field(eval_key, fof.StringField)
 
     def evaluate_samples(
-        self, samples, eval_key=None, classes=None, missing=None
+        self, samples, eval_key=None, classes=None, missing=None, progress=None
     ):
         if classes is None or len(classes) != 2:
             raise ValueError(
@@ -655,7 +724,7 @@ class BinaryEvaluation(ClassificationEvaluation):
         return results
 
 
-class ClassificationResults(BaseEvaluationResults):
+class ClassificationResults(BaseClassificationResults):
     """Class that stores the results of a classification evaluation.
 
     Args:
@@ -672,6 +741,7 @@ class ClassificationResults(BaseEvaluationResults):
             observed ground truth/predicted labels are used
         missing (None): a missing label string. Any None-valued labels are
             given this label for evaluation purposes
+        custom_metrics (None): an optional dict of custom metrics
         backend (None): a :class:`ClassificationEvaluation` backend
     """
 
@@ -695,6 +765,7 @@ class BinaryClassificationResults(ClassificationResults):
         weights (None): an optional list of sample weights
         ytrue_ids (None): a list of IDs for the ground truth labels
         ypred_ids (None): a list of IDs for the predicted labels
+        custom_metrics (None): an optional dict of custom metrics
         backend (None): a :class:`ClassificationEvaluation` backend
     """
 
@@ -710,6 +781,7 @@ class BinaryClassificationResults(ClassificationResults):
         weights=None,
         ytrue_ids=None,
         ypred_ids=None,
+        custom_metrics=None,
         backend=None,
     ):
         super().__init__(
@@ -724,6 +796,7 @@ class BinaryClassificationResults(ClassificationResults):
             ypred_ids=ypred_ids,
             classes=classes,
             missing=classes[0],
+            custom_metrics=custom_metrics,
             backend=backend,
         )
 
@@ -843,6 +916,7 @@ class BinaryClassificationResults(ClassificationResults):
         weights = d.get("weights", None)
         ytrue_ids = d.get("ytrue_ids", None)
         ypred_ids = d.get("ypred_ids", None)
+        custom_metrics = d.get("custom_metrics", None)
         return cls(
             samples,
             config,
@@ -854,24 +928,48 @@ class BinaryClassificationResults(ClassificationResults):
             weights=weights,
             ytrue_ids=ytrue_ids,
             ypred_ids=ypred_ids,
+            custom_metrics=custom_metrics,
             **kwargs,
         )
 
 
 def _parse_config(pred_field, gt_field, method, **kwargs):
     if method is None:
-        method = "simple"
+        method = fo.evaluation_config.default_classification_backend
 
-    if method == "simple":
-        return SimpleEvaluationConfig(pred_field, gt_field, **kwargs)
+    custom_metrics = kwargs.get("custom_metrics", None)
+    if etau.is_str(custom_metrics):
+        kwargs["custom_metrics"] = [custom_metrics]
 
-    if method == "binary":
-        return BinaryEvaluationConfig(pred_field, gt_field, **kwargs)
+    if inspect.isclass(method):
+        return method(pred_field, gt_field, **kwargs)
 
-    if method == "top-k":
-        return TopKEvaluationConfig(pred_field, gt_field, **kwargs)
+    backends = fo.evaluation_config.classification_backends
 
-    raise ValueError("Unsupported evaluation method '%s'" % method)
+    if method not in backends:
+        raise ValueError(
+            "Unsupported classification evaluation method '%s'. The available "
+            "methods are %s" % (method, sorted(backends.keys()))
+        )
+
+    params = deepcopy(backends[method])
+
+    config_cls = kwargs.pop("config_cls", None)
+
+    if config_cls is None:
+        config_cls = params.pop("config_cls", None)
+
+    if config_cls is None:
+        raise ValueError(
+            "Classification evaluation method '%s' has no `config_cls`"
+            % method
+        )
+
+    if etau.is_str(config_cls):
+        config_cls = etau.get_class(config_cls)
+
+    params.update(**kwargs)
+    return config_cls(pred_field, gt_field, **params)
 
 
 def _to_binary_scores(y, confs, pos_label):
