@@ -1,7 +1,7 @@
 """
 Open Images-style detection evaluation.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -10,6 +10,7 @@ from copy import deepcopy
 
 import numpy as np
 
+import fiftyone.core.labels as fol
 import fiftyone.core.plots as fop
 import fiftyone.utils.iou as foui
 
@@ -25,11 +26,13 @@ class OpenImagesEvaluationConfig(DetectionEvaluationConfig):
 
     Args:
         pred_field: the name of the field containing the predicted
-            :class:`fiftyone.core.labels.Detections` or
-            :class:`fiftyone.core.labels.Polylines`
+            :class:`fiftyone.core.labels.Detections`,
+            :class:`fiftyone.core.labels.Polylines`, or
+            :class:`fiftyone.core.labels.Keypoints`
         gt_field: the name of the field containing the ground truth
-            :class:`fiftyone.core.labels.Detections` or
-            :class:`fiftyone.core.labels.Polylines`
+            :class:`fiftyone.core.labels.Detections`,
+            :class:`fiftyone.core.labels.Polylines`, or
+            :class:`fiftyone.core.labels.Keypoints`
         iou (None): the IoU threshold to use to determine matches
         classwise (None): whether to only match objects with the same class
             label (True) or allow matches between classes (False)
@@ -41,7 +44,8 @@ class OpenImagesEvaluationConfig(DetectionEvaluationConfig):
             of the provided :class:`fiftyone.core.labels.Polyline` instances
             rather than using their actual geometries
         tolerance (None): a tolerance, in pixels, when generating approximate
-            polylines for instance masks. Typical values are 1-3 pixels
+            polylines for instance masks. Typical values are 1-3 pixels. By
+            default, IoUs are computed directly on the dense pixel masks
         max_preds (None): the maximum number of predicted objects to evaluate
             when computing mAP and PR curves
         error_level (1): the error level to use when manipulating instance
@@ -53,7 +57,7 @@ class OpenImagesEvaluationConfig(DetectionEvaluationConfig):
 
             If ``error_level > 0``, any calculation that raises a geometric
             error will default to an IoU of 0
-        hierarchy (None): an optional dict containing a hierachy of classes for
+        hierarchy (None): an optional dict containing a hierarchy of classes for
             evaluation following the structure
             ``{"LabelName": label, "Subcategory": [{...}, ...]}``
         pos_label_field (None): the name of a field containing image-level
@@ -66,6 +70,8 @@ class OpenImagesEvaluationConfig(DetectionEvaluationConfig):
             labels according to the provided ``hierarchy``
         expand_pred_hierarchy (False): whether to expand predicted objects and
             labels according to the provided ``hierarchy``
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
     """
 
     def __init__(
@@ -85,10 +91,16 @@ class OpenImagesEvaluationConfig(DetectionEvaluationConfig):
         neg_label_field=None,
         expand_gt_hierarchy=True,
         expand_pred_hierarchy=False,
+        custom_metrics=None,
         **kwargs
     ):
         super().__init__(
-            pred_field, gt_field, iou=iou, classwise=classwise, **kwargs
+            pred_field,
+            gt_field,
+            iou=iou,
+            classwise=classwise,
+            custom_metrics=custom_metrics,
+            **kwargs,
         )
 
         self.iscrowd = iscrowd
@@ -217,7 +229,13 @@ class OpenImagesEvaluation(DetectionEvaluation):
         )
 
     def generate_results(
-        self, samples, matches, eval_key=None, classes=None, missing=None
+        self,
+        samples,
+        matches,
+        eval_key=None,
+        classes=None,
+        missing=None,
+        progress=None,
     ):
         """Generates aggregate evaluation results for the samples.
 
@@ -236,6 +254,9 @@ class OpenImagesEvaluation(DetectionEvaluation):
                 purposes
             missing (None): a missing label string. Any unmatched objects are
                 given this label for results purposes
+            progress (None): whether to render a progress bar (True/False), use
+                the default value ``fiftyone.config.show_progress_bars``
+                (None), or a progress callback function to invoke instead
 
         Returns:
             a :class:`OpenImagesDetectionResults`
@@ -275,6 +296,7 @@ class OpenImagesDetectionResults(DetectionResults):
         thresholds (None): an optional dict of per-class decision thresholds
         missing (None): a missing label string. Any unmatched objects are
             given this label for evaluation purposes
+        custom_metrics (None): an optional dict of custom metrics
         backend (None): a :class:`OpenImagesEvaluation` backend
     """
 
@@ -289,6 +311,7 @@ class OpenImagesDetectionResults(DetectionResults):
         classes,
         thresholds=None,
         missing=None,
+        custom_metrics=None,
         backend=None,
     ):
         super().__init__(
@@ -298,6 +321,7 @@ class OpenImagesDetectionResults(DetectionResults):
             matches,
             classes=classes,
             missing=missing,
+            custom_metrics=custom_metrics,
             backend=backend,
         )
 
@@ -534,6 +558,11 @@ def _open_images_evaluation_setup(
                 if config.expand_pred_hierarchy and label != "all":
                     _expand_detection_hierarchy(cats, obj, config, "preds")
 
+    if isinstance(preds, fol.Keypoints):
+        sort_key = lambda p: np.nanmean(p.confidence) if p.confidence else -1
+    else:
+        sort_key = lambda p: p.confidence or -1
+
     # Compute IoUs within each category
     pred_ious = {}
     for objects in cats.values():
@@ -541,7 +570,7 @@ def _open_images_evaluation_setup(
         preds = objects["preds"]
 
         # Highest confidence predictions first
-        preds = sorted(preds, key=lambda p: p.confidence or -1, reverse=True)
+        preds = sorted(preds, key=sort_key, reverse=True)
 
         if max_preds is not None:
             preds = preds[:max_preds]
@@ -552,11 +581,8 @@ def _open_images_evaluation_setup(
         gts = sorted(gts, key=iscrowd)
 
         # Compute ``num_preds x num_gts`` IoUs
-        ious = foui.compute_ious(preds, gts, **iou_kwargs)
-
-        gt_ids = [g.id for g in gts]
-        for pred, gt_ious in zip(preds, ious):
-            pred_ious[pred.id] = list(zip(gt_ids, gt_ious))
+        ious = foui.compute_ious(preds, gts, sparse=True, **iou_kwargs)
+        pred_ious.update(ious)
 
     return cats, pred_ious, iscrowd
 
@@ -575,6 +601,13 @@ def _compute_matches(
 
         # Match each prediction to the highest available IoU ground truth
         for pred in objects["preds"]:
+            if isinstance(pred, fol.Keypoint):
+                pred_conf = (
+                    np.nanmean(pred.confidence) if pred.confidence else None
+                )
+            else:
+                pred_conf = pred.confidence
+
             if pred.id in pred_ious:
                 best_match = None
                 best_match_iou = iou_thresh
@@ -659,7 +692,7 @@ def _compute_matches(
                                 gt.label,
                                 pred.label,
                                 best_match_iou,
-                                pred.confidence,
+                                pred_conf,
                                 gt.id,
                                 pred.id,
                             )
@@ -671,15 +704,15 @@ def _compute_matches(
                             None,
                             pred.label,
                             None,
-                            pred.confidence,
+                            pred_conf,
                             None,
                             pred.id,
                         )
                     )
-            elif pred.label == cat:
+            else:
                 pred[eval_key] = "fp"
                 matches.append(
-                    (None, pred.label, None, pred.confidence, None, pred.id)
+                    (None, pred.label, None, pred_conf, None, pred.id)
                 )
 
         # Leftover GTs are false negatives
@@ -862,7 +895,7 @@ def _build_plain_hierarchy(hierarchy, skip_root=False):
             keyed_parent, keyed_child, children = _build_plain_hierarchy(node)
 
             # Update is not done through dict.update() since some children have
-            # multiple parents in the hiearchy
+            # multiple parents in the hierarchy
             _update_dict(all_keyed_parent, keyed_parent)
             _update_dict(all_keyed_child, keyed_child)
             all_children.update(children)
