@@ -1,27 +1,34 @@
 """
 Regression evaluation.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
-import logging
+
+from copy import deepcopy
+import inspect
 import itertools
-import numbers
+import logging
 
 import numpy as np
 import sklearn.metrics as skm
-from tabulate import tabulate
 
 import eta.core.utils as etau
 
-import fiftyone.core.evaluation as foe
+import fiftyone as fo
 import fiftyone.core.fields as fof
 import fiftyone.core.labels as fol
 import fiftyone.core.plots as fop
 import fiftyone.core.utils as fou
 import fiftyone.core.validation as fov
 
+from .base import (
+    BaseEvaluationMethodConfig,
+    BaseEvaluationMethod,
+    BaseEvaluationResults,
+)
+
 
 logger = logging.getLogger(__name__)
 
@@ -32,7 +39,9 @@ def evaluate_regressions(
     gt_field="ground_truth",
     eval_key=None,
     missing=None,
-    method="simple",
+    method=None,
+    custom_metrics=None,
+    progress=None,
     **kwargs,
 ):
     """Evaluates the regression predictions in the given collection with
@@ -41,7 +50,7 @@ def evaluate_regressions(
     You can customize the evaluation method by passing additional
     parameters for the method's config class as ``kwargs``.
 
-    The supported ``method`` values and their associated configs are:
+    The natively provided ``method`` values and their associated configs are:
 
     -   ``"simple"``: :class:`SimpleEvaluationConfig`
 
@@ -67,19 +76,33 @@ def evaluate_regressions(
         eval_key (None): a string key to use to refer to this evaluation
         missing (None): a missing value. Any None-valued regressions are
             given this value for results purposes
-        method ("simple"): a string specifying the evaluation method to use.
-            Supported values are ``("simple")``
+        method (None): a string specifying the evaluation method to use. The
+            supported values are
+            ``fo.evaluation_config.regression_backends.keys()`` and the default
+            is ``fo.evaluation_config.default_regression_backend``
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
+        progress (None): whether to render a progress bar (True/False), use the
+            default value ``fiftyone.config.show_progress_bars`` (None), or a
+            progress callback function to invoke instead
         **kwargs: optional keyword arguments for the constructor of the
             :class:`RegressionEvaluationConfig` being used
 
     Returns:
         a :class:`RegressionResults`
     """
+    fov.validate_non_grouped_collection(samples)
     fov.validate_collection_label_fields(
         samples, (pred_field, gt_field), fol.Regression, same_type=True
     )
 
-    config = _parse_config(pred_field, gt_field, method, **kwargs)
+    config = _parse_config(
+        pred_field,
+        gt_field,
+        method,
+        custom_metrics=custom_metrics,
+        **kwargs,
+    )
     eval_method = config.build()
     eval_method.ensure_requirements()
 
@@ -87,14 +110,20 @@ def evaluate_regressions(
     eval_method.register_samples(samples, eval_key)
 
     results = eval_method.evaluate_samples(
-        samples, eval_key=eval_key, missing=missing
+        samples, eval_key=eval_key, missing=missing, progress=progress
     )
-    eval_method.save_run_results(samples, eval_key, results)
+    eval_method.compute_custom_metrics(samples, eval_key, results)
+
+    if eval_key is not None:
+        eval_method.save_run_results(samples, eval_key, results)
+        eval_method.add_fields_to_sidebar_group(
+            samples, eval_key, omit_fields=(pred_field, gt_field)
+        )
 
     return results
 
 
-class RegressionEvaluationConfig(foe.EvaluationMethodConfig):
+class RegressionEvaluationConfig(BaseEvaluationMethodConfig):
     """Base class for configuring :class:`RegressionEvaluation` instances.
 
     Args:
@@ -102,15 +131,22 @@ class RegressionEvaluationConfig(foe.EvaluationMethodConfig):
             :class:`fiftyone.core.labels.Regression` instances
         gt_field ("ground_truth"): the name of the field containing the ground
             truth :class:`fiftyone.core.labels.Regression` instances
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
     """
 
-    def __init__(self, pred_field, gt_field, **kwargs):
+    def __init__(self, pred_field, gt_field, custom_metrics=None, **kwargs):
         super().__init__(**kwargs)
         self.pred_field = pred_field
         self.gt_field = gt_field
+        self.custom_metrics = custom_metrics
+
+    @property
+    def type(self):
+        return "regression"
 
 
-class RegressionEvaluation(foe.EvaluationMethod):
+class RegressionEvaluation(BaseEvaluationMethod):
     """Base class for regression evaluation methods.
 
     Args:
@@ -140,7 +176,9 @@ class RegressionEvaluation(foe.EvaluationMethod):
         else:
             dataset.add_sample_field(eval_key, fof.FloatField)
 
-    def evaluate_samples(self, samples, eval_key=None, missing=None):
+    def evaluate_samples(
+        self, samples, eval_key=None, missing=None, progress=None
+    ):
         """Evaluates the regression predictions in the given samples with
         respect to the specified ground truth values.
 
@@ -149,26 +187,36 @@ class RegressionEvaluation(foe.EvaluationMethod):
             eval_key (None): an evaluation key for this evaluation
             missing (None): a missing value. Any None-valued regressions are
                 given this value for results purposes
+            progress (None): whether to render a progress bar (True/False), use
+                the default value ``fiftyone.config.show_progress_bars``
+                (None), or a progress callback function to invoke instead
 
         Returns:
             a :class:`RegressionResults` instance
         """
         pass
 
-    def get_fields(self, samples, eval_key):
+    def get_fields(self, samples, eval_key, include_custom_metrics=True):
         fields = [eval_key]
 
         if samples._is_frame_field(self.config.gt_field):
             prefix = samples._FRAMES_PREFIX + eval_key
             fields.append(prefix)
 
+        if include_custom_metrics:
+            fields.extend(self.get_custom_metric_fields(samples, eval_key))
+
         return fields
 
     def rename(self, samples, eval_key, new_eval_key):
         dataset = samples._dataset
 
-        in_fields = self.get_fields(dataset, eval_key)
-        out_fields = self.get_fields(dataset, new_eval_key)
+        in_fields = self.get_fields(
+            dataset, eval_key, include_custom_metrics=False
+        )
+        out_fields = self.get_fields(
+            dataset, new_eval_key, include_custom_metrics=False
+        )
 
         in_sample_fields, in_frame_fields = fou.split_frame_fields(in_fields)
         out_sample_fields, out_frame_fields = fou.split_frame_fields(
@@ -183,6 +231,10 @@ class RegressionEvaluation(foe.EvaluationMethod):
             fields = dict(zip(in_frame_fields, out_frame_fields))
             dataset.rename_frame_fields(fields)
 
+        self.rename_custom_metrics(samples, eval_key, new_eval_key)
+
+        samples._rename_sidebar_group(eval_key, new_eval_key)
+
     def cleanup(self, samples, eval_key):
         dataset = samples._dataset
 
@@ -192,6 +244,10 @@ class RegressionEvaluation(foe.EvaluationMethod):
         if dataset._is_frame_field(self.config.gt_field):
             dataset.delete_frame_fields(fields, error_level=1)
 
+        self.cleanup_custom_metrics(samples, eval_key)
+
+        samples._delete_empty_sidebar_group(eval_key)
+
     def _validate_run(self, samples, eval_key, existing_info):
         self._validate_fields_match(eval_key, "pred_field", existing_info)
         self._validate_fields_match(eval_key, "gt_field", existing_info)
@@ -209,10 +265,21 @@ class SimpleEvaluationConfig(RegressionEvaluationConfig):
             sample/frame-level error data. Supported values are
             ``("squared_error", "absolute_error")`` or any function that
             accepts two scalar arguments ``(ypred, ytrue)``
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
     """
 
-    def __init__(self, pred_field, gt_field, metric="squared_error", **kwargs):
-        super().__init__(pred_field, gt_field, **kwargs)
+    def __init__(
+        self,
+        pred_field,
+        gt_field,
+        metric="squared_error",
+        custom_metrics=None,
+        **kwargs,
+    ):
+        super().__init__(
+            pred_field, gt_field, custom_metrics=custom_metrics, **kwargs
+        )
         self._metric = metric
 
     @property
@@ -234,7 +301,9 @@ class SimpleEvaluation(RegressionEvaluation):
         config: a :class:`SimpleEvaluationConfig`
     """
 
-    def evaluate_samples(self, samples, eval_key=None, missing=None):
+    def evaluate_samples(
+        self, samples, eval_key=None, missing=None, progress=None
+    ):
         metric = self.config._metric
 
         if metric == "squared_error":
@@ -323,7 +392,7 @@ class SimpleEvaluation(RegressionEvaluation):
         return results
 
 
-class RegressionResults(foe.EvaluationResults):
+class RegressionResults(BaseEvaluationResults):
     """Class that stores the results of a regression evaluation.
 
     Args:
@@ -340,6 +409,7 @@ class RegressionResults(foe.EvaluationResults):
             regressions
         missing (None): a missing value. Any None-valued regressions are
             given this value for results purposes
+        custom_metrics (None): an optional dict of custom metrics
         backend (None): a :class:`RegressionEvaluation` backend
     """
 
@@ -353,9 +423,16 @@ class RegressionResults(foe.EvaluationResults):
         confs=None,
         ids=None,
         missing=None,
+        custom_metrics=None,
         backend=None,
     ):
-        super().__init__(samples, config, eval_key, backend=backend)
+        super().__init__(
+            samples,
+            config,
+            eval_key,
+            custom_metrics=custom_metrics,
+            backend=backend,
+        )
 
         ytrue, ypred, confs, ids = _parse_values(
             ytrue, ypred, confs, ids, missing=missing
@@ -381,6 +458,8 @@ class RegressionResults(foe.EvaluationResults):
         -   Max error: :func:`sklearn:sklearn.metrics.max_error`
         -   Support: the number of examples
 
+        Also includes any custom metrics from :attr:`custom_metrics`.
+
         Args:
             weights (None): an optional list of weights for each example
 
@@ -410,7 +489,7 @@ class RegressionResults(foe.EvaluationResults):
             max_error = 0.0
             support = 0
 
-        return {
+        metrics = {
             "mean_squared_error": mse,
             "root_mean_squared_error": rmse,
             "mean_absolute_error": mae,
@@ -421,15 +500,19 @@ class RegressionResults(foe.EvaluationResults):
             "support": support,
         }
 
+        metrics.update(self._get_custom_metrics())
+
+        return metrics
+
     def print_metrics(self, weights=None, digits=2):
-        """Prints the regression metrics computed via :meth:`metrics`.
+        """Prints the metrics computed via :meth:`metrics`.
 
         Args:
             weights (None): an optional list of weights for each example
             digits (2): the number of digits of precision to print
         """
         metrics = self.metrics(weights=weights)
-        _print_dict_as_table(metrics, digits)
+        self._print_metrics(metrics, digits=digits)
 
     def plot_results(
         self, labels=None, sizes=None, backend="plotly", **kwargs
@@ -496,6 +579,7 @@ class RegressionResults(foe.EvaluationResults):
         confs = d.get("confs", None)
         ids = d.get("ids", None)
         missing = d.get("missing", None)
+        custom_metrics = d.get("custom_metrics", None)
         return cls(
             samples,
             config,
@@ -505,18 +589,47 @@ class RegressionResults(foe.EvaluationResults):
             confs=confs,
             ids=ids,
             missing=missing,
+            custom_metrics=custom_metrics,
             **kwargs,
         )
 
 
 def _parse_config(pred_field, gt_field, method, **kwargs):
     if method is None:
-        method = "simple"
+        method = fo.evaluation_config.default_regression_backend
 
-    if method == "simple":
-        return SimpleEvaluationConfig(pred_field, gt_field, **kwargs)
+    custom_metrics = kwargs.get("custom_metrics", None)
+    if etau.is_str(custom_metrics):
+        kwargs["custom_metrics"] = [custom_metrics]
 
-    raise ValueError("Unsupported evaluation method '%s'" % method)
+    if inspect.isclass(method):
+        return method(pred_field, gt_field, **kwargs)
+
+    backends = fo.evaluation_config.regression_backends
+
+    if method not in backends:
+        raise ValueError(
+            "Unsupported regression evaluation method '%s'. The available "
+            "methods are %s" % (method, sorted(backends.keys()))
+        )
+
+    params = deepcopy(backends[method])
+
+    config_cls = kwargs.pop("config_cls", None)
+
+    if config_cls is None:
+        config_cls = params.pop("config_cls", None)
+
+    if config_cls is None:
+        raise ValueError(
+            "Regression evaluation method '%s' has no `config_cls`" % method
+        )
+
+    if etau.is_str(config_cls):
+        config_cls = etau.get_class(config_cls)
+
+    params.update(**kwargs)
+    return config_cls(pred_field, gt_field, **params)
 
 
 def _safe_mean(values):
@@ -569,18 +682,3 @@ def _parse_values(ytrue, ypred, *args, missing=None):
         args = [np.asarray(a) if a is not None else a for a in args]
 
     return (_ytrue, _ypred, *args)
-
-
-def _print_dict_as_table(d, digits):
-    fmt = "%%.%df" % digits
-    records = []
-    for k, v in d.items():
-        k = k.replace("_", " ")
-        if isinstance(v, numbers.Integral):
-            v = str(v)
-        else:
-            v = fmt % v
-
-        records.append((k, v))
-
-    print(tabulate(records, tablefmt="plain", numalign="left"))
