@@ -1,25 +1,33 @@
 """
 Segmentation evaluation.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
+from copy import deepcopy
 import logging
+import inspect
+import itertools
 import warnings
 
 import numpy as np
 import sklearn.metrics as skm
 
 import eta.core.image as etai
+import eta.core.utils as etau
 
-import fiftyone.core.evaluation as foe
+import fiftyone as fo
 import fiftyone.core.fields as fof
 import fiftyone.core.labels as fol
 import fiftyone.core.utils as fou
 import fiftyone.core.validation as fov
 
-from .base import BaseEvaluationResults
+from .base import (
+    BaseEvaluationMethod,
+    BaseEvaluationMethodConfig,
+    BaseClassificationResults,
+)
 
 
 logger = logging.getLogger(__name__)
@@ -31,7 +39,9 @@ def evaluate_segmentations(
     gt_field="ground_truth",
     eval_key=None,
     mask_targets=None,
-    method="simple",
+    method=None,
+    custom_metrics=None,
+    progress=None,
     **kwargs,
 ):
     """Evaluates the specified semantic segmentation masks in the given
@@ -45,7 +55,7 @@ def evaluate_segmentations(
     configured by passing additional parameters for the method's config class
     as ``kwargs``.
 
-    The supported ``method`` values and their associated configs are:
+    The natively provided ``method`` values and their associated configs are:
 
     -   ``"simple"``: :class:`SimpleEvaluationConfig`
 
@@ -78,19 +88,33 @@ def evaluate_segmentations(
         eval_key (None): an evaluation key to use to refer to this evaluation
         mask_targets (None): a dict mapping pixel values or RGB hex strings to
             labels. If not provided, the observed values are used as labels
-        method ("simple"): a string specifying the evaluation method to use.
-            Supported values are ``("simple")``
+        method (None): a string specifying the evaluation method to use. The
+            supported values are
+            ``fo.evaluation_config.segmentation_backends.keys()`` and the
+            default is ``fo.evaluation_config.default_segmentation_backend``
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
+        progress (None): whether to render a progress bar (True/False), use the
+            default value ``fiftyone.config.show_progress_bars`` (None), or a
+            progress callback function to invoke instead
         **kwargs: optional keyword arguments for the constructor of the
             :class:`SegmentationEvaluationConfig` being used
 
     Returns:
         a :class:`SegmentationResults`
     """
+    fov.validate_non_grouped_collection(samples)
     fov.validate_collection_label_fields(
         samples, (pred_field, gt_field), fol.Segmentation, same_type=True
     )
 
-    config = _parse_config(pred_field, gt_field, method, **kwargs)
+    config = _parse_config(
+        pred_field,
+        gt_field,
+        method,
+        custom_metrics=custom_metrics,
+        **kwargs,
+    )
     eval_method = config.build()
     eval_method.ensure_requirements()
 
@@ -98,30 +122,56 @@ def evaluate_segmentations(
     eval_method.register_samples(samples, eval_key)
 
     results = eval_method.evaluate_samples(
-        samples, eval_key=eval_key, mask_targets=mask_targets
+        samples,
+        eval_key=eval_key,
+        mask_targets=mask_targets,
+        progress=progress,
     )
-    eval_method.save_run_results(samples, eval_key, results)
+    eval_method.compute_custom_metrics(samples, eval_key, results)
+
+    if eval_key is not None:
+        eval_method.save_run_results(samples, eval_key, results)
+        eval_method.add_fields_to_sidebar_group(
+            samples, eval_key, omit_fields=(pred_field, gt_field)
+        )
 
     return results
 
 
-class SegmentationEvaluationConfig(foe.EvaluationMethodConfig):
+class SegmentationEvaluationConfig(BaseEvaluationMethodConfig):
     """Base class for configuring :class:`SegmentationEvaluation` instances.
 
     Args:
         pred_field: the name of the field containing the predicted
             :class:`fiftyone.core.labels.Segmentation` instances
-        gt_field ("ground_truth"): the name of the field containing the ground
-            truth :class:`fiftyone.core.labels.Segmentation` instances
+        gt_field: the name of the field containing the ground truth
+            :class:`fiftyone.core.labels.Segmentation` instances
+        compute_dice (False): whether to compute the Dice coefficient for each
+            sample
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
     """
 
-    def __init__(self, pred_field, gt_field, **kwargs):
+    def __init__(
+        self,
+        pred_field,
+        gt_field,
+        compute_dice=False,
+        custom_metrics=None,
+        **kwargs,
+    ):
         super().__init__(**kwargs)
         self.pred_field = pred_field
         self.gt_field = gt_field
+        self.compute_dice = compute_dice
+        self.custom_metrics = custom_metrics
+
+    @property
+    def type(self):
+        return "segmentation"
 
 
-class SegmentationEvaluation(foe.EvaluationMethod):
+class SegmentationEvaluation(BaseEvaluationMethod):
     """Base class for segmentation evaluation methods.
 
     Args:
@@ -158,7 +208,15 @@ class SegmentationEvaluation(foe.EvaluationMethod):
             dataset.add_frame_field(pre_field, fof.FloatField)
             dataset.add_frame_field(rec_field, fof.FloatField)
 
-    def evaluate_samples(self, samples, eval_key=None, mask_targets=None):
+        if self.config.compute_dice:
+            dice_field = "%s_dice" % eval_key
+            dataset.add_sample_field(dice_field, fof.FloatField)
+            if processing_frames:
+                dataset.add_frame_field(dice_field, fof.FloatField)
+
+    def evaluate_samples(
+        self, samples, eval_key=None, mask_targets=None, progress=None
+    ):
         """Evaluates the predicted segmentation masks in the given samples with
         respect to the specified ground truth masks.
 
@@ -169,13 +227,16 @@ class SegmentationEvaluation(foe.EvaluationMethod):
                 contain a subset of the possible classes if you wish to
                 evaluate a subset of the semantic classes. By default, the
                 observed pixel values are used as labels
+            progress (None): whether to render a progress bar (True/False), use
+                the default value ``fiftyone.config.show_progress_bars``
+                (None), or a progress callback function to invoke instead
 
         Returns:
             a :class:`SegmentationResults` instance
         """
         pass
 
-    def get_fields(self, samples, eval_key):
+    def get_fields(self, samples, eval_key, include_custom_metrics=True):
         processing_frames = samples._is_frame_field(self.config.gt_field)
 
         fields = [
@@ -184,6 +245,9 @@ class SegmentationEvaluation(foe.EvaluationMethod):
             "%s_recall" % eval_key,
         ]
 
+        if self.config.compute_dice:
+            fields.append("%s_dice" % eval_key)
+
         if processing_frames:
             prefix = samples._FRAMES_PREFIX + eval_key
             fields.extend(
@@ -194,13 +258,23 @@ class SegmentationEvaluation(foe.EvaluationMethod):
                 ]
             )
 
+            if self.config.compute_dice:
+                fields.append("%s_dice" % prefix)
+
+        if include_custom_metrics:
+            fields.extend(self.get_custom_metric_fields(samples, eval_key))
+
         return fields
 
     def rename(self, samples, eval_key, new_eval_key):
         dataset = samples._dataset
 
-        in_fields = self.get_fields(dataset, eval_key)
-        out_fields = self.get_fields(dataset, new_eval_key)
+        in_fields = self.get_fields(
+            dataset, eval_key, include_custom_metrics=False
+        )
+        out_fields = self.get_fields(
+            dataset, new_eval_key, include_custom_metrics=False
+        )
 
         in_sample_fields, in_frame_fields = fou.split_frame_fields(in_fields)
         out_sample_fields, out_frame_fields = fou.split_frame_fields(
@@ -215,6 +289,10 @@ class SegmentationEvaluation(foe.EvaluationMethod):
             fields = dict(zip(in_frame_fields, out_frame_fields))
             dataset.rename_frame_fields(fields)
 
+        self.rename_custom_metrics(samples, eval_key, new_eval_key)
+
+        samples._rename_sidebar_group(eval_key, new_eval_key)
+
     def cleanup(self, samples, eval_key):
         dataset = samples._dataset
         processing_frames = samples._is_frame_field(self.config.gt_field)
@@ -225,11 +303,18 @@ class SegmentationEvaluation(foe.EvaluationMethod):
             "%s_recall" % eval_key,
         ]
 
+        if self.config.compute_dice:
+            fields.append("%s_dice" % eval_key)
+
         dataset.delete_sample_fields(fields, error_level=1)
 
         if processing_frames:
             dataset.delete_frame_fields(fields, error_level=1)
 
+        self.cleanup_custom_metrics(samples, eval_key)
+
+        samples._delete_empty_sidebar_group(eval_key)
+
     def _validate_run(self, samples, eval_key, existing_info):
         self._validate_fields_match(eval_key, "pred_field", existing_info)
         self._validate_fields_match(eval_key, "gt_field", existing_info)
@@ -243,28 +328,45 @@ class SimpleEvaluationConfig(SegmentationEvaluationConfig):
             :class:`fiftyone.core.labels.Segmentation` instances
         gt_field: the name of the field containing the ground truth
             :class:`fiftyone.core.labels.Segmentation` instances
+        compute_dice (False): whether to compute the Dice coefficient for each
+            sample
         bandwidth (None): an optional bandwidth along the contours of the
             ground truth masks to which to restrict attention when computing
             accuracies. A typical value for this parameter is 5 pixels. By
             default, the entire masks are evaluated
         average ("micro"): the averaging strategy to use when populating
             precision and recall numbers on each sample
+        custom_metrics (None): an optional list of custom metrics to compute
+            or dict mapping metric names to kwargs dicts
     """
 
+    def __init__(
+        self,
+        pred_field,
+        gt_field,
+        compute_dice=False,
+        bandwidth=None,
+        average="micro",
+        custom_metrics=None,
+        **kwargs,
+    ):
+        super().__init__(
+            pred_field,
+            gt_field,
+            compute_dice=compute_dice,
+            custom_metrics=custom_metrics,
+            **kwargs,
+        )
+        self.bandwidth = bandwidth
+        self.average = average
+
     @property
     def method(self):
         return "simple"
 
-    def __init__(
-        self, pred_field, gt_field, bandwidth=None, average="micro", **kwargs
-    ):
-        super().__init__(pred_field, gt_field, **kwargs)
-        self.bandwidth = bandwidth
-        self.average = average
-
 
 class SimpleEvaluation(SegmentationEvaluation):
-    """Stardard pixelwise segmentation evaluation.
+    """Standard pixelwise segmentation evaluation.
 
     This class can optionally be configured to evaluate along only the
     boundaries of the ground truth segmentation masks.
@@ -273,20 +375,24 @@ class SimpleEvaluation(SegmentationEvaluation):
         config: a :class:`SimpleEvaluationConfig`
     """
 
-    def evaluate_samples(self, samples, eval_key=None, mask_targets=None):
+    def evaluate_samples(
+        self, samples, eval_key=None, mask_targets=None, progress=None
+    ):
         pred_field = self.config.pred_field
         gt_field = self.config.gt_field
 
         if mask_targets is not None:
             if fof.is_rgb_mask_targets(mask_targets):
                 mask_targets = {
-                    _hex_to_int(k): v for k, v in mask_targets.items()
+                    fof.hex_to_int(k): v for k, v in mask_targets.items()
                 }
 
             values, classes = zip(*sorted(mask_targets.items()))
         else:
             logger.info("Computing possible mask values...")
-            values, classes = _get_mask_values(samples, pred_field, gt_field)
+            values, classes = _get_mask_values(
+                samples, pred_field, gt_field, progress=progress
+            )
 
         _samples = samples.select_fields([gt_field, pred_field])
         pred_field, processing_frames = samples._handle_frame_field(pred_field)
@@ -294,17 +400,22 @@ class SimpleEvaluation(SegmentationEvaluation):
 
         nc = len(values)
         confusion_matrix = np.zeros((nc, nc), dtype=int)
+        matches = []
 
         bandwidth = self.config.bandwidth
         average = self.config.average
+        compute_dice = self.config.compute_dice
+        save = eval_key is not None
 
-        if eval_key is not None:
+        if save:
             acc_field = "%s_accuracy" % eval_key
             pre_field = "%s_precision" % eval_key
             rec_field = "%s_recall" % eval_key
+            if compute_dice:
+                dice_field = "%s_dice" % eval_key
 
         logger.info("Evaluating segmentations...")
-        for sample in _samples.iter_samples(progress=True):
+        for sample in _samples.iter_samples(progress=progress, autosave=save):
             if processing_frames:
                 images = sample.frames.values()
             else:
@@ -332,26 +443,38 @@ class SimpleEvaluation(SegmentationEvaluation):
                 )
                 sample_conf_mat += image_conf_mat
 
-                # Record frame stats, if requested
-                if processing_frames and eval_key is not None:
+                for i, j in zip(*np.nonzero(image_conf_mat)):
+                    matches.append(
+                        (
+                            classes[i],
+                            classes[j],
+                            int(image_conf_mat[i, j]),
+                            gt_seg.id,
+                            pred_seg.id,
+                        )
+                    )
+
+                if processing_frames and save:
                     facc, fpre, frec = _compute_accuracy_precision_recall(
                         image_conf_mat, values, average
                     )
                     image[acc_field] = facc
                     image[pre_field] = fpre
                     image[rec_field] = frec
+                    if compute_dice:
+                        image[dice_field] = _compute_dice_score(image_conf_mat)
 
             confusion_matrix += sample_conf_mat
 
-            # Record sample stats, if requested
-            if eval_key is not None:
+            if save:
                 sacc, spre, srec = _compute_accuracy_precision_recall(
                     sample_conf_mat, values, average
                 )
                 sample[acc_field] = sacc
                 sample[pre_field] = spre
                 sample[rec_field] = srec
-                sample.save()
+                if compute_dice:
+                    sample[dice_field] = _compute_dice_score(confusion_matrix)
 
         if nc > 0:
             missing = classes[0] if values[0] in (0, "#000000") else None
@@ -364,12 +487,13 @@ class SimpleEvaluation(SegmentationEvaluation):
             eval_key,
             confusion_matrix,
             classes,
+            matches=matches,
             missing=missing,
             backend=self,
         )
 
 
-class SegmentationResults(BaseEvaluationResults):
+class SegmentationResults(BaseClassificationResults):
     """Class that stores the results of a segmentation evaluation.
 
     Args:
@@ -378,7 +502,11 @@ class SegmentationResults(BaseEvaluationResults):
         eval_key: the evaluation key
         pixel_confusion_matrix: a pixel value confusion matrix
         classes: a list of class labels corresponding to the confusion matrix
+        matches (None): a list of
+            ``(gt_label, pred_label, pixel_count, gt_id, pred_id)``
+            matches
         missing (None): a missing (background) class
+        custom_metrics (None): an optional dict of custom metrics
         backend (None): a :class:`SegmentationEvaluation` backend
     """
 
@@ -389,13 +517,23 @@ class SegmentationResults(BaseEvaluationResults):
         eval_key,
         pixel_confusion_matrix,
         classes,
+        matches=None,
         missing=None,
+        custom_metrics=None,
         backend=None,
     ):
         pixel_confusion_matrix = np.asarray(pixel_confusion_matrix)
-        ytrue, ypred, weights = self._parse_confusion_matrix(
-            pixel_confusion_matrix, classes
-        )
+
+        if matches is None:
+            ytrue, ypred, weights = self._parse_confusion_matrix(
+                pixel_confusion_matrix, classes
+            )
+            ytrue_ids = None
+            ypred_ids = None
+        elif matches:
+            ytrue, ypred, weights, ytrue_ids, ypred_ids = zip(*matches)
+        else:
+            ytrue, ypred, weights, ytrue_ids, ypred_ids = [], [], [], [], []
 
         super().__init__(
             samples,
@@ -404,25 +542,53 @@ class SegmentationResults(BaseEvaluationResults):
             ytrue,
             ypred,
             weights=weights,
+            ytrue_ids=ytrue_ids,
+            ypred_ids=ypred_ids,
             classes=classes,
             missing=missing,
+            custom_metrics=custom_metrics,
             backend=backend,
         )
 
         self.pixel_confusion_matrix = pixel_confusion_matrix
 
-    def attributes(self):
-        return ["cls", "pixel_confusion_matrix", "classes", "missing"]
+    def dice_score(self):
+        """Computes the Dice score across all samples in the evaluation.
+
+        Returns:
+            the Dice score in ``[0, 1]``
+        """
+        return _compute_dice_score(self.pixel_confusion_matrix)
 
     @classmethod
     def _from_dict(cls, d, samples, config, eval_key, **kwargs):
+        ytrue = d.get("ytrue", None)
+        ypred = d.get("ypred", None)
+        weights = d.get("weights", None)
+        ytrue_ids = d.get("ytrue_ids", None)
+        ypred_ids = d.get("ypred_ids", None)
+
+        if ytrue is not None and ypred is not None and weights is not None:
+            if ytrue_ids is None:
+                ytrue_ids = itertools.repeat(None)
+
+            if ypred_ids is None:
+                ypred_ids = itertools.repeat(None)
+
+            matches = list(zip(ytrue, ypred, weights, ytrue_ids, ypred_ids))
+        else:
+            # Legacy format segmentations
+            matches = None
+
         return cls(
             samples,
             config,
             eval_key,
             d["pixel_confusion_matrix"],
             d["classes"],
+            matches=matches,
             missing=d.get("missing", None),
+            custom_metrics=d.get("custom_metrics", None),
             **kwargs,
         )
 
@@ -445,22 +611,50 @@ class SegmentationResults(BaseEvaluationResults):
 
 def _parse_config(pred_field, gt_field, method, **kwargs):
     if method is None:
-        method = "simple"
+        method = fo.evaluation_config.default_segmentation_backend
 
-    if method == "simple":
-        return SimpleEvaluationConfig(pred_field, gt_field, **kwargs)
+    custom_metrics = kwargs.get("custom_metrics", None)
+    if etau.is_str(custom_metrics):
+        kwargs["custom_metrics"] = [custom_metrics]
 
-    raise ValueError("Unsupported evaluation method '%s'" % method)
+    if inspect.isclass(method):
+        return method(pred_field, gt_field, **kwargs)
+
+    backends = fo.evaluation_config.segmentation_backends
+
+    if method not in backends:
+        raise ValueError(
+            "Unsupported segmentation evaluation method '%s'. The available "
+            "methods are %s" % (method, sorted(backends.keys()))
+        )
+
+    params = deepcopy(backends[method])
+
+    config_cls = kwargs.pop("config_cls", None)
+
+    if config_cls is None:
+        config_cls = params.pop("config_cls", None)
+
+    if config_cls is None:
+        raise ValueError(
+            "Segmentation evaluation method '%s' has no `config_cls`" % method
+        )
+
+    if etau.is_str(config_cls):
+        config_cls = etau.get_class(config_cls)
+
+    params.update(**kwargs)
+    return config_cls(pred_field, gt_field, **params)
 
 
 def _compute_pixel_confusion_matrix(
     pred_mask, gt_mask, values, bandwidth=None
 ):
     if pred_mask.ndim == 3:
-        pred_mask = _rgb_array_to_int(pred_mask)
+        pred_mask = fof.rgb_array_to_int(pred_mask)
 
     if gt_mask.ndim == 3:
-        gt_mask = _rgb_array_to_int(gt_mask)
+        gt_mask = fof.rgb_array_to_int(gt_mask)
 
     if pred_mask.shape != gt_mask.shape:
         msg = (
@@ -487,6 +681,14 @@ def _compute_pixel_confusion_matrix(
         return np.zeros((num_classes, num_classes), dtype=int)
 
 
+def _compute_dice_score(confusion_matrix):
+    confusion_matrix = np.asarray(confusion_matrix)
+    tp = np.diag(confusion_matrix).sum()
+    fp = confusion_matrix.sum() - tp
+    fn = fp
+    return 2 * tp / (2 * tp + fp + fn)
+
+
 def _extract_contour_band_values(pred_mask, gt_mask, bandwidth):
     band_mask = etai.get_contour_band_mask(gt_mask, bandwidth)
     return pred_mask[band_mask], gt_mask[band_mask]
@@ -504,7 +706,7 @@ def _compute_accuracy_precision_recall(confusion_matrix, values, average):
     return metrics["accuracy"], metrics["precision"], metrics["recall"]
 
 
-def _get_mask_values(samples, pred_field, gt_field):
+def _get_mask_values(samples, pred_field, gt_field, progress=None):
     _samples = samples.select_fields([gt_field, pred_field])
     pred_field, processing_frames = samples._handle_frame_field(pred_field)
     gt_field, _ = samples._handle_frame_field(gt_field)
@@ -512,7 +714,7 @@ def _get_mask_values(samples, pred_field, gt_field):
     values = set()
     is_rgb = False
 
-    for sample in _samples.iter_samples(progress=True):
+    for sample in _samples.iter_samples(progress=progress):
         if processing_frames:
             images = sample.frames.values()
         else:
@@ -525,37 +727,15 @@ def _get_mask_values(samples, pred_field, gt_field):
                     mask = seg.get_mask()
                     if mask.ndim == 3:
                         is_rgb = True
-                        mask = _rgb_array_to_int(mask)
+                        mask = fof.rgb_array_to_int(mask)
 
                     values.update(mask.ravel())
 
     values = sorted(values)
 
     if is_rgb:
-        classes = [_int_to_hex(v) for v in values]
+        classes = [fof.int_to_hex(v) for v in values]
     else:
         classes = [str(v) for v in values]
 
     return values, classes
-
-
-def _rgb_array_to_int(mask):
-    return (
-        np.left_shift(mask[:, :, 0], 16, dtype=int)
-        + np.left_shift(mask[:, :, 1], 8, dtype=int)
-        + mask[:, :, 2]
-    )
-
-
-def _hex_to_int(hex_str):
-    r = int(hex_str[1:3], 16)
-    g = int(hex_str[3:5], 16)
-    b = int(hex_str[5:7], 16)
-    return (r << 16) + (g << 8) + b
-
-
-def _int_to_hex(value):
-    r = (value >> 16) & 255
-    g = (value >> 8) & 255
-    b = value & 255
-    return "#%02x%02x%02x" % (r, g, b)
