@@ -2,13 +2,12 @@
 Utilities for working with datasets in
 `KITTI format <http://www.cvlibs.net/datasets/kitti/eval_object.php>`_.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
 import csv
 import logging
-import multiprocessing
 import struct
 import os
 
@@ -20,6 +19,8 @@ import eta.core.web as etaw
 import fiftyone as fo
 import fiftyone.core.labels as fol
 import fiftyone.core.metadata as fom
+import fiftyone.core.storage as fos
+import fiftyone.core.threed as fo3d
 import fiftyone.core.utils as fou
 import fiftyone.utils.data as foud
 
@@ -176,7 +177,7 @@ class KITTIDetectionDatasetImporter(
         )
 
         if self.labels_path is not None and os.path.isdir(self.labels_path):
-            labels_path = fou.normpath(self.labels_path)
+            labels_path = fos.normpath(self.labels_path)
             labels_paths_map = {
                 os.path.splitext(p)[0]: os.path.join(labels_path, p)
                 for p in etau.list_files(labels_path, recursive=True)
@@ -264,7 +265,7 @@ class KITTIDetectionDatasetExporter(
             and labels file. This argument allows for populating nested
             subdirectories that match the shape of the input paths. The path is
             converted to an absolute path (if necessary) via
-            :func:`fiftyone.core.utils.normalize_path`
+            :func:`fiftyone.core.storage.normalize_path`
         image_format (None): the image format to use when writing in-memory
             images to disk. By default, ``fiftyone.config.default_image_ext``
             is used
@@ -416,6 +417,7 @@ def download_kitti_multiview_dataset(
     scratch_dir=None,
     overwrite=False,
     cleanup=False,
+    num_workers=None,
 ):
     """Downloads and prepares the multiview KITTI dataset.
 
@@ -457,12 +459,14 @@ def download_kitti_multiview_dataset(
         dataset_dir: the directory in which to construct the dataset
         splits (None): the split or list of splits to download. Supported
             values are ``("train", "test")``
-        scratch_dir (None): a scratch directory to use to downoad any necessary
+        scratch_dir (None): a scratch directory to use to download any necessary
             temporary files
-        overwrite (False): whether to redownload/regnerate files if they
+        overwrite (False): whether to redownload/regenerate files if they
             already exist
         cleanup (False): whether to delete the downloaded zips and scratch
             directory
+        num_workers (None): a suggested number of processes to use when
+            converting LiDAR to PCD
     """
     if splits is None:
         splits = ("test", "train")
@@ -529,7 +533,11 @@ def download_kitti_multiview_dataset(
 
     for split in splits:
         split_dir = os.path.join(dataset_dir, split)
-        _prepare_kitti_split(split_dir, overwrite=overwrite)
+        samples_path = os.path.join(split_dir, "samples.json")
+        if overwrite or not os.path.isfile(samples_path):
+            _prepare_kitti_split(
+                split_dir, overwrite=overwrite, num_workers=num_workers
+            )
 
     if cleanup:
         etau.delete_dir(scratch_dir)
@@ -566,7 +574,7 @@ def download_kitti_detection_dataset(
         dataset_dir: the directory in which to construct the dataset
         splits (None): the split or list of splits to download. Supported
             values are ``("train", "test")``
-        scratch_dir (None): a scratch directory to use to downoad any necessary
+        scratch_dir (None): a scratch directory to use to download any necessary
             temporary files
         overwrite (False): whether to redownload the zips if they already exist
         cleanup (False): whether to delete the downloaded zips and scratch
@@ -809,11 +817,7 @@ def _make_kitti_detection_row(detection, frame_size):
 #
 
 
-def _prepare_kitti_split(split_dir, overwrite=False):
-    samples_path = os.path.join(split_dir, "samples.json")
-    if not overwrite and os.path.isfile(samples_path):
-        return
-
+def _prepare_kitti_split(split_dir, overwrite=False, num_workers=None):
     group_field = "group"
     label_field = "ground_truth"
 
@@ -831,6 +835,7 @@ def _prepare_kitti_split(split_dir, overwrite=False):
     left_images_dir = os.path.join(split_dir, "left")
     right_images_dir = os.path.join(split_dir, "right")
     pcd_dir = os.path.join(split_dir, "pcd")
+    fo3d_dir = os.path.join(split_dir, "fo3d")
 
     make_map = lambda d: {
         os.path.splitext(os.path.basename(p))[0]: p
@@ -848,11 +853,14 @@ def _prepare_kitti_split(split_dir, overwrite=False):
         pcd_dir,
         uuids,
         overwrite=overwrite,
+        num_workers=num_workers,
     )
 
+    _write_fo3d_files(pcd_dir, fo3d_dir, overwrite=overwrite)
+
     left_map = make_map(left_images_dir)
     right_map = make_map(right_images_dir)
-    pcd_map = make_map(pcd_dir)
+    fo3d_map = make_map(fo3d_dir)
 
     is_labeled = os.path.isdir(labels_dir)
     if is_labeled:
@@ -867,7 +875,7 @@ def _prepare_kitti_split(split_dir, overwrite=False):
 
             left_filepath = left_map[uuid]
             right_filepath = right_map[uuid]
-            pcd_filepath = pcd_map[uuid]
+            scene_filepath = fo3d_map[uuid]
 
             left_sample = fo.Sample(filepath=left_filepath)
             left_sample[group_field] = group.element("left")
@@ -875,7 +883,7 @@ def _prepare_kitti_split(split_dir, overwrite=False):
             right_sample = fo.Sample(filepath=right_filepath)
             right_sample[group_field] = group.element("right")
 
-            pcd_sample = fo.Sample(filepath=pcd_filepath)
+            pcd_sample = fo.Sample(filepath=scene_filepath)
             pcd_sample[group_field] = group.element("pcd")
 
             if is_labeled:
@@ -909,11 +917,13 @@ def _prepare_kitti_split(split_dir, overwrite=False):
 
             samples.extend([left_sample, right_sample, pcd_sample])
 
+    logger.info("Adding samples...")
     dataset.add_samples(samples)
 
     if is_labeled:
         dataset.compute_metadata()
 
+    logger.info("Writing samples...")
     dataset.export(
         export_dir=split_dir,
         dataset_type=fo.types.FiftyOneDataset,
@@ -926,6 +936,10 @@ def _prepare_kitti_split(split_dir, overwrite=False):
 
 def _load_kitti_annotations(labels_path, frame_size):
     gt2d = load_kitti_detection_annotations(labels_path, frame_size)
+
+    for index, detection in enumerate(gt2d.detections, 1):
+        detection.index = index
+
     gt3d = gt2d.copy()
 
     for detection in gt2d.detections:
@@ -1053,7 +1067,7 @@ def _proj_3d_to_right_camera(detections3d, calib, frame_size):
 
 
 def _convert_velodyne_to_pcd(
-    velodyne_map, calib_map, pcd_dir, uuids, overwrite=False
+    velodyne_map, calib_map, pcd_dir, uuids, overwrite=False, num_workers=None
 ):
     inputs = []
     for uuid in uuids:
@@ -1072,9 +1086,11 @@ def _convert_velodyne_to_pcd(
     etau.ensure_dir(pcd_dir)
 
     logger.info("Converting Velodyne scans to PCD format...")
-    num_workers = multiprocessing.cpu_count()
+    num_workers = fou.recommend_process_pool_workers(num_workers)
     with fou.ProgressBar(total=len(inputs)) as pb:
-        with multiprocessing.Pool(processes=num_workers) as pool:
+        with fou.get_multiprocessing_context().Pool(
+            processes=num_workers
+        ) as pool:
             for _ in pb(pool.imap_unordered(_do_conversion, inputs)):
                 pass
 
@@ -1112,3 +1128,19 @@ def _do_conversion(input):
     pcd.colors = o3d.utility.Vector3dVector(colors)
 
     o3d.io.write_point_cloud(pcd_path, pcd)
+
+
+def _write_fo3d_files(pcd_dir, fo3d_dir, overwrite=False, abs_paths=False):
+    for pcd_filepath in etau.list_files(pcd_dir, abs_paths=True):
+        scene_filepath = os.path.join(
+            fo3d_dir,
+            os.path.splitext(os.path.basename(pcd_filepath))[0] + ".fo3d",
+        )
+
+        if overwrite or not os.path.isfile(scene_filepath):
+            if not abs_paths:
+                pcd_filepath = os.path.relpath(pcd_filepath, fo3d_dir)
+
+            scene = fo3d.Scene(camera=fo3d.PerspectiveCamera(up="Z"))
+            scene.add(fo3d.PointCloud("point cloud", pcd_filepath))
+            scene.write(scene_filepath)
