@@ -2,24 +2,26 @@
 Utilities for working with annotations in
 `Label Studio <https://labelstud.io>`_.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
 from copy import deepcopy
 from datetime import datetime
+from collections import defaultdict
 import itertools
 import json
 import logging
+import os
 from packaging import version
 import random
 import string
 import webbrowser
 
+import os
 from bson import ObjectId
 import numpy as np
 
-import fiftyone as fo
 import fiftyone.core.labels as fol
 import fiftyone.core.media as fom
 import fiftyone.core.utils as fou
@@ -29,6 +31,14 @@ ls = fou.lazy_import(
     "label_studio_sdk",
     callback=lambda: fou.ensure_import("label_studio_sdk>=0.0.13"),
 )
+brush = fou.lazy_import(
+    "label_studio_converter.brush",
+    callback=lambda: fou.ensure_import("label_studio_converter.brush"),
+)
+etree = fou.lazy_import(
+    "lxml.etree",
+    callback=lambda: fou.ensure_import("lxml.etree"),
+)
 
 
 logger = logging.getLogger(__name__)
@@ -90,6 +100,7 @@ class LabelStudioBackend(foua.AnnotationBackend):
     def supported_label_types(self):
         return [
             "classification",
+            "classifications",
             "detection",
             "detections",
             "instance",
@@ -216,12 +227,8 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
                 break
 
         # generate label config
-        assert len(label_schema) == 1
-        _, label_info = label_schema.popitem()
         label_config = generate_labeling_config(
-            media=samples.media_type,
-            label_type=label_info["type"],
-            labels=label_info["classes"],
+            label_schema, samples.media_type
         )
 
         project = self._client.start_project(
@@ -231,8 +238,6 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
 
     def _prepare_tasks(self, samples, label_schema, media_field):
         """Prepares Label Studio tasks for the given data."""
-        samples.compute_metadata()
-
         ids, mime_types, filepaths = samples.values(
             ["id", "metadata.mime_type", media_field]
         )
@@ -247,14 +252,17 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
             for _id, _mime_type, _filepath in zip(ids, mime_types, filepaths)
         ]
 
-        predictions, id_map = {}, {}
+        predictions = {}
+        id_map = {}
         for label_field, label_info in label_schema.items():
+            label_type = label_info["type"]
             if label_info["existing_field"]:
                 predictions[label_field] = {
                     smp.id: export_label_to_label_studio(
                         smp[label_field],
+                        label_type=label_type,
                         full_result={
-                            "from_name": "label",
+                            "from_name": label_field,
                             "to_name": "image",
                             "original_width": smp.metadata["width"],
                             "original_height": smp.metadata["height"],
@@ -281,40 +289,72 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
         Returns:
             a dict mapping ``task_id`` to ``sample_id``
         """
-        files = [
-            (
-                one["source_id"],
-                (
-                    one["source_id"],
-                    open(one[one["media_type"]], "rb"),
-                    one["mime_type"],
-                ),
+        ls_root = os.getenv("LABEL_STUDIO_LOCAL_FILES_DOCUMENT_ROOT", None)
+        ls_local_enabled = os.getenv(
+            "LABEL_STUDIO_LOCAL_FILES_SERVING_ENABLED", None
+        )
+        local_storage_enabled = (
+            ls_root is not None
+            and ls_local_enabled is not None
+            and ls_local_enabled.lower() == "true"
+        )
+
+        # If files are not inside local storage root, we can't use local storage
+        common_prefix = os.path.commonprefix(
+            [one[one["media_type"]] for one in tasks]
+        ).rstrip("/")
+        common_prefix = os.path.dirname(common_prefix)
+
+        if local_storage_enabled and ls_root in common_prefix:
+            logger.debug("Using local LabelStudio storage")
+            project.connect_local_import_storage(common_prefix)
+
+            def _get_file_path(file_path):
+                return "/data/local-files?d=" + os.path.relpath(
+                    file_path, ls_root
+                )
+
+            files_data = [
+                {one["media_type"]: _get_file_path(one[one["media_type"]])}
+                for one in tasks
+            ]
+
+            # Make the request to import tasks using URLs
+            upload_resp = self._client.make_request(
+                "POST",
+                f"/api/projects/{project.id}/import",
+                json=files_data,
             )
-            for one in tasks
-        ]
+        else:
+            logger.debug("Uploading as files")
 
-        # upload files first and get their upload ids
-        upload_resp = self._client.make_request(
-            "POST",
-            f"/api/projects/{project.id}/import",
-            params={"commit_to_project": True},
-            files=files,
-        )
+            files = [
+                (one["source_id"], open(one[one["media_type"]], "rb"))
+                for one in tasks
+            ]
 
-        # create tasks out of the uploaded files
-        payload = json.dumps(
-            {
-                "file_upload_ids": upload_resp.json()["file_upload_ids"],
-                "files_as_tasks_list": False,
-            }
-        )
-        self._client.headers.update({"Content-Type": "application/json"})
-        self._client.make_request(
-            "POST", f"/api/projects/{project.id}/reimport", data=payload
-        )
+            # upload files first and get their upload ids
+            upload_resp = self._client.make_request(
+                "POST",
+                f"/api/projects/{project.id}/import",
+                params={"commit_to_project": True},
+                files=files,
+            )
+
+            # create tasks out of the uploaded files
+            payload = json.dumps(
+                {
+                    "file_upload_ids": upload_resp.json()["file_upload_ids"],
+                    "files_as_tasks_list": False,
+                }
+            )
+            self._client.headers.update({"Content-Type": "application/json"})
+            self._client.make_request(
+                "POST", f"/api/projects/{project.id}/reimport", data=payload
+            )
 
         # get uploaded task ids
-        uploaded_ids = project.get_tasks(only_ids=True)[-len(files) :]
+        uploaded_ids = project.get_tasks(only_ids=True)[-len(tasks) :]
         uploaded_tasks = {
             i: t["source_id"] for i, t in zip(uploaded_ids, tasks)
         }
@@ -343,7 +383,7 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
 
         return list(filter(task_filter, matched_tasks))
 
-    def _import_annotations(self, tasks, task_map, label_type):
+    def _import_annotations(self, tasks, task_map):
         results = {}
         for t in tasks:
             # convert latest annotation results
@@ -357,29 +397,44 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
                 if len(annotations) == 0
                 else sorted(annotations, key=lambda x: x["updated_at"])[-1]
             )
-            if label_type == "keypoints":
-                labels = import_label_studio_annotation(
-                    latest_annotation["result"]
-                )
-            else:
-                labels = [
-                    import_label_studio_annotation(r)
-                    for r in latest_annotation.get("result", [])
-                ]
+            latest_result = latest_annotation.get("result", [])
+            # we need logic below to group keypoints with same label together
+            result_kps = [
+                r for r in latest_result if _result_type(r) == "keypointlabels"
+            ]
+            result_not_kps = [
+                r for r in latest_result if _result_type(r) != "keypointlabels"
+            ]
+            labels = [
+                import_label_studio_annotation(r) for r in result_not_kps
+            ]
+            if result_kps:
+                labels += [import_label_studio_annotation(result_kps)]
+
+            if not labels:
+                continue
 
             # add to dict
-            if labels:
-                label_ids = (
-                    {l.id: l for l in labels}
-                    if not isinstance(labels[0], fol.Regression)
-                    else labels[0]
-                )
-                sample_id = task_map[t["id"]]
-                results[sample_id] = label_ids
+            sample_id = task_map[t["id"]]
+            # we save and pass both id and the name of the label field
+            results[sample_id] = {}
+            for ln, l in labels:
+                if isinstance(l, fol.Classifications):
+                    for classification_obj in l.classifications:
+                        label_id = classification_obj.id
+                        results[sample_id][label_id] = (ln, classification_obj)
+                else:
+                    results[sample_id][l.id] = (ln, l)
 
         return results
 
     def _export_to_label_studio(self, labels, label_type):
+        if label_type == "instances":
+            raise ValueError(
+                "This method does not support uploading instance "
+                "segmentations. Use upload_samples() instead"
+            )
+
         if _LABEL_TYPES[label_type]["multiple"] is None:
             return export_label_to_label_studio(labels)
 
@@ -401,6 +456,9 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
 
         project = self._init_project(config, samples)
 
+        samples.compute_metadata()
+
+        # @todo can we add support for uploading tasks in batches?
         tasks, predictions, id_map = self._prepare_tasks(
             samples,
             config.label_schema,
@@ -433,14 +491,34 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
             project, list(results.uploaded_tasks.keys())
         )
         annotations = {}
+        all_labels = self._import_annotations(
+            labeled_tasks, results.uploaded_tasks
+        )
+        annotations = defaultdict(lambda: defaultdict(dict))
+        for task_id, task_labels in all_labels.items():
+            for label_id, (label_field, label) in task_labels.items():
+                annotations[label_field][task_id][label_id] = label
+
+        # convert to expected format
+        final_annotations = {}
         for label_field, label_info in results.config.label_schema.items():
             return_type = foua._RETURN_TYPES_MAP[label_info["type"]]
-            labels = self._import_annotations(
-                labeled_tasks, results.uploaded_tasks, return_type
-            )
-            annotations.update({label_field: {return_type: labels}})
+            # need to convert masks to detections
+            if label_info["type"] == "instances":
+                new_label_field = defaultdict(dict)
+                for task_id, task_labels in annotations[label_field].items():
+                    for label_id, label in task_labels.items():
+                        detections = label.to_detections(
+                            {255: label.label}
+                        ).detections
+                        detections = detections[0] if detections else None
+                        new_label_field[task_id][label_id] = detections
+                annotations[label_field] = new_label_field
+            final_annotations[label_field] = {
+                return_type: annotations[label_field]
+            }
 
-        return annotations
+        return final_annotations
 
     def upload_predictions(self, project, tasks, sample_labels, label_type):
         """Uploads the given predictions to an existing Label Studio project.
@@ -463,10 +541,7 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
             task_ids: list of task ids
         """
         for t_id in task_ids:
-            self._client.make_request(
-                "DELETE",
-                f"/api/tasks/{t_id}",
-            )
+            self._client.make_request("DELETE", f"/api/tasks/{t_id}")
 
     def delete_project(self, project_id):
         """Deletes the project from Label Studio.
@@ -474,10 +549,7 @@ class LabelStudioAnnotationAPI(foua.AnnotationAPI):
         Args:
             project_id: project id
         """
-        self._client.make_request(
-            "DELETE",
-            f"/api/projects/{project_id}",
-        )
+        self._client.make_request("DELETE", f"/api/projects/{project_id}")
 
 
 class LabelStudioAnnotationResults(foua.AnnotationResults):
@@ -530,45 +602,50 @@ class LabelStudioAnnotationResults(foua.AnnotationResults):
         )
 
 
-def generate_labeling_config(media, label_type, labels=None):
+def generate_labeling_config(label_schema, media):
     """Generates a labeling config for a Label Studio project.
 
     Args:
+        label_shema: dict defining each label type
         media: The media type to label
-        label_type: The type of labels to use
-        labels (None): the labels to use
 
     Returns:
         a labeling config
     """
-    etree = fou.lazy_import(
-        "lxml.etree", callback=lambda: fou.ensure_import("lxml.etree")
-    )
-
     assert media in ["image", "video"]
-    assert (
-        label_type in _LABEL_TYPES.keys()
-        or label_type in foua._RETURN_TYPES_MAP.keys()
+    possible_types = list(_LABEL_TYPES.keys()) + list(
+        foua._RETURN_TYPES_MAP.keys()
     )
 
     # root view and media view
     root = etree.Element("View")
     etree.SubElement(root, media.capitalize(), name=media, value=f"${media}")
 
-    # labels view
-    parent_tag, child_tag, tag_kwargs = _ls_tags_from_type(label_type)
-    # parent_name = child_tag.lower() if child_tag else parent_tag.lower()
-    label_view = etree.SubElement(
-        root, parent_tag, name="label", toName=media, **tag_kwargs
-    )
-    if labels:
-        for one in labels:
-            etree.SubElement(label_view, child_tag, value=one)
+    for label_field, label_info in label_schema.items():
+        assert label_info["type"] in possible_types
+        parent_tag, child_tag, tag_kwargs = _ls_tags_from_type(
+            label_info["type"]
+        )
+        label_view = etree.SubElement(
+            root, parent_tag, name=label_field, toName=media, **tag_kwargs
+        )
+        if "classes" in label_info and label_info["classes"]:
+            for one in label_info["classes"]:
+                etree.SubElement(label_view, child_tag, value=one)
 
     config_str = etree.tostring(root, pretty_print=True).decode()
     return config_str
 
 
+def _result_type(result):
+    if isinstance(result, dict):
+        return result["type"]
+    elif isinstance(result, list):
+        return result[0]["type"]
+    else:
+        raise TypeError("Result type %s is not understood" % type(result))
+
+
 def import_label_studio_annotation(result):
     """Imports an annotation from Label Studio.
 
@@ -580,12 +657,12 @@ def import_label_studio_annotation(result):
     """
     # TODO link keypoints by parent id
     # TODO handle multiple classes for segmentation
+    ls_type = _result_type(result)
+    # we return this to match with label field later
     if isinstance(result, dict):
-        ls_type = result["type"]
+        from_name = result.get("from_name", "from_name")
     elif isinstance(result, list):
-        ls_type = result[0]["type"]
-    else:
-        raise TypeError("Result type %s is not understood" % type(result))
+        from_name = result[0].get("from_name", "from_name")
 
     if ls_type == "choices":
         label = _from_choices(result)
@@ -608,8 +685,7 @@ def import_label_studio_annotation(result):
         label.id = label_id
     except:
         pass
-
-    return label
+    return from_name, label
 
 
 def _update_dict(src_dict, update_dict):
@@ -618,12 +694,17 @@ def _update_dict(src_dict, update_dict):
     return new
 
 
-def export_label_to_label_studio(label, full_result=None):
+def export_label_to_label_studio(label, label_type=None, full_result=None):
     """Exports a label to the Label Studio format.
 
     Args:
         label: a :class:`fiftyone.core.labels.Label` or list of
             :class:`fiftyone.core.labels.Label` instances
+        label_type (None): the label type to use when exporting the annotation.
+            This argument is only used when exporting object detections. By
+            default, only the bounding boxes are exported, but you can pass
+            ``label_type="instances"`` to export them as brush labels encoding
+            the instance masks instead
         full_result (None): if non-empty, return the full Label Studio result
 
     Returns:
@@ -633,11 +714,18 @@ def export_label_to_label_studio(label, full_result=None):
     if label is None:
         result_value = {}
         ls_type = None
-        ids = []
+        ids = None
     elif _check_type(label, fol.Classification, fol.Classifications):
         result_value, ls_type, ids = _to_classification(label)
     elif _check_type(label, fol.Detection, fol.Detections):
-        result_value, ls_type, ids = _to_detection(label)
+        if label_type == "instances":
+            size = (
+                full_result["original_width"],
+                full_result["original_height"],
+            )
+            result_value, ls_type, ids = _to_instance(label, size)
+        else:
+            result_value, ls_type, ids = _to_detection(label)
     elif _check_type(label, fol.Polyline, fol.Polylines):
         result_value, ls_type, ids = _to_polyline(label)
     elif _check_type(label, fol.Keypoint, fol.Keypoints):
@@ -645,9 +733,7 @@ def export_label_to_label_studio(label, full_result=None):
     elif isinstance(label, fol.Segmentation):
         result_value, ls_type, ids = _to_segmentation(label)
     elif isinstance(label, fol.Regression):
-        result_value = {"number": label.value}
-        ls_type = "number"
-        ids = label.id
+        result_value, ls_type, ids = _to_regression(label)
     else:
         raise ValueError("Label type %s is not supported" % type(label))
 
@@ -655,6 +741,7 @@ def export_label_to_label_studio(label, full_result=None):
         # return full LS result
         if not isinstance(result_value, (list, tuple)):
             result_value = [result_value]
+            ids = [ids]
 
         return [
             _update_dict(
@@ -720,6 +807,28 @@ def _to_detection(label):
     return result, ls_type, label.id
 
 
+def _to_instance(label, size):
+    ls_type = "brushlabels"
+
+    if isinstance(label, list):
+        return (
+            [_to_instance(l, size)[0] for l in label],
+            ls_type,
+            [l.id for l in label],
+        )
+
+    if isinstance(label, fol.Detections):
+        return (
+            [_to_instance(l, size)[0] for l in label.detections],
+            ls_type,
+            [l.id for l in label.detections],
+        )
+
+    rle = brush.mask2rle(label.to_segmentation(frame_size=size).get_mask())
+    result = {"format": "rle", "rle": rle, "brushlabels": [label.label]}
+    return result, ls_type, label.id
+
+
 def _to_polyline(label):
     ls_type = "polygonlabels"
 
@@ -775,22 +884,24 @@ def _to_keypoint(label):
 
 
 def _to_segmentation(label):
-    brush = fou.lazy_import(
-        "label_studio_converter.brush",
-        callback=lambda: fou.ensure_import("label_studio_converter.brush"),
-    )
     rle = brush.mask2rle(label.get_mask())
     result = {"format": "rle", "rle": rle, "brushlabels": [label.label]}
     return result, "brushlabels", label.id
 
 
+def _to_regression(label):
+    return {"number": label.value}, "number", label.id
+
+
 def _from_choices(result):
     label_values = result["value"]["choices"]
     if len(label_values) == 1:
         return fol.Classification(label=label_values[0])
 
     # multi-label classification
-    return [fol.Classification(label=l) for l in label_values]
+    return fol.Classifications(
+        classifications=[fol.Classification(label=l) for l in label_values]
+    )
 
 
 def _from_rectanglelabels(result):
@@ -823,21 +934,21 @@ def _from_keypointlabels(result):
         points = _normalize_values(points)
         keypoints.append(fol.Keypoint(label=key, points=points))
 
-    return keypoints
+    if len(keypoints) == 1:
+        return keypoints[0]
+    else:
+        return fol.Keypoints(keypoints=keypoints)
 
 
 def _from_brushlabels(result):
-    brush = fou.lazy_import(
-        "label_studio_converter.brush",
-        callback=lambda: fou.ensure_import("label_studio_converter.brush"),
-    )
-
     label_values = result["value"]["brushlabels"]
     img = brush.decode_rle(result["value"]["rle"])
-    mask = img.reshape(
-        (result["original_height"], result["original_width"], 4)
-    )[:, :, 3]
-    return fol.Segmentation(label=label_values[0], mask=mask)
+    shape = (result["original_height"], result["original_width"], 4)
+    mask = img.reshape(shape)[:, :, 3]
+
+    label = label_values[0]
+    segmentation = fol.Segmentation(label=label, mask=mask)
+    return segmentation
 
 
 def _from_number(result):
@@ -866,7 +977,7 @@ def _ls_tags_from_type(label_type):
 
 
 def _label_class_from_tag(label_type):
-    """Maps Label Studio parent tag to fo.Label."""
+    """Maps Label Studio parent tag to FiftyOne Label type."""
     reverse = {
         v["parent_tag"].lower(): v["label"] for v in _LABEL_TYPES.values()
     }
@@ -938,6 +1049,11 @@ _LABEL_TYPES = {
         child_tag="Label",
         label=fol.Detection,
     ),
+    "instances": dict(
+        parent_tag="BrushLabels",
+        child_tag="Label",
+        label=fol.Detection,
+    ),
     "polylines": dict(
         parent_tag="PolygonLabels",
         child_tag="Label",
