@@ -3,7 +3,7 @@ Utilities for working with the
 `Open Images <https://storage.googleapis.com/openimages/web/index.html>`
 dataset.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -27,6 +27,7 @@ import fiftyone.core.fields as fof
 import fiftyone.core.labels as fol
 import fiftyone.utils.aws as foua
 import fiftyone.utils.data as foud
+import fiftyone.utils.image as foui
 
 
 logger = logging.getLogger(__name__)
@@ -65,7 +66,7 @@ class OpenImagesDatasetImporter(foud.LabeledImageDatasetImporter):
         only_matching (False): whether to only load labels that match the
             ``classes`` or ``attrs`` requirements that you provide (True), or
             to load all labels for samples that match the requirements (False)
-        load_hierarchy (True): whether to load the classes hiearchy and add it
+        load_hierarchy (True): whether to load the classes hierarchy and add it
             to the dataset's ``info`` dictionary
         shuffle (False): whether to randomly shuffle the order in which the
             samples are imported
@@ -405,7 +406,7 @@ class OpenImagesV6DatasetImporter(OpenImagesDatasetImporter):
         only_matching (False): whether to only load labels that match the
             ``classes`` or ``attrs`` requirements that you provide (True), or
             to load all labels for samples that match the requirements (False)
-        load_hierarchy (True): whether to load the classes hiearchy and add it
+        load_hierarchy (True): whether to load the classes hierarchy and add it
             to the dataset's ``info`` dictionary
         shuffle (False): whether to randomly shuffle the order in which the
             samples are imported
@@ -484,7 +485,7 @@ class OpenImagesV7DatasetImporter(OpenImagesDatasetImporter):
         only_matching (False): whether to only load labels that match the
             ``classes`` or ``attrs`` requirements that you provide (True), or
             to load all labels for samples that match the requirements (False)
-        load_hierarchy (True): whether to load the classes hiearchy and add it
+        load_hierarchy (True): whether to load the classes hierarchy and add it
             to the dataset's ``info`` dictionary
         shuffle (False): whether to randomly shuffle the order in which the
             samples are imported
@@ -719,9 +720,8 @@ def download_open_images_split(
             -   the path to a text (newline-separated), JSON, or CSV file
                 containing the list of image IDs to load in either of the first
                 two formats
-        num_workers (None): the number of processes to use when downloading
-            individual images. By default, ``multiprocessing.cpu_count()`` is
-            used
+        num_workers (None): a suggested number of threads to use when
+            downloading individual images
         shuffle (False): whether to randomly shuffle the order in which samples
             are chosen for partial downloads
         seed (None): a random seed to use when shuffling
@@ -1842,7 +1842,7 @@ def _create_segmentations(seg_data, image_id, classes_map, dataset_dir):
             warnings.warn(msg)
             return None
 
-        rgb_mask = etai.read(mask_path)
+        rgb_mask = foui.read(mask_path)
         mask = etai.rgb_to_gray(rgb_mask) > 122
         h, w = mask.shape
         cropped_mask = mask[
