@@ -2,7 +2,7 @@
 Utilities for working with datasets in
 `OpenLABEL format <https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f>`_.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -17,7 +17,7 @@ import eta.core.utils as etau
 
 import fiftyone.core.labels as fol
 import fiftyone.core.metadata as fom
-import fiftyone.core.utils as fou
+import fiftyone.core.storage as fos
 import fiftyone.utils.data as foud
 
 
@@ -218,7 +218,7 @@ class OpenLABELImageDatasetImporter(
         annotations = OpenLABELAnnotations()
 
         if self.labels_path is not None:
-            labels_path = fou.normpath(self.labels_path)
+            labels_path = fos.normpath(self.labels_path)
 
             base_dir = None
             if os.path.isfile(labels_path):
@@ -435,7 +435,7 @@ class OpenLABELVideoDatasetImporter(
         annotations = OpenLABELAnnotations()
 
         if self.labels_path is not None:
-            labels_path = fou.normpath(self.labels_path)
+            labels_path = fos.normpath(self.labels_path)
 
             base_dir = None
             if os.path.isfile(labels_path):
@@ -990,7 +990,7 @@ class OpenLABELStreams(OpenLABELGroup):
         for a given media file identifier.
 
         Args:
-            uri: the unique media file identifer for which to get all stream
+            uri: the unique media file identifier for which to get all stream
                 infos
 
         Returns:
@@ -1169,6 +1169,10 @@ class OpenLABELPoly2D(OpenLABELShape):
 class OpenLABELPoint(OpenLABELShape):
     """An OpenLABEL keypoint."""
 
+    @classmethod
+    def _get_empty_value_for_type(cls, example_type):
+        return example_type()
+
     @classmethod
     def _sort_by_skeleton(cls, points, attrs, label_order, skeleton_order):
         if len(points) != len(label_order):
@@ -1184,6 +1188,8 @@ class OpenLABELPoint(OpenLABELShape):
             if isinstance(v, list) and len(v) == len(points):
                 attrs_to_sort[k] = v
                 sorted_attrs[k] = []
+            elif isinstance(v, list) and len(v) == 0:
+                continue
             else:
                 sorted_attrs[k] = v
 
@@ -1191,7 +1197,9 @@ class OpenLABELPoint(OpenLABELShape):
             if label not in label_order:
                 sorted_points.append([float("nan"), float("nan")])
                 for k in attrs_to_sort:
-                    sorted_attrs[k].append(None)
+                    example_type = type(attrs_to_sort[k][0])
+                    empty_value = cls._get_empty_value_for_type(example_type)
+                    sorted_attrs[k].append(empty_value)
             else:
                 ind = label_order.index(label)
                 sorted_points.append(points[ind])
@@ -2020,7 +2028,7 @@ def _remove_ext(p):
 
 def _merge_frame_labels(sample_labels, frame_labels, seg_type):
     # Add frame labels to sample labels, if there is a key collision, merge the
-    # labels if they are a list field otherewise skip
+    # labels if they are a list field otherwise skip
     for labels in frame_labels.values():
         for name, value in labels.items():
             if name == "detections":
