@@ -1,7 +1,7 @@
 """
 Video utilities.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -18,6 +18,7 @@ import eta.core.video as etav
 
 import fiftyone as fo
 import fiftyone.core.metadata as fom
+import fiftyone.core.storage as fos
 import fiftyone.core.utils as fou
 import fiftyone.core.validation as fov
 
@@ -35,7 +36,7 @@ def extract_clip(
 ):
     """Extracts the specified clip from the video.
 
-    Provide either ``suppport`` or ``timestamps`` to this method.
+    Provide either ``support`` or ``timestamps`` to this method.
 
     When fast=False, the following ffmpeg command is used::
 
@@ -92,14 +93,20 @@ def reencode_videos(
     output_field=None,
     output_dir=None,
     rel_dir=None,
+    update_filepaths=True,
     delete_originals=False,
     skip_failures=False,
     verbose=False,
+    progress=None,
     **kwargs,
 ):
     """Re-encodes the videos in the sample collection as H.264 MP4s that can be
     visualized in the FiftyOne App.
 
+    If no ``output_dir`` is specified and ``delete_originals`` is False, then
+    if a transformation would result in overwriting an existing file with the
+    same filename, the original file is renamed to ``<name>-original.<ext>``.
+
     By default, the re-encoding is performed via the following ``ffmpeg``
     command::
 
@@ -138,13 +145,17 @@ def reencode_videos(
             ``output_dir`` to generate an output path for each video. This
             argument allows for populating nested subdirectories in
             ``output_dir`` that match the shape of the input paths
+        update_filepaths (True): whether to store the output paths on the
+            sample collection
         delete_originals (False): whether to delete the original videos after
-            re-encoding. This parameter has no effect if the videos are being
-            updated in-place
+            re-encoding
         skip_failures (False): whether to gracefully continue without raising
             an error if a video cannot be re-encoded
         verbose (False): whether to log the ``ffmpeg`` commands that are
             executed
+        progress (None): whether to render a progress bar (True/False), use the
+            default value ``fiftyone.config.show_progress_bars`` (None), or a
+            progress callback function to invoke instead
         **kwargs: keyword arguments for ``eta.core.video.FFmpeg(**kwargs)``
     """
     fov.validate_video_collection(sample_collection)
@@ -157,9 +168,11 @@ def reencode_videos(
         output_field=output_field,
         output_dir=output_dir,
         rel_dir=rel_dir,
+        update_filepaths=update_filepaths,
         delete_originals=delete_originals,
         skip_failures=skip_failures,
         verbose=verbose,
+        progress=progress,
         **kwargs,
     )
 
@@ -178,14 +191,20 @@ def transform_videos(
     output_field=None,
     output_dir=None,
     rel_dir=None,
+    update_filepaths=True,
     delete_originals=False,
     skip_failures=False,
     verbose=False,
+    progress=None,
     **kwargs,
 ):
     """Transforms the videos in the sample collection according to the provided
     parameters using ``ffmpeg``.
 
+    If no ``output_dir`` is specified and ``delete_originals`` is False, then
+    if a transformation would result in overwriting an existing file with the
+    same filename, the original file is renamed to ``<name>-original.<ext>``.
+
     In addition to the size and frame rate parameters, if ``reencode == True``,
     the following basic ``ffmpeg`` command structure is used to re-encode the
     videos as H.264 MP4s::
@@ -237,12 +256,17 @@ def transform_videos(
             ``output_dir`` to generate an output path for each video. This
             argument allows for populating nested subdirectories in
             ``output_dir`` that match the shape of the input paths
+        update_filepaths (True): whether to store the output paths on the
+            sample collection
         delete_originals (False): whether to delete the original videos after
             re-encoding
         skip_failures (False): whether to gracefully continue without raising
             an error if a video cannot be transformed
         verbose (False): whether to log the ``ffmpeg`` commands that are
             executed
+        progress (None): whether to render a progress bar (True/False), use the
+            default value ``fiftyone.config.show_progress_bars`` (None), or a
+            progress callback function to invoke instead
         **kwargs: keyword arguments for ``eta.core.video.FFmpeg(**kwargs)``
     """
     fov.validate_video_collection(sample_collection)
@@ -261,9 +285,11 @@ def transform_videos(
         output_field=output_field,
         output_dir=output_dir,
         rel_dir=rel_dir,
+        update_filepaths=update_filepaths,
         delete_originals=delete_originals,
         skip_failures=skip_failures,
         verbose=verbose,
+        progress=progress,
         **kwargs,
     )
 
@@ -279,14 +305,15 @@ def sample_videos(
     max_size=None,
     original_frame_numbers=True,
     force_sample=False,
-    save_filepaths=False,
     media_field="filepath",
     output_field=None,
     output_dir=None,
     rel_dir=None,
+    save_filepaths=False,
     delete_originals=False,
     skip_failures=False,
     verbose=False,
+    progress=None,
     **kwargs,
 ):
     """Samples the videos in the sample collection into directories of
@@ -366,8 +393,6 @@ def sample_videos(
             the frames as 1, 2, ... (False)
         force_sample (False): whether to resample videos whose sampled frames
             already exist
-        save_filepaths (False): whether to save the sampled frame paths in the
-            ``output_field`` field of each frame of the input collection
         media_field ("filepath"): the input field containing the video paths to
             sample
         output_field (None): an optional frame field in which to store the
@@ -377,17 +402,22 @@ def sample_videos(
             the same basename of each video
         rel_dir (None): a relative directory to remove from the filepath of
             each video, if possible. The path is converted to an absolute path
-            (if necessary) via :func:`fiftyone.core.utils.normalize_path`. This
-            argument can be used in conjunction with ``output_dir`` to cause
-            the sampled frames to be written in a nested directory structure
-            within ``output_dir`` matching the shape of the input video's
-            folder structure
+            (if necessary) via :func:`fiftyone.core.storage.normalize_path`.
+            This argument can be used in conjunction with ``output_dir`` to
+            cause the sampled frames to be written in a nested directory
+            structure within ``output_dir`` matching the shape of the input
+            video's folder structure
+        save_filepaths (False): whether to save the sampled frame paths in the
+            ``output_field`` field of each frame of the input collection
         delete_originals (False): whether to delete the original videos after
             sampling
         skip_failures (False): whether to gracefully continue without raising
             an error if a video cannot be sampled
         verbose (False): whether to log the ``ffmpeg`` commands that are
             executed
+        progress (None): whether to render a progress bar (True/False), use the
+            default value ``fiftyone.config.show_progress_bars`` (None), or a
+            progress callback function to invoke instead
         **kwargs: keyword arguments for ``eta.core.video.FFmpeg(**kwargs)``
     """
     fov.validate_video_collection(sample_collection)
@@ -412,6 +442,7 @@ def sample_videos(
         delete_originals=delete_originals,
         skip_failures=skip_failures,
         verbose=verbose,
+        progress=progress,
         **kwargs,
     )
 
@@ -700,20 +731,23 @@ def _transform_videos(
     original_frame_numbers=True,
     reencode=False,
     force_reencode=False,
-    save_filepaths=False,
     media_field="filepath",
     output_field=None,
     output_dir=None,
     rel_dir=None,
+    save_filepaths=False,
+    update_filepaths=True,
     delete_originals=False,
     skip_failures=False,
     verbose=False,
+    progress=None,
     **kwargs,
 ):
     if output_field is None:
         output_field = media_field
 
     diff_field = output_field != media_field
+    save = (save_filepaths and sample_frames) or update_filepaths
 
     if sample_frames:
         reencode = True
@@ -727,8 +761,10 @@ def _transform_videos(
     if frames is None:
         frames = itertools.repeat(None)
 
-    with fou.ProgressBar(total=len(view)) as pb:
-        for sample, _frames in pb(zip(view, frames)):
+    with fou.ProgressBar(total=view, progress=progress) as pb:
+        for sample, _frames in pb(
+            zip(view.iter_samples(autosave=save), frames)
+        ):
             inpath = sample[media_field]
 
             _outpath = _get_outpath(
@@ -782,7 +818,7 @@ def _transform_videos(
                         _frames = range(
                             1, sample.metadata.total_frame_count + 1
                         )
-                    except Exception as e:
+                    except BaseException as e:
                         if not skip_failures:
                             raise
 
@@ -794,11 +830,12 @@ def _transform_videos(
                     if os.path.isfile(frame_path):
                         sample.frames[fn][output_field] = frame_path
 
-                sample.save()
-
-            if (diff_field or outpath != inpath) and not sample_frames:
+            if (
+                update_filepaths
+                and not sample_frames
+                and (diff_field or outpath != inpath)
+            ):
                 sample[output_field] = outpath
-                sample.save()
 
 
 def _transform_video(
@@ -820,8 +857,8 @@ def _transform_video(
     verbose=False,
     **kwargs,
 ):
-    inpath = fou.normalize_path(inpath)
-    outpath = fou.normalize_path(outpath)
+    inpath = fos.normalize_path(inpath)
+    outpath = fos.normalize_path(outpath)
     in_ext = os.path.splitext(inpath)[1]
     out_ext = os.path.splitext(outpath)[1]
 
@@ -837,6 +874,7 @@ def _transform_video(
         max_fps = None
 
     did_transform = False
+    moves = []
 
     try:
         if (
@@ -881,12 +919,11 @@ def _transform_video(
             or in_ext.lower() != out_ext.lower()
         )
 
-        if (inpath == outpath) and should_reencode:
-            _inpath = inpath
-            inpath = etau.make_unique_path(inpath, suffix="-original")
-            etau.move_file(_inpath, inpath)
-
-        diff_path = inpath != outpath
+        if should_reencode and inpath == outpath:
+            orig_path = etau.make_unique_path(inpath, suffix="-original")
+            etau.move_file(inpath, orig_path)
+            moves.append((inpath, orig_path))
+            inpath = orig_path
 
         if frames is not None:
             etav.sample_select_frames(
@@ -903,12 +940,19 @@ def _transform_video(
                 ffmpeg.run(inpath, outpath, verbose=verbose)
 
             did_transform = True
-        elif diff_path:
+        elif inpath != outpath:
             etau.copy_file(inpath, outpath)
 
-        if delete_original and diff_path:
+        if delete_original and inpath != outpath:
             etau.delete_file(inpath)
-    except Exception as e:
+    except BaseException as e:
+        try:
+            # Undo any moves
+            for from_path, to_path in moves:
+                etau.move_file(to_path, from_path)
+        except:
+            pass
+
         if not skip_failures:
             raise
 
@@ -989,9 +1033,9 @@ def _get_outpath(inpath, output_dir=None, rel_dir=None):
         return inpath
 
     if rel_dir is not None:
-        rel_dir = fou.normalize_path(rel_dir)
+        rel_dir = fos.normalize_path(rel_dir)
         filename = os.path.relpath(inpath, rel_dir)
     else:
         filename = os.path.basename(inpath)
 
-    return os.path.join(fou.normalize_path(output_dir), filename)
+    return os.path.join(fos.normalize_path(output_dir), filename)
