@@ -1,7 +1,7 @@
 """
 FiftyOne Zoo Datasets provided by ``tensorflow_datasets``.
 
-| Copyright 2017-2023, Voxel51, Inc.
+| Copyright 2017-2025, Voxel51, Inc.
 | `voxel51.com <https://voxel51.com/>`_
 |
 """
@@ -63,6 +63,10 @@ class MNISTDataset(TFDSDataset):
     def name(self):
         return "mnist"
 
+    @property
+    def license(self):
+        return "CC-BY-SA-3.0"
+
     @property
     def tags(self):
         return ("image", "classification")
@@ -120,6 +124,10 @@ class FashionMNISTDataset(TFDSDataset):
     def name(self):
         return "fashion-mnist"
 
+    @property
+    def license(self):
+        return "MIT"
+
     @property
     def tags(self):
         return ("image", "classification")
@@ -178,6 +186,10 @@ class CIFAR10Dataset(TFDSDataset):
     def name(self):
         return "cifar10"
 
+    @property
+    def license(self):
+        return None
+
     @property
     def tags(self):
         return ("image", "classification")
@@ -236,6 +248,10 @@ class CIFAR100Dataset(TFDSDataset):
     def name(self):
         return "cifar100"
 
+    @property
+    def license(self):
+        return None
+
     @property
     def tags(self):
         return ("image", "classification")
@@ -270,10 +286,10 @@ class CIFAR100Dataset(TFDSDataset):
 class ImageNet2012Dataset(TFDSDataset):
     """The ImageNet 2012 dataset.
 
-    ImageNet, as known as ILSVRC 2012, is an image dataset organized according
+    ImageNet, also known as ILSVRC 2012, is an image dataset organized according
     to the WordNet hierarchy. Each meaningful concept in WordNet, possibly
     described by multiple words or word phrases, is called a "synonym set" or
-    "synset". There are more than 100,000 synsets in WordNet, majority of them
+    "synset". There are more than 100,000 synsets in WordNet, the majority of them
     are nouns (80,000+). ImageNet provides on average 1,000 images to
     illustrate each synset. Images of each concept are quality-controlled and
     human-annotated. In its completion, we hope ImageNet will offer tens of
@@ -328,6 +344,10 @@ class ImageNet2012Dataset(TFDSDataset):
     def name(self):
         return "imagenet-2012"
 
+    @property
+    def license(self):
+        return "https://image-net.org/download"
+
     @property
     def tags(self):
         return ("image", "classification", "manual")
@@ -402,6 +422,10 @@ class VOC2007Dataset(TFDSDataset):
     def name(self):
         return "voc-2007"
 
+    @property
+    def license(self):
+        return None
+
     @property
     def tags(self):
         return ("image", "detection")
@@ -466,6 +490,10 @@ class VOC2012Dataset(TFDSDataset):
     def name(self):
         return "voc-2012"
 
+    @property
+    def license(self):
+        return None
+
     @property
     def tags(self):
         return ("image", "detection")
